/* $Workfile: GimleBurstEx.c $       $Revision: 2 $      $Modtime: 10/23/12 5:00p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    GimleBurstEx.c

    A simple example console application demonstrating the use of the
    DMA Direct Receiving with the OTX Driver using the
    board's Burst device to access the data using DMA.

    The Thor-4-ExpressCard can monitor 2 spans while configure as a 
    monitor only device. In this mode the transmit pin is set to 
    tri-state and receivers from line 2 and 3 are connected to what is 
    normally the transmit signal. This allow the Thor-4-ExpressCard
    to monitor both side of the span.

      Gimle-16-PCI-Plus
      Gimle-16-PCI-Basic-32
      Gimle-16-PCI-Basic-64
      Thor-4-ExpressCard

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 16 Lines which should be connected to other T1/E1 devices.

    The direct receives operate on a highway basis and are
    hosted by the Burst device.

    The Gimle-16-PCI-Plus has four 8Mbps (128TS) burst highways. In this demo 
    we connect all timeslots from the Li0-Li3 to the first burst highway, 
    Li4-Li7 to the second burst highway, and so on. This means that this application
    will receive all 512 timeslots from all 16 E1 spans. 
    
    If the application
    is used in T1 mode, then we still connecte 32 timeslots per span, but only 
    TS0-TS24 will carry data. TS0 will carry the DL-bit (as one bit of TS0), and
    TS1-TS24 will carry the 24 timeslots of the T1 span.

    If this application is used with the Gimle-16-PCI-Basic-32, it uses one 2Mpbs
    (32TS) burst highway. In this application we connect TS10 and TS11 from each 
    of the 16 Lis to the burst highway.

    If this application is used with the Gimle-16-PCI-Basic-64, it uses two 2Mpbs
    (32TS) burst highways. In this application we connect TS10 and TS11 from each 
    of the 16 Lis to burst highway 0, and TS12 and TS13 to burst highway 1.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"


#define DEMO_DEFAULT_LOG_FILE   "GimleBurstEx.log"
#define DEMO_DEFAULT_BOARD_NO   0

#define DEMO_RAW_TX_BUF_SIZE    2000
#define DEMO_BURST_TX_BUF_SIZE  0x8000    // 32k
#define DEMO_DEFAULT_RX_THRESHOLD 500     // in mV

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k

typedef struct Demo2IoInLoopAllTsS
{
  short        bLoopAllTs;
} Demo2IoInLoopAllTsS;

typedef struct DemoEnvT {
  char                   *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int                    coDsp;
  int                    coLi;
  int                    coQuadE1T1;
  int                    coBurstHighways;
  int                    nHighwaySpeed;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  char                   dspFileName[OTX_MAX_PATH];
  int                    boardNo;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxBrdClockSourceE     eClkSource;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(const char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoBurstReceiver(OTX_UINT32 nSourceId);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE        *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;
int         g_nCurrentLi = 0;
OTX_UINT16  g_nDspCommand = 0; // DEMO2_IOCTL_LOOP_ALL_TS which is defined by 55x or 54x
OTX_BOOL    g_nLoopBackMode = OTX_TRUE;


struct sPhysDevs {
  OTX_HANDLE hBoardController;
  OTX_HANDLE hLi[16];
  OTX_HANDLE hQuadLi[4];
  OTX_HANDLE hBurst;
  OTX_HANDLE hVidarAsmController0;  // Optional for Gimle-16-PCI-Plus
  OTX_HANDLE hDsp[4];               // Optional for Gimle-16-PCI-Plus
  OTX_HANDLE hUserApp[4];           // Optional for Gimle-16-PCI-Plus
  OTX_HANDLE hTss;
} g_hPhysDevs;

struct sPipes {
  OTX_HANDLE hLiOut[16];
  OTX_HANDLE hDspOut[4];
} g_hPipes;

// Direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nBurstTsToRecord;  // Default is -1 (The entire Burst highway; All TS)
  OTX_UINT32  nRecvCnt;
  OTX_UINT32  nEventCnt;
  OTX_UINT64  nFrameRefId;

  OTX_BOOL    bGammaTestMode;
  OTX_INT32   nGammaTestErrors[128];
  OTX_UINT8   sqGammaTestData[128];
};

struct sLogicDevs {
  struct sBurstReceiver mBurstReceiver[4];    // One for each possible highway hosted by the Burst device.
                                              // There are 4 8Mbps burst highways on Gimle-16-PCI-Plus.
                                              // There is  1 2Mbps burst highways on Gimle-16-PCI-Basic-32.
                                              // There are 2 2Mbps burst highways on Gimle-16-PCI-Basic-64.
} g_hLogicDevs;

struct sT1E1CfgData {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;




/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_GIMLE_16_PCI_PLUS:
      DemoLogf("Opening Gimle-16-PCI-Plus board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCIE:
      DemoLogf("Opening Gimle-16-PCIe board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
      DemoLogf("Opening Gimle-16-PCI-Basic-32 board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCIE_BASIC_32:
      DemoLogf("Opening Gimle-16-PCIe-Basic-32 board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
      DemoLogf("Opening Gimle-16-PCI-Basic-64 board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCIE_BASIC_64:
      DemoLogf("Opening Gimle-16-PCIe-Basic-64 board\n");
      break;
    case OTX_DEVICE_THOR_4_EXPRESSCARD:
      DemoLogf("Opening Thor-4-ExpressCard as Gimle monitor only\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController); DemoCheckRc(result);

  for (i = 0; i < g_demoEnv.coQuadE1T1; i++) {
    // Get handles to the four quad T1/E1 line interface devices
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(result);
  }
    
  // Get handles to the 16 individual T1/E1 line interface members within the Quad device
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i/4], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
  }

  // The T4E does not have a TSS device
  if(g_demoEnv.nBoardType != OTX_DEVICE_THOR_4_EXPRESSCARD) {
    // Get a handle to the physical Tss device 
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(result);
    result = OtxDrvEnable(g_hPhysDevs.hTss,OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }
  else {
    // Not used on T4E
    g_hPhysDevs.hTss = OTX_INVALID_HANDLE_VALUE;
  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(result);

  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(result);
  DemoLogf("Burst TX disable for GIMLE\n");
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SEND_DISABLED,  OTX_TRUE); 
  if(OTX_E_NOT_SUPPORTED != OTX_RESULT_CODE(result)) {
    DemoCheckRc(result);
  }

  if(g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
    OTX_BOOL bMonitorMode;
    // Thor-4-ExpressCard by default is set to receive on the four lines 
    // (monitor mode) with all transmitters disabled (tristate). This is
    // just a check to make sure it is in the correct mode
    result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hBoardController, 
                                         OTX_ATTR_BRD_MONITOR_ONLY, 
                                         &bMonitorMode); 
    DemoCheckRc(result);
    DemoLogf("Configure Thor-4-ExpressCard is in %s mode\n", 
             bMonitorMode ? "Monitor only" : "NOT IN MONITOR ONLY");
    if(bMonitorMode == OTX_FALSE) {
      // Transmit pins could be enabled, need to tell user
      DemoLogf("\nError Thor-4-ExpressCard is not in monitor only mode?\n\n");
    }

    // we'll source the Falc from the Burst device, and source the burst device from the FALC
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(result);
    //  Burst Recv (S2P) <-- Falc
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_FALC);  DemoCheckRc(result);
  }
  else if (g_demoEnv.dspFileName[0]) {
    
    // Open physical DSP devices
    switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    case OTX_DEVICE_GIMLE_16_PCIE:
      {
        OTX_UINT32   eDevType;
        OTX_ITERATOR itrChildTypes;
        OtxDrvIteratorBegin(&itrChildTypes);
        while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hBoardController, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
          OTX_BOOL bOpenAsm = OTX_FALSE;
          switch (eDevType) {
          case OTX_DEVICE_VIDAR_5x_ASM:
          case OTX_DEVICE_VIDAR_5X4_ASM_V90:
            DemoLogf("Opening Vidar-5x4-ASM board\n");
            DemoLogf("Vidar Board Type (%d) not supported by this demo program\n", eDevType);
            exit(1);
            break;
          case OTX_DEVICE_VIDAR_5X4_ASM_CST:
            DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
            g_nDspCommand =  0x010A;
            bOpenAsm = OTX_TRUE;
            break;
          case OTX_DEVICE_VIDAR_5X4_ASM_EX:
            DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
            g_nDspCommand =  0x010A;
            bOpenAsm = OTX_TRUE;
            break;
          case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
            DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
            g_nDspCommand =  0x010A;
            bOpenAsm = OTX_TRUE;
            break;
          case OTX_DEVICE_VIDAR_55x4_ASM:
            DemoLogf("Opening Vidar-55x4-ASM board\n");
            g_nDspCommand = 0x0106;
            bOpenAsm = OTX_TRUE;
            break;
          }
          if (bOpenAsm) {
            // Get a handle to to the ASM Daughter Board
            result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(result);
          }
        }
        
        for (i = 0; i < g_demoEnv.coDsp; i++) {
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(result);
        }
      }
      break;
    default:
      DemoLogf("Board Type (%d) does not support ASM daughter boards\n", g_demoEnv.nBoardType);
      exit(1);
      break;
    }

    // Load DSP Program
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      DemoLogf("Loading DSP#%d program %s\n", i, g_demoEnv.dspFileName);
      result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], g_demoEnv.dspFileName, "_c_int00"); DemoCheckRc(result);
    }

  } // g_demoEnv.dspFileName[0]
}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
static void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  
  // Set default configuration data
  DemoInitT1E1ConfigData(&(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));
  
  for (i = 0; i < g_demoEnv.coLi; i++) {
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1){ // T1
      // Configure T1 Transceivers
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
      // Set them in T1 mode
      result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(result);
    }
    else{  // E1
      // Configure E1 Transceivers
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
      // Set them in E1 mode
      result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(result);
    }
    
    // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
    result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(result);

    // Set line termination mode
    {

      DemoLogf("Li[%d]: Setting Line Termination mode to ", i);
      switch (g_demoEnv.eTermMode) {
      case OTX_T1E1_TERM_120_OHM:
        DemoLogf("OTX_T1E1_TERM_120_OHM\n");
        break;
      case OTX_T1E1_TERM_100_OHM:
        DemoLogf("OTX_T1E1_TERM_100_OHM\n");
        break;
      case OTX_T1E1_TERM_75_OHM:
        DemoLogf("OTX_T1E1_TERM_75_OHM\n");
        break;
      case OTX_T1E1_TERM_HIGHZ:
        DemoLogf("OTX_T1E1_TERM_HIGHZ\n");
        break;
      default:
        DemoLogf("Invalid Termination mode.\n");
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], g_demoEnv.eTermMode); DemoCheckRc(result);

    }
    

    // Set receive threshold
    DemoLogf("Li[%d]: Setting RX Input Threshold to %d\n", i, g_demoEnv.nRxThreshold);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nRxThreshold); DemoCheckRc(result);
    { // Read back the threshold
      OTX_UINT32 nRxThreshold;
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(result);
      DemoLogf("Li[%d]: RX Input Threshold = %d\n", i, nRxThreshold);
    }
    
    // Set Monitor (-20dB) mode
    DemoLogf("Li[%d]: Setting Monitor mode to %s\n", i, g_demoEnv.bMonitorMode ? "ON" : "OFF");
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[i], g_demoEnv.bMonitorMode); DemoCheckRc(result);
    
    // Turning front end Amplifier ON/OFF      
    DemoLogf("Li[%d]: Setting Front End Amplifier to %s\n", i, g_demoEnv.bAmplifier ? "ON" : "OFF");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bAmplifier); DemoCheckRc(result);
  }
  
}


void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hBoardController\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  for (i = 0; i < g_demoEnv.coQuadE1T1; i++) {
    // Enable QuadFalcs
    DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Enable Physical Devices, Li's
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  // by enabling the Burst device, we start the process of transferring data
  // from the hardware to the driver via DMA.  The data is stored in the Burst
  // device's buffer for later processing to connected logical devices.
  DemoLogf("Enable g_hPhysDevs.hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[128];

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
    OTX_HANDLE hPipe;

    // Although the timeslots connections are hardcoded on this card, we still need to connect
    // to the timeslots of the Li devices that we want to control. If we don't connect any pipes 
    // to the Li devices then it will send an Idle code on each timeslot.

    memset(sqTimeSlotMask, 0xFF, sizeof(sqTimeSlotMask));

    result = OtxDrvCreatePipe(64*32, &hPipe);
    DemoCheckRc(result);
    result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[0],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(result);

    result = OtxDrvCreatePipe(64*32, &hPipe);
    DemoCheckRc(result);
    result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[1],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(result);

    return; // No need to connect the other end of the pipe since the pipe 
            // connections are hard-coded on this board.
  }
  
  if (g_demoEnv.dspFileName[0]) {
    // DSP configuration
    OTX_HANDLE hPipes[4*128];
    OTX_HANDLE hBurstPipes[4*128];
    OTX_INT32 k;
    OTX_INT32 i;
    OTX_INT32 j;


    // Connect each QuadFalc to each DSP. Map the timeslot in the following fashion on 
    // the four DSP highways:
    //    [ 0..31]  - LI0:TS0-TS31, LI4:TS0-TS31, LI8:TS0-TS31,  LI12:TS0-TS31
    //    [32..63]  - LI1:TS0-TS31, LI5:TS0-TS31, LI9:TS0-TS31,  LI13:TS0-TS31
    //    [64..95]  - LI2:TS0-TS31, LI6:TS0-TS31, LI10:TS0-TS31, LI14:TS0-TS31
    //    [96..127] - LI3:TS0-TS31, LI7:TS0-TS31, LI11:TS0-TS31, LI15:TS0-TS31
    for (k = 0; k < 4; k++) {
      for (j = 0; j < 4; j++) {
        for (i = 0; i < 32; i++) {
          
          OTX_INT32 nPipe = k*128 + j*32 + i;
          
          result = OtxDrvCreatePipe(64, &hPipes[nPipe]); DemoCheckRc(result);
          
          memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
          sqTimeSlotMask[i*4+j] = 0xFF;
          result = OtxDrvConnectPipe(hPipes[nPipe], OTX_PIPE_INPUT,  g_hPhysDevs.hTss,  k, sqTimeSlotMask, 128); DemoCheckRc(result);
          
          memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
          sqTimeSlotMask[i+j*32] = 0xFF;
          result = OtxDrvConnectPipe(hPipes[nPipe], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  k+4, sqTimeSlotMask, 128); DemoCheckRc(result);
          
        }
      }
    }

    
    // Connect from the DSP highways to the burst highways
    for (k = 0; k < 4; k++) {
      if(g_nLoopBackMode == OTX_FALSE) {

        for (i = 0; i < 128; i++) {
        
          OTX_INT32 nPipe = k*128 + i;
        
          result = OtxDrvCreatePipe(64, &hBurstPipes[nPipe]); DemoCheckRc(result);
        
          memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
          sqTimeSlotMask[i] = 0xFF;
        
          // Connect from DSP output highway...
          result = OtxDrvConnectPipe(hBurstPipes[nPipe], OTX_PIPE_INPUT,  g_hPhysDevs.hTss,  k+4, sqTimeSlotMask, 128); DemoCheckRc(result);
        
          // ... to the Burst highway
          result = OtxDrvConnectPipe(hBurstPipes[nPipe], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  k,   sqTimeSlotMask, 128); DemoCheckRc(result);
        }
      }
      else {
      
        Demo2IoInLoopAllTsS sIn;
        OTX_UINT16 nDemo2Rc;
        OTX_UINT32 nBytesReturned;
        OTX_INT32 nStartingLi;
        OTX_UINT8 sqBurstTimeSlotMask[128];

        // Create a pipe to connect the DSP to Burst
        // LI0 - 3 to DSP 0  
        // LI4 - 7 to DSP 1
        // LI8 -11 to DSP 2 
        // LI12-15 to DSP 3 
        nStartingLi = k * 4; // used to map 1/4 of the lines to each DSP
        DemoLogf("Configure DSP%d for Li %02d-%02d\n", k, nStartingLi, nStartingLi+3);

        switch (g_demoEnv.nBoardType) {
        case OTX_DEVICE_GIMLE_16_PCI_PLUS:
        case OTX_DEVICE_GIMLE_16_PCIE:
          DemoLogf("Pipe configuration for a Gimle-16-PCI-Plus board DSP %d\n", k);
          break;

        case OTX_DEVICE_GIMLE_16_PCIE_BASIC_32:
        case OTX_DEVICE_GIMLE_16_PCIE_BASIC_64:
        case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
        case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
        // Not supported at this time
        default:
          DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
          OsaConsoleExit();
          exit(1);
          break;
        }

        // --------------------------------------------------------------------
        // Create an incoming Pipe representing the DSP to Burst
        // Create a pipe to connect the DSP to Burst
        // LI0 - 3 of DSP 0 to Burst HWY 0
        // LI4 - 7 of DSP 1 to Burst HWY 1
        // LI8 -11 of DSP 2 to Burst HWY 2
        // LI12-15 of DSP 3 to Burst HWY 3
        // --------------------------------------------------------------------
        // All TS are used, note TS 0 of Li is not actually available but it is passsed on anyway
        memset(sqBurstTimeSlotMask, 0xFF, sizeof(sqBurstTimeSlotMask));

        result = OtxDrvCreatePipe(64*128, &g_hPipes.hDspOut[k]);
        DemoCheckRc(result);
  
        DemoLogf("  Configure DSP%d to Burst HWY%d  for loopback test\n", k, k);

        // -------------------------
        // Connect pipe: DSP -> Burst
        // -------------------------
        result = OtxDrvConnectPipe(g_hPipes.hDspOut[k], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[k],  -1, sqBurstTimeSlotMask, 128);
        DemoCheckRc(result);

        // We must define the highway to use [0-3] for burst device
        result = OtxDrvConnectPipe(g_hPipes.hDspOut[k], OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, k, sqBurstTimeSlotMask, 128);
        DemoCheckRc(result);

        // Instruct the DSP program to loop all timeslots
        sIn.bLoopAllTs = 1;

        DemoLogf("  Configure DSP%d for TS loopback test\n", k);
        result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[k], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, 0, &g_hPhysDevs.hUserApp[k]); 
        DemoCheckRc(result);

        // Loop the TS from receive to transmit to send data 
        // to the host on the busrt highway.
        result = OtxDspIoControl(
          g_hPhysDevs.hUserApp[k], 
          g_nDspCommand,    // DEMO2_IOCTL_LOOP_ALL_TS which is defined by 55x or 54x 
          &sIn, 
          sizeof(sIn), 
          &nDemo2Rc, 
          sizeof(nDemo2Rc), 
          &nBytesReturned, 
          OTX_TASK_SYNC, 
          5000 //OTX_TIME_INFINITY
          );
        DemoCheckRc(result);
      }
    }
  } else {
    // Non-DSP configuration
    OTX_INT32 nLi;
    OTX_UINT32 iHighwayNo;
    OTX_UINT8 sqBurstTimeSlotMask[128];
    

    switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    case OTX_DEVICE_GIMLE_16_PCIE:
      DemoLogf("Pipe configuration for a Gimle-16-PCI-Plus board\n");

      // Create the Pipes from the LIs to the Burst Device.
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        OTX_INT32 nTs;
        
        // We must define the highway to use [0-3]
        iHighwayNo = nLi / 4;
        
        // Create pipes. Use all timeslots except TS0.
        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
        
        DemoLogf("Creating pipe for Burst Receiver\n");
        
        for (nTs = 0; nTs < 32; nTs++) {
          
          sqTimeSlotMask[nTs] = 0xFF;
          
          // for this demo, we'll lay out the burst time slot mask as follows:
          //    [ 0..31]  - LI0:TS0-TS31, LI4:TS0-TS31, LI8:TS0-TS31,  LI12:TS0-TS31
          //    [32..63]  - LI1:TS0-TS31, LI5:TS0-TS31, LI9:TS0-TS31,  LI13:TS0-TS31
          //    [64..95]  - LI2:TS0-TS31, LI6:TS0-TS31, LI10:TS0-TS31, LI14:TS0-TS31
          //    [96..127] - LI3:TS0-TS31, LI7:TS0-TS31, LI11:TS0-TS31, LI15:TS0-TS31
          //
          // upon connecting the pipe (see below), we have to pass the highway
          // number (0-3 - one highway per each QuadFalc (128TS per highway)
          // as well as the size of the highway (128 TS)
          //
          // this is just a simple algorithm to accomplish the above
          sqBurstTimeSlotMask[nTs + ((nLi % 4) * 32)] = 0xFF;
          
        }
        
        
        // --------------------------------------------------------------------
        // Create an incoming Pipe representing the external T1/E1
        // --------------------------------------------------------------------
        result = OtxDrvCreatePipe(64*32, &g_hPipes.hLiOut[nLi]);
        DemoCheckRc(result);
        
        // -------------------------
        // Connect pipe: Li -> Burst
        // -------------------------
        result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
        DemoCheckRc(result);
        
        result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, 128);
        DemoCheckRc(result);
        
      }
      break;

    case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
    case OTX_DEVICE_GIMLE_16_PCIE_BASIC_32:
      DemoLogf("Pipe configuration for a Gimle-16-PCI-Basic-32 and Gimle-16-PCIe-Basic-32 board\n");

      // Create the Pipes from the LIs to the Burst Device. We can use 32 64kpbs pipes on this board.
      // In this example we pick up TS10 and TS11 from each of the 16 Lis. 
      //
      // There is only one 2Mbps (32 timeslot) burst highway on this board.
      //
      // We connect the TS10 timeslots from each Li to TS0-TS15 of the burst highway.
      // And we connect the TS11 timeslots from each Li to TS16-TS31 of the burst highway.
      // 
      // However, in practise 
      // you can pick up any timeslot from any Li device just as long as the total number of timeslots are
      // less than or equal to 32 (for this particular board).
      //

      // Loop over the 16 Line interfaces and connect TS10 from each Li
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        OTX_HANDLE hPipe;
        OTX_INT32 nTs = 10;
        
        // There is only 1 burst highway on this board
        iHighwayNo = 0;
          
        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
        
        sqTimeSlotMask[nTs] = 0xFF;  // E1/T1 timeslot
        sqBurstTimeSlotMask[nLi] = 0xFF;  // Timeslot on the burst highway
        
        // Create a pipe for each T1/E1 timeslot
        result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
        
        // Connect pipe: Li -> Burst
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
        DemoCheckRc(result);
        
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, 32); // Note, the burst highway is 2Mbps, so therefore we supply a mask size of 32
        DemoCheckRc(result);
        
      }
      // Loop over the 16 Line interfaces and connect TS11 from each Li
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        OTX_HANDLE hPipe;
        OTX_INT32 nTs = 11;
        
        // There is only 1 burst highway on this board
        iHighwayNo = 0;
          
        memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
        memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
        
        sqTimeSlotMask[nTs] = 0xFF;  // E1/T1 timeslot
        sqBurstTimeSlotMask[nLi + 16] = 0xFF;  // Timeslot on the burst highway
        
        // Create a pipe for each T1/E1 timeslot
        result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
        
        // Connect pipe: Li -> Burst
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
        DemoCheckRc(result);
        
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, 32); // Note, the burst highway is 2Mbps, so therefore we supply a mask size of 32
        DemoCheckRc(result);
        
      }
      break;


    case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
    case OTX_DEVICE_GIMLE_16_PCIE_BASIC_64:
      DemoLogf("Pipe configuration for a Gimle-16-PCI-Basic-64 and Gimle-16-PCIe-Basic-64 board\n");

      // Create the Pipes from the LIs to the Burst Device. We can use 64 64kpbs pipes on this board.
      // In this example we pick up TS10, TS11, TS12 and TS13 from each of the 16 Lis. 
      //
      // There are two 2Mbps burst highways on this board. Each 2Mbps highways have 32 timeslots each. 
      //
      // We connect the TS10 timeslots from each Li to TS0-TS15 of the first burst highway.
      // And we connect the TS11 timeslots from each Li to TS16-TS31 of the first burst highway.
      //
      // We connect the TS12 timeslots from each Li to TS0-TS15 of the second burst highway.
      // And we connect the TS13 timeslots from each Li to TS16-TS31 of the second burst highway.
      //
      // However, in practise 
      // you can pick up any timeslot from any Li device just as long as the total number of timeslots are
      // less than or equal to 64 (for this particular board).
      //
 
      // Loop over the two burst highways
      for (iHighwayNo = 0; iHighwayNo < 2; iHighwayNo++) {

        // Loop over the 16 Line interfaces and connect TS10 from each Li to 16 timeslots on burst highway 0
        for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
          OTX_HANDLE hPipe;
          OTX_INT32 nTs = 1 + 2 * iHighwayNo;
          
          memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
          memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
          
          sqTimeSlotMask[nTs] = 0xFF;  // E1/T1 timeslot
          sqBurstTimeSlotMask[nLi] = 0xFF;  // Timeslot on the burst highway
          
          // Create a pipe for each T1/E1 timeslot
          result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
          
          // Connect pipe: Li -> Burst
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
          DemoCheckRc(result);
          
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, 32); // Note, the burst highway is 2Mbps, so therefore we supply a mask size of 32
          DemoCheckRc(result);
          
        }
        // Loop over the 16 Line interfaces and connect TS11 from each Li to 16 timeslots on burst highway 0
        for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
          OTX_HANDLE hPipe;
          OTX_INT32 nTs = 2 + 2 * iHighwayNo;
          
          memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
          memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
          
          sqTimeSlotMask[nTs] = 0xFF;  // E1/T1 timeslot
          sqBurstTimeSlotMask[nLi + 16] = 0xFF;  // Timeslot on the burst highway
          
          // Create a pipe for each T1/E1 timeslot
          result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
          
          // Connect pipe: Li -> Burst
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
          DemoCheckRc(result);
          
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, 32); // Note, the burst highway is 2Mbps, so therefore we supply a mask size of 32
          DemoCheckRc(result);
          
        }
      }
      break;

    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
    }


  }
  
}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(const char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF((char *)formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {
    
    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
  
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}

void DemoCheckEventCount()
{
  OTX_INT32   nHighway;
  OTX_INT32   nHighwaySpeed;
  OTX_INT32   nBurstHighways;
  OTX_BOOL    bHwy;
  OTX_UINT64  nRange;

  nHighwaySpeed = g_demoEnv.nHighwaySpeed;
  nBurstHighways = g_demoEnv.coBurstHighways;
  bHwy = OTX_FALSE;
  for (nHighway = 0; nHighway < nBurstHighways; nHighway++) {
    if(g_hLogicDevs.mBurstReceiver[nHighway].nEventCnt >= 2) {
      bHwy = OTX_TRUE;
    }
    if (!g_hLogicDevs.mBurstReceiver[nHighway].bActive) {
      // All burst highways need to be active, so just return
      return;
    }
  }

  if(bHwy == OTX_TRUE) {
    // One of the highways is receiving more events then the others
    DemoLogf("Burst receiver event count not equal\n");
    // Display the event count for the highways
    for(nHighway = 0; nHighway < nBurstHighways; nHighway++) {
      DemoLogf("    Highway %d event count %d\n", 
               nHighway, 
               g_hLogicDevs.mBurstReceiver[nHighway].nEventCnt);
    }
  }

  // Zero counters when all highways have an event or
  // an error is found above  
  for (nHighway = 0; nHighway < nBurstHighways; nHighway++) {
    if(g_hLogicDevs.mBurstReceiver[nHighway].nEventCnt >= 1) {
      bHwy = OTX_TRUE;
    }
  }
  if(bHwy == OTX_TRUE) {
    nRange = (DEMO_BURST_BUF_SIZE/nHighwaySpeed);
    for (nHighway = 0; nHighway < nBurstHighways; nHighway++) {
      // Reset the counters to zero
      g_hLogicDevs.mBurstReceiver[nHighway].nEventCnt = 0;
    }
  }
}

/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_INT32 nHighway;
  OTX_INT32 nBurstHighways;

  nBurstHighways = g_demoEnv.coBurstHighways;
  for (nHighway = 0; nHighway < nBurstHighways; nHighway++) {
    g_hLogicDevs.mBurstReceiver[nHighway].nEventCnt = 0;
    g_hLogicDevs.mBurstReceiver[nHighway].nFrameRefId = 0;
  }

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        switch (eventData.m_eDeviceType) {
          case OTX_DEVICE_BURST:
            switch (eventData.m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              DemoBurstReceiver(eventData.m_nRequestId);
              g_hLogicDevs.mBurstReceiver[eventData.m_nRequestId].nEventCnt++;

              // Check to make sure all highways are generating events
              DemoCheckEventCount();

              continue;
            default:
              break;
            }
            break;
          default:
            break;
        }

        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.
*******************************************************************************/
void *DemoStatsThread(
  void *lpParameter   // thread data
)
{
  OTX_RESULT        result;
  OTX_DATETIME      nTimeStamp;

  OTX_UINT32        stat_Starts;          // Number of actual dma transfers initiated
  OTX_UINT32        stat_Completes;       // Number of actual dma transfers completed
  OTX_UINT32        stat_Partial;         // Number of partial transfer due to framemask
  OTX_UINT32        stat_Full;            // Number of count of max dma size transfers
  OTX_UINT32        stat_Aborts;          // Number of actual dma transfers aborts
  OTX_UINT32        stat_Amount;          // Number of amount of data transfered
  OTX_UINT32        stat_Errors;          // Number of count of processing errors
  OTX_UINT32        stat_OUruns;          // Number of amount of over/under runs
  OTX_UINT32        stat_OUrunsUser;      // Number of amount of over/under runs caused by user mode
  OTX_UINT32        stat_DmaBusy;         // Number of count of DMA busy occurances
  OTX_UINT32        stat_Notify;          // Number of logical device notifies
  OTX_UINT32        stat_NotifyPartial;   // Number of logical device notifies - partial

  while (g_bStatsThreadActive) {
    OtxDrvGetCurrentTime(&nTimeStamp);
    
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS,         &stat_Starts        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES,      &stat_Completes     ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL,        &stat_Partial       ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_FULL,           &stat_Full          ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS,         &stat_Aborts        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT,         &stat_Amount        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS,         &stat_Errors        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS,         &stat_OUruns        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER,    &stat_OUrunsUser    ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY,        &stat_DmaBusy       ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY,         &stat_Notify        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL, &stat_NotifyPartial ); DemoCheckRc(result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Recv:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_OUrunsUser,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial
    );

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS,         &stat_Starts        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES,      &stat_Completes     ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL,        &stat_Partial       ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_FULL,           &stat_Full          ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS,         &stat_Aborts        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT,         &stat_Amount        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS,         &stat_Errors        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS,         &stat_OUruns        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER,    &stat_OUrunsUser    ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY,        &stat_DmaBusy       ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY,         &stat_Notify        ); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL, &stat_NotifyPartial ); DemoCheckRc(result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Send:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_OUrunsUser,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial
    );

    OsaThreadSleep(1000);
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;
}


/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{

  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    nBufferSize;
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_UINT64    nRange;
  OTX_DATETIME  dtFrameRefTime;
  OTX_UINT32 nTestValue;
  OTX_UINT64 nMaskValue = 0xFFFFFFFF;

  // validate parameters
  if ( (nSourceId < 0) || (nSourceId > 3) ) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  if(ui32 != DEMO_BURST_BUF_SIZE) {
    if(ui32 == 0) {
      // Nothing for us to process buffer empty
      return;
    }
    // Process what we received
/*
    DemoLogf("DemoBurstReceiver[%d] Receive a short buffer size %d should be %d\n", 
             nSourceId, 
             ui32, 
             DEMO_BURST_BUF_SIZE);
*/
  }

  // We may not always get a full buffer
  nBufferSize = ui32;
  // Every frame in the buffer will need data, 
  // check for a partial frame.
  if(nBufferSize % g_demoEnv.nHighwaySpeed) {
    DemoLogf("DemoBurstReceiver[%d] Buffer contains a partial frame\n", 
             nSourceId);
    DemoLogf("DemoBurstReceiver[%d] Buf size %d, Frame size %d\n", 
             nSourceId, 
             nBufferSize,
             g_demoEnv.nHighwaySpeed);

    // The size needs to be a multiple of the timeslots
    nBufferSize = (ui32/g_demoEnv.nHighwaySpeed) * g_demoEnv.nHighwaySpeed;
  }

  // There was a data receive event, add one to the counter
  g_hLogicDevs.mBurstReceiver[nSourceId].nRecvCnt++;
  if(g_hLogicDevs.mBurstReceiver[nSourceId].nRecvCnt == 0) {
    // If count is zero set it to 1. Zero is used to indicate no data received
    g_hLogicDevs.mBurstReceiver[nSourceId].nRecvCnt++;
  }

  // Check the frame refence ID. It should be different for each frame.
  if(g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId >= nFrameRefId) {

    if(ui32 != DEMO_BURST_BUF_SIZE) {
      DemoLogf("DemoBurstReceiver[%d] Receive a short buffer size %d should be %d\n", 
               nSourceId, 
               ui32, 
               DEMO_BURST_BUF_SIZE);
    }

    // frame reference should be incrementing
    DemoLogf("DemoBurstReceiver[%d] nFrameRefId did not increment previous value\n", 
             nSourceId);

    DemoLogf("  Previous frame ID %llu\n", 
             g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId);
    nTestValue = (OTX_UINT32) (g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId & nMaskValue);
    DemoLogf("Lower bits of nFrameRefId 0x%X, mask %llu\n", 
             nTestValue,
             nMaskValue);
    nTestValue =  (OTX_UINT32) (g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId >> 32);
    DemoLogf("Upper bits of nFrameRefId 0x%X\n", 
             nTestValue);


    DemoLogf("  This frame ID %llu\n", 
             nFrameRefId);
    nTestValue = (OTX_UINT32) (nFrameRefId & nMaskValue);
    DemoLogf("Lower bits of nFrameRefId 0x%X, mask %llu\n", 
             nTestValue,
             nMaskValue);
    nTestValue =  (OTX_UINT32) (nFrameRefId >> 32);
    DemoLogf("Upper bits of nFrameRefId 0x%X\n", 
             nTestValue);


    if(nFrameRefId == 0) {
      // Counter is set to zero rolled over
      DemoLogf("DemoBurstReceiver[%d] ID is zero\n", 
               nSourceId);
    }    
    if(nFrameRefId == g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId) {
      // The ID is the same
      DemoLogf("DemoBurstReceiver[%d] Frame Reference ID is equal\n", 
               nSourceId);
    }
  }

  nRange = (ui32/g_demoEnv.nHighwaySpeed);
  if(g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId > (nFrameRefId - nRange)) {
    // The buffer has multiple frames
    DemoLogf("DemoBurstReceiver[%d] nFrameRefId range is not correct for buffer size %d\n", 
             nSourceId,
             ui32);
    DemoLogf("  Previous frame ID %llu, this frame ID %llu\n", 
             g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId,
             nFrameRefId);

    DemoLogf("  Previous frame ID %llu\n", 
             g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId);
    nTestValue = (OTX_UINT32) (g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId & nMaskValue);
    DemoLogf("Lower bits of nFrameRefId 0x%X, mask %llu\n", 
             nTestValue,
             nMaskValue);
    nTestValue =  (OTX_UINT32) (g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId >> 32);
    DemoLogf("Upper bits of nFrameRefId 0x%X\n", 
             nTestValue);

    DemoLogf("  This frame ID %llu\n", 
             nFrameRefId);
    nTestValue = (OTX_UINT32) (nFrameRefId & nMaskValue);
    DemoLogf("Lower bits of nFrameRefId 0x%X, mask %llu\n", 
             nTestValue,
             nMaskValue);
    nTestValue =  (OTX_UINT32) (nFrameRefId >> 32);
    DemoLogf("Upper bits of nFrameRefId 0x%X\n", 
             nTestValue);

  }

  g_hLogicDevs.mBurstReceiver[nSourceId].nFrameRefId = nFrameRefId;


  // echo if we're in verbose mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bVerbose) {
      int x;
      OTX_DATETIME nCurrentTime;

      OtxDrvGetCurrentTime(&nCurrentTime);
      DemoPrintTimeStamp(nCurrentTime);
      DemoLogf("BurstReceiver[%d][%d] (len=%.4d) ", nSourceId, g_nCurrentLi, sizeof(buf));
      // Display a sample of the highway's currentLi's data
      switch (g_demoEnv.nBoardType) {
      case OTX_DEVICE_GIMLE_16_PCI_PLUS:
      case OTX_DEVICE_GIMLE_16_PCIE:
        // Print the first byte of the 32 timeslots in the buffer for the current Li
        for (x = 0; (x < 32) && (x < (OTX_INT32) sizeof(buf)); x++) {
          DemoLogf("%.2X ", buf[x + ((g_nCurrentLi % 4) * 32)]);
        }
        break;
      case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
      case OTX_DEVICE_GIMLE_16_PCIE_BASIC_32:
        // We connected up TS10 and TS11 on this board. Print the first byte of these 2 timeslots in the buffer for the current Li
        DemoLogf("TS10=%.2X ", buf[g_nCurrentLi]);  
        DemoLogf("TS11=%.2X ", buf[g_nCurrentLi + 16]); 
        break;
      case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
      case OTX_DEVICE_GIMLE_16_PCIE_BASIC_64:
        // We connected up TS10, TS11, TS12, and TS13 on this board. 
        // TS10 and TS11 was connected to highway (nSourceId) 0.
        // TS12 and TS13 was connected to highway (nSourceId) 1.
        // Print the first byte of these 2 timeslots in the buffer for the current Li
        if (nSourceId == 0) {
          DemoLogf("TS10=%.2X ", buf[g_nCurrentLi]);  
          DemoLogf("TS11=%.2X ", buf[g_nCurrentLi + 16]); 
        } else {
          DemoLogf("TS12=%.2X ", buf[g_nCurrentLi]);  
          DemoLogf("TS13=%.2X ", buf[g_nCurrentLi + 16]); 
        }
        break;
      case OTX_DEVICE_THOR_4_EXPRESSCARD:
        // Print the first byte of the 32 timeslots in the buffer for the current Li
        for (x = 0; (x < 32) && (x < (OTX_INT32) sizeof(buf)); x++) {
          DemoLogf("%.2X ", buf[(g_nCurrentLi) + x*4]);
        }
        break;
      default:
	      break;
      }
      DemoLogf("\n");

  }

  // Check the received value if we are in Gamma Test Mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
    int x;
    int y;

    switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    case OTX_DEVICE_GIMLE_16_PCIE:
    case OTX_DEVICE_THOR_4_EXPRESSCARD:
      // GAMMA test
      for (x = 0; x < (int) nBufferSize; x += 128) {
        for (y = 0; y < 128; y++) {
          if (buf[x+y] == g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y]) {
            // Data matched. No action needed.
          } else {
            // Data mismatch. Step error count and resync
            g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y] = buf[x+y];
            g_hLogicDevs.mBurstReceiver[nSourceId].nGammaTestErrors[y]++;
          }

          // Step the GammaTestData value, which is the next value that we expect
          g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y]++;
        }
      }
      break;

    case OTX_DEVICE_GIMLE_16_PCIE_BASIC_32:
    case OTX_DEVICE_GIMLE_16_PCIE_BASIC_64:
    case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
    case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
      // Gamma test is not supported on these boards.
      break;

    default:
	  break;
    }
  }

  // write data to a file if desired
  if (g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile) {
    int cbWritten;

    if (g_hLogicDevs.mBurstReceiver[nSourceId].nBurstTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
      }
    } else {
      // Single TS from the burst highway
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;
      
      byteCount = 0;
            
      for (x = g_hLogicDevs.mBurstReceiver[nSourceId].nBurstTsToRecord; x < (OTX_INT32) sizeof(buf); x += g_demoEnv.nHighwaySpeed) {
        singleTsBuf[byteCount++] = buf[x];
      }
      
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    } // Single TS from the burst highway
  } // if (g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile)
}


void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}

OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  OTX_RESULT result;
  static OTX_UINT32 g_nBurstHighwayNo = 0;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-%d Set Burst Receiver number (currently %d selected)\n", g_demoEnv.coBurstHighways, g_nBurstHighwayNo);
    DemoLogf(" n/N Increment/decrement active Li (currently Li#%d selected)\n", g_nCurrentLi);
    DemoLogf(" a   Abort all Burst Receivers\n");
    DemoLogf(" r   Start Burst Receiver[%d] (no file logging)\n", g_nBurstHighwayNo);
    DemoLogf(" R   Start Burst Receiver[%d] Prompt for file to recv\n", g_nBurstHighwayNo);
    DemoLogf(" X   Start Burst Receiver All Highways Prompt for file to recv\n", g_nBurstHighwayNo);
    DemoLogf(" v   Toggle verbose (print received data) setting Burst Receiver[%d]\n", g_nBurstHighwayNo);
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" t   Toggle verbose buffer usage statistics\n");
    DemoLogf(" f   Forcing resynchronization\n");
    DemoLogf(" T   Reset buffer usage statistics\n");
    DemoLogf(" D   Dump buffer usage statistics\n");
    DemoLogf(" #   Set burst notification value (1-32)\n");
    DemoLogf(" F   Query line frequency of the current clock source\n");
    DemoLogf(" [   Read T1/E1 span voltage for current Li\n");
    DemoLogf(" g   Enable/Disable Gamma Test\n");
    DemoLogf(" G   Clear Gamma Test error counters\n");
    DemoLogf(" s   Print Gamma Test Results (condensed form)\n");
    DemoLogf(" S   Print Gamma Test Results (verbose form)\n");
    DemoLogf(" z   Set T1E1 Line Termination\n");
    DemoLogf(" Z   Test T1E1 line Terminations\n");

    break;

  case '0':
  case '1':
  case '2':
  case '3':
    // Select current burst highway number
    {
      OTX_INT32 nBurstHighwayNo = commandChar - '0';
      
      if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_BASIC_32 || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE_BASIC_32 ) {

        if (nBurstHighwayNo > 0) {
          DemoLogf("Invalid Burst Highway number (%d) for Gimle-16 Basic-32 which has one highway. \n", nBurstHighwayNo);

          // first highway is selected after incorrect pick 
          nBurstHighwayNo = 0;
        }
      
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_BASIC_64 || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE_BASIC_64) {
        
        if (nBurstHighwayNo > 1) {
          DemoLogf("Invalid Burst Highway number (%d) for Gimle-16 Basic-64 which has two highways. \n", nBurstHighwayNo);

          nBurstHighwayNo = 0;
        }

      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
        
        if (nBurstHighwayNo > 0) {
          DemoLogf("Invalid Burst Highway number (%d) for Thor-4-ExpressCard which has one highway. \n", nBurstHighwayNo);
          
          nBurstHighwayNo = 0;
        }
      }
      
      g_nBurstHighwayNo = nBurstHighwayNo;
      DemoLogf("Burst Receiver #%d is selected ....\n", g_nBurstHighwayNo);
    }
    break;

  case '9':
    {
      OTX_BOOL bMonitor;
      OTX_HANDLE nBoardHandle = g_hPhysDevs.hBoardController;

      result = OtxDrvGetAttributeValueBool(nBoardHandle, 
                                           OTX_ATTR_BRD_MONITOR_ONLY, 
                                           &bMonitor); 
      DemoCheckRc(result);

      if(OTX_SUCCESS(result)) {
        bMonitor = !bMonitor;

        result = OtxDrvSetAttributeValueBool(nBoardHandle, 
                                             OTX_ATTR_BRD_MONITOR_ONLY, 
                                             bMonitor); 
        DemoCheckRc(result);
      }

      DemoLogf("OTX_ATTR_BRD_MONITOR_ONLY for data transmit %d\n", bMonitor);
    }
    break;

    // E1/T1 span voltmeter
  case '[':
    {
      OTX_UINT32 nVoltage;
      double nVolts;
      
      DemoLogf("Reading T1/E1 Span Voltage for Li%d: \n", g_nCurrentLi);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[g_nCurrentLi], OTX_ATTR_T1E1_LINE_VOLTAGE, &nVoltage); DemoCheckRc(result);
      
      //DemoLogf("  Voltage = %d \n", nVoltage);
      nVolts = (double) nVoltage/1000;
      DemoLogf("  Voltage = %2.3f V   \n", (double)nVolts);
      
    }
    break;

  case 'F':
    {
      OTX_UINT32 nFrequencyCount;

      DemoLogf("Querying frequency of current clock source ...\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_FREQUENCY_COUNTER, &nFrequencyCount); DemoCheckRc(result);
      if (OTX_SUCCESS(result)) {
        OTX_UINT32 nFrequency;

        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          nFrequency = (nFrequencyCount+1)*193/2048;
        } else {
          // E1 mode
          nFrequency = (nFrequencyCount+1)/8;
        }
        DemoLogf("Board frequency is %d Hz (nFrequencyCount=%d)\n", 
                 nFrequency, 
                 nFrequencyCount);
      }

      // Also get the current Li frequency
      if(g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
        
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[g_nCurrentLi], 
                                               OTX_ATTR_T1E1_LINE_FREQUENCY, 
                                               &nFrequencyCount); 
        DemoCheckRc(result);
        if (OTX_SUCCESS(result)) {
          OTX_UINT32 nFrequency;

          if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
            // T1 mode
            nFrequency = (nFrequencyCount+1)*193/2048;
          } else {
            // E1 mode
            nFrequency = (nFrequencyCount+1)/8;
          }
          DemoLogf("Line[%d] frequency is %d Hz (nFrequencyCount=%d)\n", 
                   g_nCurrentLi, 
                   nFrequency, 
                   nFrequencyCount);
        }
      }

      DemoLogf("Done.\n");
    }
    break;


  case 't':
    {
      g_bStatsThreadActive = !g_bStatsThreadActive;

      if (g_bStatsThreadActive) {
        DemoLogf("Starting Stats Thread\n");
        OsaThreadStart(DemoStatsThread);
      }

    }
    break;

  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d\n", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(result);
      }
      DemoLogf("Done.\n");
    }
    break;

  case 'a':
        
    g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bActive = OTX_FALSE;
    g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].nRecvCnt = 0;
    g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bGammaTestMode = OTX_FALSE;

    // close file
    if (g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].pRxFile) {
      DemoLogf("DemoBurstReceiver[%d]: Closing file\n", g_nBurstHighwayNo);
      fclose(g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].pRxFile);
      g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].pRxFile = NULL;
    }
    DemoLogf("DemoBurstReceiver[%d]: abort\n", g_nBurstHighwayNo);

    break;

  case 'v':
    g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bVerbose = !g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Receiver[%d] Thread\n", (g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bVerbose)?"ON":"OFF", g_nBurstHighwayNo);
    break;

  case 'g':
    {
      if ((g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) ||
          (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) ||
          (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD)) {
        int nActiveBurstReceivers = 0;
        int nHwy;

        for (nHwy = 0; nHwy < g_demoEnv.coBurstHighways; nHwy++) {
          if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_TRUE) {
            nActiveBurstReceivers++;
          }
        }

        if (nActiveBurstReceivers == g_demoEnv.coBurstHighways) {
          // Good. All receivers are active
          for (nHwy = 0; nHwy < g_demoEnv.coBurstHighways; nHwy++) {
            int n;
            memset(g_hLogicDevs.mBurstReceiver[nHwy].sqGammaTestData, 0, 128);
            for (n = 0; n < 128; n++) {
              g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[n] = 0;
            }
          }
          for (nHwy = 0; nHwy < g_demoEnv.coBurstHighways; nHwy++) {
            g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_TRUE;
          }
          DemoLogf("Gamma Test Mode is now enabled on all Burst Receivers\n");
        } else {
          DemoLogf("All Burst Receivers must be running before enabling Gamma Test Mode\n");
        }
      } else {
        DemoLogf("Gamma Test is only implemented for the Gimle-16-PCI-Plus and Gimle-16-PCIe-Plus boards\n");
      }
    }
    break;

  case 'G':  // Zero Gamma Test Error Counters
    {
      if ((g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) ||
          (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) ||
          (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD)) {
        int nHwy;
        int n;
        for (nHwy = 0; nHwy < g_demoEnv.coBurstHighways; nHwy++) {
          for (n = 0; n < 128; n++) {
            g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[n] = 0;
          }
          // Used to make sure data is received
          g_hLogicDevs.mBurstReceiver[nHwy].nRecvCnt = 0;
        }
        DemoLogf("Gamma Test Errors are now reset to zero\n");
      } else {
        DemoLogf("Gamma Test is only implemented for the Gimle-16-PCI-Plus and Gimle-16-PCIe-Plus boards\n");
      }
    }
    break;

  case 's':  // Gamma Test Results (condensed form)
    {
      if ((g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) ||
          (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE)     ||
          (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD)) {
        if (g_hLogicDevs.mBurstReceiver[0].bActive == OTX_TRUE) {
          int nHwy;
          int x;
          int y;
          OTX_UINT32 nStart;    // Number of dma transfers initiated
          OTX_UINT32 nComplete; // Number of dma transfers completed
          OTX_UINT32 nCheck1;    // Used to check data transfer active
          OTX_UINT32 nCheck2;    // Used to check data transfer active
          OTX_HANDLE hBurst = g_hPhysDevs.hBurst;

          // Need to check to make sure the burst device is receiving data
          OtxDrvGetAttributeValueUint32(hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &nStart);
          OtxDrvGetAttributeValueUint32(hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &nComplete);

          if (g_hLogicDevs.mBurstReceiver[0].bGammaTestMode) {
            DemoLogf("Burst Receiver Gamma Test Results (Condensed Form):\n");
            for (nHwy = 0; nHwy < g_demoEnv.coBurstHighways; nHwy++) {
              for (x = 0; x < 4; x++) {
                DemoLogf("Li[%.2d]", nHwy*4 + x);
                for (y = 0; y < 32; y++) {
                  if ((y % 4) == 0) {
                    DemoLogf(" "); // Group by 4 timeslots to make it easier to read
                  }
                  if ((y % 8) == 0) {
                    DemoLogf(" "); // Group by 8 timeslots to make it easier to read
                  }
                  if (g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[x*32 + y] > 9) {
                    DemoLogf("*"); // Indicates a number larger than 9
                  } else {
                    DemoLogf("%1d", g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[x*32 + y]);
                  }
                }
                DemoLogf("\n");
              }
            } // end of for nHwy
            DemoLogf("\n");

            // wait for a little bit, otherwise, OTX_ATTR_BURST_STAT_RECV_STARTS and 
            // OTX_ATTR_BURST_STAT_RECV_COMPLETES don't change 
            OsaThreadSleep(4);

            // Make sure that the Burst device is receiving data        
            OtxDrvGetAttributeValueUint32(hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &nCheck1);
            OtxDrvGetAttributeValueUint32(hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &nCheck2);
            if((nStart == nCheck1) || (nComplete == nCheck2)) {
              // Count is 0, the burst device is not receiving data
              DemoLogf("\n");
              DemoLogf("Burst HW is not receiving data check burst stats\n");
            }
            else {
              DemoLogf("\n");
              DemoLogf("Burst start is receiving data OK\n");
            }

            for (nHwy = 0; nHwy < g_demoEnv.coBurstHighways; nHwy++) {
              if(g_hLogicDevs.mBurstReceiver[nHwy].nRecvCnt == 0) {
                // Count is the same, the burst device is not receiving events
                DemoLogf("\n");
                DemoLogf("Burst[%d] event counter is 0, no data\n", nHwy);
                nHwy = g_demoEnv.coBurstHighways; // exit on error
              }
              else {
                DemoLogf("Burst[%d] event counter is OK\n", nHwy);
              }
            } 
          } else {
            DemoLogf("Gamma Test Mode is not enabled\n");
          }
        } else {
          DemoLogf("Burst Receivers Threads are not running. Start then with 'r', 'R', or 'X'\n");
        }
      } else {
        DemoLogf("Gamma Test is only implemented for the Gimle-16-PCI-Plus and Gimle-16-PCIe-Plus boards\n");
      }
    }
    break;

  case 'S':  // Gamma Test Results (verbose form)
    {
      if ((g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) ||
          (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) ||
          (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD)) {
        if (g_hLogicDevs.mBurstReceiver[0].bActive == OTX_TRUE) {
          int nHwy;
          int x;
          int y;
          if (g_hLogicDevs.mBurstReceiver[0].bGammaTestMode) {
            DemoLogf("Burst Receiver Gamma Test Results (Verbose Form):\n");
            for (nHwy = 0; nHwy < g_demoEnv.coBurstHighways; nHwy++) {
              for (x = 0; x < 4; x++) {
                for (y = 0; y < 32; y++) {
                  DemoLogf("Li[%.2d] TS%.2d:  ", nHwy*4 + x, y);
                  DemoLogf("%d", g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[x*32 + y]);
                  DemoLogf("\n");
                }
              }
            }
            DemoLogf("\n");
          } else {
            DemoLogf("Gamma Test Mode is not enabled\n");
          }
        } else {
          DemoLogf("Burst Receivers Threads are not running. Start then with 'r', 'R', or 'X'\n");
        }
      } else {
        DemoLogf("Gamma Test is only implemented for the Gimle-16-PCI-Plus and Gimle-16-PCIe-Plus boards\n");
      }
    }
    break;

  case 'r':
    {
      if (g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bActive == OTX_FALSE) {
        g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bActive = OTX_TRUE;
      } else if (g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].pRxFile) {
        DemoLogf("DemoBurstReceiver[%d]: Closing file\n", g_nBurstHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].pRxFile = NULL;
      } else {
        DemoLogf("Burst Receiver[%d] already running.\n", g_nBurstHighwayNo);
      }
    }
    break;


  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      static OTX_INT32 nBurstTsToRecord = -1;

      if (g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("BurstReceiver[%d]. Please enter file to record to: ", g_nBurstHighwayNo);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          break;
        }

        if ((g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].pRxFile = fopen(sqzBuf, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }

        DemoLogf("BurstReceiver[%d]. Please enter timeslot[0-%d] to record:\n", g_nBurstHighwayNo, g_demoEnv.nHighwaySpeed-1);
        DemoLogf(" -1  = All %d timeslots on the burst highway\n", g_demoEnv.nHighwaySpeed);

        switch (g_demoEnv.nBoardType) {
        case OTX_DEVICE_GIMLE_16_PCI_PLUS:
        case OTX_DEVICE_GIMLE_16_PCIE:
          // The burst highway (8Mpbs) on this board is connected to all of the timeslots of 4 E1/T1 spans.
          DemoLogf("  1  = TS1 of Li%d\n", g_nBurstHighwayNo * 4 + 0);
          DemoLogf("  33 = TS1 of Li%d\n", g_nBurstHighwayNo * 4 + 1);
          DemoLogf("  65 = TS1 of Li%d\n", g_nBurstHighwayNo * 4 + 2);
          DemoLogf("  97 = TS1 of Li%d\n", g_nBurstHighwayNo * 4 + 3);
          DemoLogf("  etc.\n");
          break;
        case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
        case OTX_DEVICE_GIMLE_16_PCIE_BASIC_32:
          // We connected TS10 of each Li to the TS0-TS15 of the one and only burst highway of this board.
          // We connected TS11 of each Li to the TS15-TS31 of the one and only burst highway of this board.
          DemoLogf("  0  = TS10 of Li0\n");
          DemoLogf("  1  = TS10 of Li1\n");
          DemoLogf("  ..\n");
          DemoLogf("  16 = TS11 of Li0\n");
          DemoLogf("  17 = TS11 of Li1\n");
          DemoLogf("  etc.\n");
          break;
        case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
        case OTX_DEVICE_GIMLE_16_PCIE_BASIC_64:
          // We connected up TS10, TS11, TS12, and TS13 on this board. 
          // TS10 and TS11 was connected to burst highway 0 (to TS0-TS15 and TS16-TS31 respectively).
          // TS12 and TS13 was connected to burst highway 1 (to TS0-TS15 and TS16-TS31 respectively).
          DemoLogf("  0  = TS%d of Li0\n", g_nBurstHighwayNo*2 + 10);
          DemoLogf("  1  = TS%d of Li1\n", g_nBurstHighwayNo*2 + 10);
          DemoLogf("  ..\n");
          DemoLogf("  16 = TS%d of Li0\n", g_nBurstHighwayNo*2 + 11);
          DemoLogf("  17 = TS%d of Li1\n", g_nBurstHighwayNo*2 + 11);
          DemoLogf("  etc.\n");
          break;
     
	      default:
	        break;

        }

        DemoLogf("Or press Enter to accept this value [%d]: ", nBurstTsToRecord);

        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          break;
        } else {
          nBurstTsToRecord = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].nBurstTsToRecord = nBurstTsToRecord;

        if (nBurstTsToRecord == -1) {
          DemoLogf("Recording all timeslots of the burst highway\n");
        } else {
          DemoLogf("Recording TS%d of the burst highway %d\n", nBurstTsToRecord, g_nBurstHighwayNo);
        }

        DemoLogf("Press 'a' to stop the recording\n");
        g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].nSourceId = g_nBurstHighwayNo;
        g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bGammaTestMode = OTX_FALSE;
        g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bActive = OTX_TRUE;

      } else {
        DemoLogf("Burst Receiver[%d] Thread already running.\n", g_nBurstHighwayNo);
      }
    }
    break;

  case 'X':
    {
      char sqzBufBase[OTX_MAX_PATH];
      char sqzBufH0[OTX_MAX_PATH];
      char sqzBufH1[OTX_MAX_PATH];
      char sqzBufH2[OTX_MAX_PATH];
      char sqzBufH3[OTX_MAX_PATH];
      int i;


      sqzBufBase[0] = 0;
      sqzBufH0[0]   = 0;
      sqzBufH1[0]   = 0;
      sqzBufH2[0]   = 0;
      sqzBufH3[0]   = 0;

      if (g_hLogicDevs.mBurstReceiver[g_nBurstHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("BurstReceiver: Please enter file to record to: ");
        OsaConsoleGetStr(sqzBufBase);
        if (sqzBufBase[0] == '\0') {
          break;
        }

        strcat(sqzBufH0, "highway0");
        strcat(sqzBufH0, sqzBufBase);
        if ((g_hLogicDevs.mBurstReceiver[0].pRxFile = fopen(sqzBufH0, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBufH0);
          break;
        }

        strcat(sqzBufH1, "highway1");
        strcat(sqzBufH1, sqzBufBase);
        if ((g_hLogicDevs.mBurstReceiver[1].pRxFile = fopen(sqzBufH1, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBufH1);
          break;
        }

        strcat(sqzBufH2, "highway2");
        strcat(sqzBufH2, sqzBufBase);
        if ((g_hLogicDevs.mBurstReceiver[2].pRxFile = fopen(sqzBufH2, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBufH2);
          break;
        }

        strcat(sqzBufH3, "highway3");
        strcat(sqzBufH3, sqzBufBase);
        if ((g_hLogicDevs.mBurstReceiver[3].pRxFile = fopen(sqzBufH3, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBufH3);
          break;
        }


        DemoLogf("Recording all timeslots of the burst highways\n");
        for(i = 0; i<4; i++) {
          g_hLogicDevs.mBurstReceiver[i].nBurstTsToRecord = -1;

          g_hLogicDevs.mBurstReceiver[i].nSourceId = i;
          g_hLogicDevs.mBurstReceiver[i].bGammaTestMode = OTX_FALSE;
          g_hLogicDevs.mBurstReceiver[i].bActive = OTX_TRUE;
        }
        DemoLogf("Press 'a' to stop the recording\n");
      } else {
        DemoLogf("Burst Receiver[%d] Thread already running.\n", g_nBurstHighwayNo);
      }
    }
    break;


  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case 'n':
    {
      // Select current g_nCurrentLi
      g_nCurrentLi++;
      if (g_nCurrentLi > (g_demoEnv.coLi-1)) {
        // Wrap 
        g_nCurrentLi = 0;
      }
      
      DemoLogf("Li#%d Selected \n", g_nCurrentLi);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current g_nCurrentLi
      g_nCurrentLi--;
      if (g_nCurrentLi < 0) {
        // Wrap
        g_nCurrentLi = g_demoEnv.coLi-1;
      }
      
      DemoLogf("Li#%d Selected \n", g_nCurrentLi);
    }
    break;

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;
      int c;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      DemoLogf("4 = OTX_T1E1_TERM_300_OHM (for highly attenuated signals)\n");
      c = OsaConsoleGetCh();
      DemoLogf("'%c' pressed\n", c);
      switch (c - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      case 4:
        eTermMode = OTX_T1E1_TERM_300_OHM;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[g_nCurrentLi], eTermMode); 
      DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

  case 'Z':
    {
      OtxT1E1TermModeE eTermMode;
      int c;
      int nLi;
      OTX_UINT32 nVoltage;
      double nVolts;
      double nTestValue = 0;
      double nTestTolerance = 0;
      double nTestValue120[16];
      double nTestValue75[16];
      double nTestValue300[16];

      OTX_HANDLE hLiDUT;
      char *pTestString; // Used to print a message when test fails
      char pTest120[]   = "120 ohm";
      char pTest75[]    = "75 ohm";
      char pTestHighZ[] = "HighZ";
      char pTest300[]   = "300 ohm";
      OTX_BOOL bTestResult = OTX_TRUE; // Start with good

      /* This test uses the voltmeter for each span to verify 
      ** the termination resistor is working correctly.
      ** This is also a test of the voltmeter.
      **
      ** The test requires that 100 Ohm resistors are used to
      ** connect each line to the data source. The data source
      ** used for this test was a thor-4-pci-plus card.
      */ 
      DemoLogf("Test line terminations using voltmeter\n");

      for(nLi=0; nLi < g_demoEnv.coLi; nLi++) {
        hLiDUT = g_hPhysDevs.hLi[nLi];
        // Make sure all Li are in High Z to start
        result = OtxT1E1SetLineTermination(hLiDUT, OTX_T1E1_TERM_HIGHZ); 
        DemoCheckRc(result);
      }

      for(nLi=0; nLi < g_demoEnv.coLi; nLi++) {
        bTestResult = OTX_TRUE;
        hLiDUT = g_hPhysDevs.hLi[nLi];
        DemoLogf("   Testing span%02d\n", nLi);

        for(c=1; c < 5; c++) {
          switch(c) {
          case 0:
          case 1:
            // test as if one case
            eTermMode = OTX_T1E1_TERM_120_OHM;
            pTestString = pTest120;
            // The following test values were collected from several
            // test cards as meassures by the cards voltmeter.
            nTestValue = 1.4; // 120 ohm
            nTestTolerance = .70; // set tolerance to .7 Volts
            break;
          case 2:
            eTermMode = OTX_T1E1_TERM_75_OHM;
            pTestString = pTest75;
            // The following test values were collected from several
            // test cards as meassures by the cards voltmeter.
            nTestValue = 1.0; // 75 ohm
            nTestTolerance = .70; // set tolerance to .7 Volts
            break;
          case 3:
            eTermMode = OTX_T1E1_TERM_300_OHM;
            pTestString = pTest300;
            // The following test values were collected from several
            // test cards as meassures by the cards voltmeter.
            nTestValue = 2.0; // 300 ohm
            nTestTolerance = .70; // set tolerance to .7 Volts
            break;
          case 4:
            // Perform this setting last so DUT is left in HighZ
            eTermMode = OTX_T1E1_TERM_HIGHZ;
            pTestString = pTestHighZ;
            // The following test values were collected from several
            // test cards as meassures by the cards voltmeter.
            nTestValue = 3.2; // HighZ
            nTestTolerance = .90; // set tolerance to .9 Volts
            break;
          default:
            DemoLogf("Invalid entry. Setting Line Termination to 100 OHM\n");
            eTermMode = OTX_T1E1_TERM_100_OHM;
            pTestString = pTest120;
            break;
          }

          result = OtxT1E1SetLineTermination(hLiDUT, eTermMode); 
          DemoCheckRc(result);

          // Sleep a little while to let setting time to adjust
          OsaThreadSleep(100); 

          result = OtxDrvGetAttributeValueUint32(hLiDUT, 
                                                 OTX_ATTR_T1E1_LINE_VOLTAGE, 
                                                 &nVoltage); 
          DemoCheckRc(result);
      
          // Check using voltage tolerance
          nVolts = (double) nVoltage/1000;
          if(nVolts > (nTestValue + nTestTolerance)) {
            // The Li failed
            DemoLogf("   Testing span%02d, %s - FAILED\n", nLi, pTestString);
            DemoLogf("     Expected %2.3f V, Actual %2.3f V\n", 
                    (double)(nTestValue + nTestTolerance), 
                    (double)nVolts);

            bTestResult = OTX_FALSE;
          }
          if(nVolts < (nTestValue - nTestTolerance)) {
            // The Li failed
            DemoLogf("   Testing span%02d, %s - FAILED\n", nLi, pTestString);
            DemoLogf("     Expected %2.3f V, Actual %2.3f V\n", 
                    (double)(nTestValue - nTestTolerance), 
                    (double)nVolts);

            bTestResult = OTX_FALSE;
          }

          /* 
          ** There can be a large difference in reading because of the date
          ** source and connections to source but difference between the
          ** values of 120 75, and 300 Ohms should be close to the tolerance
          ** of just the parts on the card. Save off the value for each to be
          ** check later. 
          */
          switch(c) {
          case 0:
          case 1:
            // test as if one case
            nTestValue120[nLi] = nVolts;
            break;
          case 2:
            nTestValue75[nLi] = nVolts;
            break;
          case 3:
            nTestValue300[nLi] = nVolts;
            break;
          case 4:
            // No additional check performed using this value
            break;
          default:
            DemoLogf("Invalid entry. Setting Line Termination to 100 OHM\n");
            eTermMode = OTX_T1E1_TERM_100_OHM;
            pTestString = pTest120;
            break;
          }
        } // end of for(c=1; c < 5; c++) Termination selection

        // Check the difference between readings and expected

        // Reading for 120 Ohm should be close to 1.4 and 75 Ohm should be 1.0.
        // The difference is .4. Check to see if the two values are about 
        // .4 Volt apart.
        nVolts = nTestValue120[nLi] - nTestValue75[nLi];
        if(nTestValue120[nLi] < nTestValue75[nLi]) {
          // This is not correct.
          DemoLogf("   Testing 120 to 75 Ohm value span%02d - FAILED\n", 
                   nLi);
          DemoLogf("    Expected V120 (%2.3f) to be larger then V75 (%2.3f)\n", 
                  (double)nTestValue120[nLi],
                  (double)nTestValue75[nLi]);

          bTestResult = OTX_FALSE;
        }
        else
        if((nVolts > .7) || (nVolts < .1)) {
          DemoLogf("   Testing 120 to 75 Ohm value span%02d - FAILED\n", 
                   nLi);
          DemoLogf("    Expected .4V, with .3V tolerance, Actual %2.3fV\n", 
                  (double)nVolts);

          bTestResult = OTX_FALSE;
        }

        // Reading for 120 Ohm should be close to 1.4 and 300 Ohm should be 2.0.
        // The difference is .6. Check to see if the two values are about 
        // .6 Volt apart.
        nVolts = nTestValue300[nLi] - nTestValue120[nLi];
        if(nTestValue300[nLi] < nTestValue120[nLi]) {
          // This is not correct.
          DemoLogf("   Testing 300 to 120 Ohm value span%02d - FAILED\n", 
                   nLi);
          DemoLogf("    Expected V300 (%2.3f) to be larger then V120 (%2.3f)\n", 
                  (double)nTestValue300[nLi],
                  (double)nTestValue120[nLi]);

          bTestResult = OTX_FALSE;
        }
        else
        if((nVolts > .9) || (nVolts < .3)) {
          DemoLogf("   Testing 120 to 300 Ohm value span%02d - FAILED\n", 
                   nLi);
          DemoLogf("    Expected .6V, with .3V tolerance, Actual %2.3fV\n", 
                  (double)nVolts);

          bTestResult = OTX_FALSE;
        }

        // Reading for 75 Ohm should be close to 1.0 and 300 Ohm should be 2.0.
        // The difference is 1. Check to see if the two values are about 
        // 1 Volt apart.
        nVolts = nTestValue300[nLi] - nTestValue75[nLi];
        if(nTestValue300[nLi] < nTestValue75[nLi]) {
          // This is not correct.
          DemoLogf("   Testing 300 to 75 Ohm value span%02d - FAILED\n", 
                   nLi);
          DemoLogf("    Expected V300 (%2.3f) to be larger then V75 (%2.3f)\n", 
                  (double)nTestValue300[nLi],
                  (double)nTestValue75[nLi]);

          bTestResult = OTX_FALSE;
        }
        else
        if((nVolts > 1.3) || (nVolts < .7)) {
          DemoLogf("   Testing 75 to 300 Ohm value span%02d - FAILED\n", 
                   nLi);
          DemoLogf("    Expected 1V, with .3V tolerance, Actual %2.3fV\n", 
                  (double)nVolts);

          bTestResult = OTX_FALSE;
        }


        if(bTestResult == OTX_FALSE) {
          DemoLogf("   Testing span%02d -FAILED-\n", nLi);
        }
        else {
          DemoLogf("span%02d OK\n", nLi);
        }
      }
    }
              
    DemoLogf("Test finished\n");
    break;

  case 'T':
    // reset stats
    OtxDrvDbgDump(g_hPhysDevs.hBurst, 101);
    break;

  case 'D':
    // dump stats to debugger
    OtxDrvDbgDump(g_hPhysDevs.hBurst, 100);
    break;

  case 'x':
    // Hardcoded cross-connects
    OtxDrvDbgDump(g_hPhysDevs.hBoardController, 200);
    break;

  case '#':
    {
      OTX_UINT32 nValue;
      char sValue[80];

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(result);
      DemoLogf("Enter queue notification value (1-32) [currently %d]: ", nValue);
      OsaConsoleGetStr(sValue);
      if (sValue[0] == '\0') {
        break;
      }
      nValue = strtol(sValue, NULL, 0);

      // before setting the new trigger value, the burst device must not be enabled
      result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, nValue); DemoCheckRc(result);
      result = OtxDrvEnable(g_hPhysDevs.hBurst,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(result);
      DemoLogf("New queue notification value is %d\n", nValue);

      break;
    }

  case 'q':
  case 'Q':
    return 1;
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP)
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_GIMLE_16_PCI_PLUS;
  envP->coLi              = 16;
  envP->coQuadE1T1        = 4;
  envP->coBurstHighways   = 4;
  envP->nHighwaySpeed     = OTX_HIGHWAY_SPEED_8MBS;
  envP->coDsp             = 4;
  envP->dspFileName[0]    = '\0';

  envP->eClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->eTermMode         = OTX_T1E1_TERM_HIGHZ;
  envP->bMonitorMode      = OTX_FALSE;
  envP->bAmplifier        = OTX_FALSE;
  envP->nRxThreshold      = DEMO_DEFAULT_RX_THRESHOLD; // in mV

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'P') {   // DSP file
          strcpy(envP->dspFileName, envP->cwd);
          strcat(envP->dspFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'G') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->coLi = 16;
            envP->coQuadE1T1 = 4;
            envP->coBurstHighways = 4;
            envP->nHighwaySpeed = OTX_HIGHWAY_SPEED_8MBS;
          } else if ( *(argv[i]+2) == '9') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCIE;
            envP->coLi = 16;
            envP->coQuadE1T1 = 4;
            envP->coBurstHighways = 4;
            envP->nHighwaySpeed = OTX_HIGHWAY_SPEED_8MBS;
          } else if ( *(argv[i]+2) == '3') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCI_BASIC_32;
            envP->coLi = 16;
            envP->coQuadE1T1 = 4;
            envP->coBurstHighways = 1;
            envP->nHighwaySpeed = OTX_HIGHWAY_SPEED_2MBS;
          } else if ( *(argv[i]+2) == '6') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCI_BASIC_64;
            envP->coLi = 16;
            envP->coQuadE1T1 = 4;
            envP->coBurstHighways = 2;
            envP->nHighwaySpeed = OTX_HIGHWAY_SPEED_2MBS;
          } else if ( *(argv[i]+2) == '5') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCIE_BASIC_32;
            envP->coLi = 16;
            envP->coQuadE1T1 = 4;
            envP->coBurstHighways = 1;
            envP->nHighwaySpeed = OTX_HIGHWAY_SPEED_2MBS;
          } else if ( *(argv[i]+2) == '7') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCIE_BASIC_64;
            envP->coLi = 16;
            envP->coQuadE1T1 = 4;
            envP->coBurstHighways = 2;
            envP->nHighwaySpeed = OTX_HIGHWAY_SPEED_2MBS;
          } else if ( *(argv[i]+2) == 'R') {
            envP->nBoardType = OTX_DEVICE_THOR_4_EXPRESSCARD;
            envP->coLi = 4;
            envP->coBurstHighways = 1;
            envP->nHighwaySpeed = OTX_HIGHWAY_SPEED_8MBS;
            envP->coQuadE1T1 = 1;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i] + 1) == 'A') {
          // Use Front End Amplifier = ON
          envP->bAmplifier = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'E') {
          // Use ReceiveEqualizer  = ON
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i] + 1) == '2') {
          // Use MonitorMode  = ON
          envP->bMonitorMode = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'H') {
          // RxThreshold (in mV)
          envP->nRxThreshold = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'Y') {
          // TermMode
          if ( *(argv[i]+2) == '1') {
            envP->eTermMode = OTX_T1E1_TERM_100_OHM;
          } else if ( *(argv[i]+2) == '3') {
            envP->eTermMode = OTX_T1E1_TERM_300_OHM;
          } else if ( *(argv[i]+2) == '7') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM;
          } else if ( *(argv[i]+2) == 'Z') {
            envP->eTermMode = OTX_T1E1_TERM_HIGHZ;
          } else if ( *(argv[i]+2) == 'S') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'O') {   // Clock Source
          if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '0') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_10;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '1') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_11;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '2') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_12;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '3') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_13;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '4') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_14;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '5') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_15;
          } else if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == '2') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else if (*(argv[i]+2) == '6') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_6;
          } else if (*(argv[i]+2) == '7') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_7;
          } else if (*(argv[i]+2) == '8') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_8;
          } else if (*(argv[i]+2) == '9') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_9;
          } else if ( (*(argv[i]+2) == 'H') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_A_CLOCKS_ETCF;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'N') {
            envP->eT1FrameFormat = OTX_T1_FF_UNFRAMED; // Transparent mode
          } else if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'N') {
            envP->eE1FrameFormat = OTX_E1_FF_UNFRAMED; // Transparent mode
          } else if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[G|9|3|6|R]  : Board Type used for the demo, where\n");
  DemoLogf("                  G = Use Gimle-16-PCI-Plus Board\n");
  DemoLogf("                  9 = Use Gimle-16-PCIe Board\n");
  DemoLogf("                  3 = Use Gimle-16-PCI-Basic-32 Board\n");
  DemoLogf("                  6 = Use Gimle-16-PCI-Basic-64 Board\n");
  DemoLogf("                  5 = Use Gimle-16-PCIe-Basic-32 Board\n");
  DemoLogf("                  7 = Use Gimle-16-PCIe-Basic-64 Board\n");
  DemoLogf("                  R = Use Thor-4-ExpressCard in receive only mode\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /O[x]        : Clock Source, where\n");
  DemoLogf("                  x = 0 through 15 indicating Li to recover the clock from or\n");
  DemoLogf("                  x = H to recover from H.100 A-clock (ETCF mode)\n");
  DemoLogf(" /A           : Front End Amplifier = ON (default is OFF)\n");
  DemoLogf(" /E           : Receive Equalizer = ON (longhaul) (default is OFF)\n");
  DemoLogf(" /2           : -20dB MonitorMode = ON (default is OFF)\n");
  DemoLogf(" /H           : Receive Input Threshold in mV (default is %d mV)\n", DEMO_DEFAULT_RX_THRESHOLD);
  DemoLogf(" /Y           : Termination Mode, where\n");
  DemoLogf("                  1 = 100/120 ohm balanced\n");
  DemoLogf("                  3 = 300 ohm balanced (for highly attenuated signals)\n");
  DemoLogf("                  7 = 75 ohm unbalanced\n");
  DemoLogf("                  Z = High-impedance (default)\n");
  DemoLogf("                  S = 75 ohm balanced\n");
  DemoLogf(" /F[N|E|S|9]  : T1 Frame Format, where\n");
  DemoLogf("                  N = Not framed/Unframed\n");
  DemoLogf("                  E = ESF (default)\n");
  DemoLogf("                  4 = F4\n");
  DemoLogf("                  S = SF [D3/D4] (F12)\n");
  DemoLogf("                  9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]      : T1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  B = B8ZS (default)\n");
  DemoLogf(" /G[N|D|C|G]  : E1 Frame Format, where\n");
  DemoLogf("                  N = Not framed/Unframed\n");
  DemoLogf("                  D = Doubleframe\n");
  DemoLogf("                  C = CRC4 Multiframe (default)\n");
  DemoLogf("                  G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]      : E1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  H = HDB3 (default)\n");
  DemoLogf(" /P<fileName> : Connect the Burst Device to the DSPs\n");
  DemoLogf("                and load the specified DSP program.\n");
  DemoLogf("                This option requires a Vidar-55x4-ASM DSP daughter board.\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(const char *formatStr, ...)
{
  va_list argList;

  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF((char *)formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coBurstHighways; i++) {
    if (g_hLogicDevs.mBurstReceiver[i].pRxFile) {
      fclose(g_hLogicDevs.mBurstReceiver[i].pRxFile);
      g_hLogicDevs.mBurstReceiver[i].pRxFile = NULL;
    }
  }

  fclose(g_pDemoLogFile);
}

int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  printf("T8pDma, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  //OtxDrvDbgDump(g_hPhysDevs.hBoardController, 200);

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  {
    char szClockSource[64];

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szClockSource, "Internal");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
    case OTX_CLOCK_SOURCE_LOCAL_8:
    case OTX_CLOCK_SOURCE_LOCAL_9:
    case OTX_CLOCK_SOURCE_LOCAL_10:
    case OTX_CLOCK_SOURCE_LOCAL_11:
    case OTX_CLOCK_SOURCE_LOCAL_12:
    case OTX_CLOCK_SOURCE_LOCAL_13:
    case OTX_CLOCK_SOURCE_LOCAL_14:
    case OTX_CLOCK_SOURCE_LOCAL_15:
      sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
      break;
    case OTX_CLOCK_SOURCE_A_CLOCKS_ETCF:
      sprintf(szClockSource, "H.100 A-Clocks (ETCF mode)");
      break;
    default:
      strcpy(szClockSource, "Unknown");
      break;
    }

    DemoLogf("Clock Source: %s\n", szClockSource);

    result = OtxBrdSetClocks(g_hPhysDevs.hBoardController, OTX_CLOCK_MODE_H100_ETCF, g_demoEnv.eClkSource); DemoCheckRc(result);
  }

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\nExiting...\n");
  {
    OTX_INT32 i;


    g_bStatsThreadActive = OTX_FALSE;
    for (i = 0; i < g_demoEnv.coBurstHighways; i++) {
      g_hLogicDevs.mBurstReceiver[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}
