/* $Workfile: T1E1HostedHdlc.c $       $Revision: 6 $      $Modtime: 12/03/12 4:55p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T1E1HostedHdlc.c

    A simple example console application demonstrating the use of OTX Driver
    with the following boards:
    
      Thor-2-PCI-Express 
      Thor-8-PCI-Express 
      Thor-1-PCI-Plus
      Thor-2-PCI-Plus
      Thor-4-PCI-Plus
      Thor-8-PCI-Plus, 
      Sleipnir-1-PCI
      Thor-2-ExpressCard
      Thor-4-ExpressCard
      Thor-8-Usb
      Alvis-2-CSI
      Alvis-4-CSI
      Alvis-4M-CSI
      Alvis-8-CSI
      Alvis-8M-CSI
      Alvis-2-PBX
      Alvis-4-PBX
      Alvis-8-PBX

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures a number of E1 Lines and expects them to be looped with an external
    loop cable. The LI Internal HDLC controller is used to send HDLC frames from
    one Line Interface to be received by the other.

    4 Handsets are connected to the codecs and a through connection is
    demonstrated through the Loop Cable (only available for certain OTX boards).

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDsp.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"

#define DEMO_DEFAULT_LOG_FILE "T1E1HostedHdlc.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_SS7_FISU_FILTERING 1

#define DEMO_MAX_MULTIPLE_FRAME_READ 128

//#define DEMO_USE_BURST_DEVICE 1

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  OtxDataFormatE         eLaw;
  int                    boardNo;
  int                    coLi;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL               bSs7FisuFiltering;
  OTX_BOOL               bLoopHandsetsInternally;
  const char            *pszDongleType;
  OTX_UINT16             nLineLength;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OTX_BOOL               b56kbpsChannel;
  int                    nTs;
  OTX_BOOL               bHdlcCrcCheck;  // Default OTX_TRUE
  OTX_BOOL               bHdlcSharedFlags;
  OTX_BOOL               bHdlcBitInversion; 
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(const char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_BOOL    g_bHdlcPayloadCheck = OTX_TRUE;  // Only in affect when g_bHdlcVerbose = OTX_FALSE
OTX_UINT32  g_coHdlcRxFrames[32];
OTX_UINT32  g_coHdlcBadRxFrames[32];
OTX_UINT8   g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_INT32   g_nSleepTimeBetweenHdlcFrames = 0;
OTX_INT32   g_cbHdlcFrame = 10;
OTX_INT32   g_coHdlcSenderFrames = 1;
OTX_UINT8   g_nSs7FilterBitMask = 0;

struct PhysDevs{
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; 
  OTX_HANDLE hTss0;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hCodec[4];
#ifdef DEMO_USE_BURST_DEVICE
  OTX_HANDLE hBurst; 
#endif
} g_hPhysDevs;


struct Pipes {
  OTX_HANDLE hLiCodec[2];
  OTX_HANDLE hCodecLi[2];
  OTX_HANDLE hCodecCodec[2];
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
  OTX_HANDLE hHs3MicPipe;
  OTX_HANDLE hHs3SpkrPipe;
  OTX_HANDLE hHs2MicPipe;
  OTX_HANDLE hHs2SpkrPipe;
} g_hPipes;


struct LogicDevs {
  OTX_HANDLE hHdlcSender[8];
  OTX_HANDLE hHdlcReceiver[8];
} g_hLogicDevs;


struct T1E1CfgData {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************

  @func

  Sets up the board for RX/RX mode. Controls the T1E1 Xmit Pin Control bit

*******************************************************************************/
void DemoSetRxRxMode(void)
{
  OTX_RESULT result;
#if 0
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(result);
    DemoLogf("OtxT1E1XmitPinControl for Line #%d set %s\n", i, "FALSE (Tri-state)");
  }
#else
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, 
                                         OTX_ATTR_BRD_MONITOR_ONLY, 
                                         OTX_TRUE); 
  DemoCheckRc(result);
  DemoLogf("Configured board for RX/RX operation\n"); 
#endif
}

/*******************************************************************************

  @func

  Sets up the board for TX/RX mode. Controls the T1E1 Xmit Pin Control bit

*******************************************************************************/
void DemoSetTxRxMode(void)
{
  OTX_RESULT result;
#if 0
  OTX_INT32 i;

  for (i = 0; i < 3; i++) {
    result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_TRUE); DemoCheckRc(result);
    DemoLogf("OtxT1E1XmitPinControl for Line #%d set %s\n", i, "TRUE (Active)");
  }
  for (i = 4; i < 7; i++) {
    result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(result);
    DemoLogf("OtxT1E1XmitPinControl for Line #%d set %s\n", i, "FALSE (Tri-state)");
  }
#else
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, 
                                         OTX_ATTR_BRD_MONITOR_ONLY, 
                                         OTX_FALSE); 
  DemoCheckRc(result);
  DemoLogf("Configured board for TX/RX operation\n"); 
#endif
}


typedef struct FalcRegValT {
  OTX_INT32 address;
  OTX_UINT8 value;
} FalcRegValT;



/*******************************************************************************

  @func

  Dump the contents of the falc registers for the purpose of debugging..

*******************************************************************************/
void DemoDumpFalcRegisters(void)
{
  OTX_INT32 i;
  OTX_UINT32 nAddress, readVal;
  OTX_RESULT result;
  FILE *fout;
  FILE *fout2;

  fout = fopen("./FalcRegisterDump", "wt");
  if (!fout) {
    DemoLogf("\nCould not open FalcRegisterDump file\n");
    return;
  }
  fout2 = fopen("./ThorAccessRegFile", "wt");
  if (!fout2) {
    DemoLogf("\nCould not open ThorAccessRegFile file\n");
    return;
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    DemoLogf("\nFalc[%d] register dump\n=====================\n", i);
    for (nAddress = 0; nAddress < 0x100; nAddress++) {
      if (nAddress >= 0x00 && nAddress <= 0x02) continue;
      if (nAddress >= 0x04 && nAddress <= 0x07) continue;
      if (nAddress == 0x0D) continue;
      if (nAddress == 0x0E) continue;
      if (nAddress == 0x0F) continue;
      if (nAddress == 0x11) continue;
      if (nAddress == 0x12) continue;
      if (nAddress == 0x13) continue;
      if (nAddress >= 0x2D && nAddress <= 0x30) continue;
      if (nAddress >= 0x33 && nAddress <= 0x35) continue;
      if (nAddress == 0x3D) continue;
      if (nAddress >= 0x49 && nAddress <= 0x5F) continue;
      if (nAddress >= 0x61 && nAddress <= 0x6F) continue;
      if (nAddress >= 0x72 && nAddress <= 0x7F) continue;
      if (nAddress >= 0x9A && nAddress <= 0x9B) continue;
      if (nAddress == 0xA6) continue;
      if (nAddress >= 0xA9 && nAddress <= 0xBF) continue;
      if (nAddress >= 0xC2 && nAddress <= 0xD2) continue;
      if (nAddress >= 0xE0 && nAddress <= 0xFF) continue;

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[i], nAddress, &readVal);
      if (OTX_RESULT_CODE(result) != OTX_E_INVALIDARG) {
        DemoCheckRc(result);
      }
      fprintf(fout, "./ThorAccess /Zw0x%.4X^0x%.2X^q\n", nAddress + i*0x100, readVal); 
      fprintf(fout2, "0x%.4X,0x%.2X\n", nAddress + i*0x100, readVal); 
      if ((nAddress & 0xf) == 0x0) {
        DemoLogf("0x%.2X  ", nAddress);
      }
      if ((nAddress & 0xf) == 0x8) {
        DemoLogf("  ");
      }
      DemoLogf(" %.2X", (OTX_UINT8) readVal);
      if ((nAddress & 0xf) == 0xf) {
        DemoLogf("\n");
      }
    }
    if ((nAddress & 0xf) == 0xf) {
      DemoLogf("\n");
    }
  }

  fclose(fout);
  fclose(fout2);
}


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects  
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor-2-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-8-PCI
  OTX_UINT32 nDeviceId;
  OTX_UINT32 nSubsystemId;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI Plus board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_THOR_8_PCIE:
    DemoLogf("Opening Thor-8-PCIe board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI Plus board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir board\n");
    g_demoEnv.coLi = 1;
    break;
  case OTX_DEVICE_HUGIN_2_PC104:
    DemoLogf("Opening Hugin-2-PC104 board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    DemoLogf("Opening Thor-1-PCI-Plus board\n");
    g_demoEnv.coLi = 1;
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    g_demoEnv.coLi = 2; 
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    DemoLogf("Opening Thor-4-ExpressCard board\n");
    g_demoEnv.coLi = 4; // set line count 2 transmit/receive or 4 receive
    break;
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
    DemoLogf("Opening Alvis-8-PBX board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
    DemoLogf("Opening Alvis-4-PBX board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
    DemoLogf("Opening Alvis-2-PBX board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
    DemoLogf("Opening Alvis-2-CSI board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
    DemoLogf("Opening Alvis-4-CSI board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    DemoLogf("Opening Alvis-4M-CSI board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
    DemoLogf("Opening Alvis-8-CSI board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    DemoLogf("Opening Alvis-8M-CSI board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_THOR_8M_USB:
    DemoLogf("Opening Thor-8M-USB board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_THOR_8_USB:
    DemoLogf("Opening Thor-8-USB board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_THOR_4M_USB:
    DemoLogf("Opening Thor-4M-USB board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_THOR_2_USB:
    DemoLogf("Opening Thor-2-USB board\n");
    g_demoEnv.coLi = 2;
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(result);
    
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_HUGIN_2_PC104:
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
  case OTX_DEVICE_THOR_8M_USB:
  case OTX_DEVICE_THOR_8_USB:
  case OTX_DEVICE_THOR_4M_USB:
  case OTX_DEVICE_THOR_2_USB:
    break; // No TSS device on these boards
  default:
    // Get a handle to the Time-Space Switch (TSS) on the Thor-PCI board
    DemoLogf("Open OTX_DEVICE_TSS\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss0); DemoCheckRc(result);
    DemoLogf("Open OTX_DEVICE_TSS\n");
    result = OtxDrvEnable(g_hPhysDevs.hTss0, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    break;
  }

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(result);
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SUBSYSTEM_ID, &nSubsystemId); DemoCheckRc(result);
  DemoLogf("Board Device ID: 0x%.4X, SubsystemId: 0x%.4X\n", nDeviceId, nSubsystemId);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_HUGIN_2_PC104:
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;

  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCIE:
    // On the Thor-8-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;

  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    // Thor-4-ExpressCard by default is set to receive on the four lines 
    // (monitor mode) with all transmitters disabled (tristate). This 
    // demo however needs the card to support the transmit data so the 
    // card will use 2 receivers and the 2 transmitters.
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, 
                                         OTX_ATTR_BRD_MONITOR_ONLY, 
                                         OTX_FALSE); 
    DemoCheckRc(result);
    DemoLogf("Configure Thor-4-ExpressCard for data transmit and receive\n"); 
    // no 'break' here, fall through

  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    // On the Thor-4-PCI board and Thor-4-ExpressCard, the E1T1 devices are hosted by 
    // a quad E1T1 device.
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
  case OTX_DEVICE_THOR_8_USB:
  case OTX_DEVICE_THOR_8M_USB:
    // On the Alvis-8-CSI board and Alvis-8M-CSI, the E1T1 devices are hosted by 
    // a octal E1T1 device.
    nQuadE1T1 = 0;
    DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = 0;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_4M_USB:
    nQuadE1T1 = 0;
    DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      OTX_INT32 nLiNumber;
      nQuadE1T1 = 0;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      switch (i) {
      case 0: nLiNumber = 1; DemoLogf("Mapping Li#%d to physical Li#%d\n", i, nLiNumber); break;
      case 1: nLiNumber = 5; DemoLogf("Mapping Li#%d to physical Li#%d\n", i, nLiNumber);  break;
      case 2: nLiNumber = 2; DemoLogf("Mapping Li#%d to physical Li#%d\n", i, nLiNumber);  break;
      case 3: nLiNumber = 6; DemoLogf("Mapping Li#%d to physical Li#%d\n", i, nLiNumber);  break;
      }
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, nLiNumber, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_2_USB:
    nQuadE1T1 = 0;
    DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      OTX_INT32 nLiNumber;
      nQuadE1T1 = 0;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      switch (i) {
      case 0: nLiNumber = 1; DemoLogf("Mapping Li#%d to physical Li#%d\n", i, nLiNumber);  break;
      case 1: nLiNumber = 2; DemoLogf("Mapping Li#%d to physical Li#%d\n", i, nLiNumber);  break;
      }
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, nLiNumber, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
    // On the Alvis-PBX board the E1T1 devices are hosted by a dual E1T1 device. 
    DemoLogf("Open OTX_DEVICE_DUAL_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[0]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[0], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    // On the Thor-2-ExpressCard board the E1T1 devices are hosted by a dual E1T1 device. 
    DemoLogf("Open OTX_DEVICE_DUAL_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[0]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[0], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_HUGIN_2_PC104:
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
  case OTX_DEVICE_THOR_8M_USB:
  case OTX_DEVICE_THOR_8_USB:
  case OTX_DEVICE_THOR_4M_USB:
  case OTX_DEVICE_THOR_2_USB:
    // These boards do not have a Codec
    break;
  default:
    // Get handles to Codec devices #0,#1,#2 and #3
    for (i = 0; i < 4; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(result);

      // Set PCM compression law for the Codec
      DemoLogf("Set Law to %s\n", g_demoEnv.eLaw == OTX_DATA_FORMAT_COMPAND8_A_LAW ? "A-law" : "U-law");
      result = OtxBrdCodecSetLaw(g_hPhysDevs.hCodec[i], g_demoEnv.eLaw); DemoCheckRc(result);

    }
  }

#ifdef DEMO_USE_BURST_DEVICE
  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(result);

  //DemoLogf("Burst TX interrupts will be turned off when the Burst device is enabled\n");
  //result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SEND_DISABLED,  OTX_TRUE); DemoCheckRc(result);

  //DemoLogf("Burst RX interrupts will be turned off when the Burst device is enabled\n");
  //result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECV_DISABLED,  OTX_TRUE); DemoCheckRc(result);

  DemoLogf("Enable g_hPhysDevs.hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);

#endif

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_SLEIPNIR_1_PCI:
    case OTX_DEVICE_DMP_ALVIS_2_PBX:
    case OTX_DEVICE_DMP_ALVIS_4_PBX:
    case OTX_DEVICE_DMP_ALVIS_8_PBX:
      // These boards do not have software configurable line termination
      break;
    default:
      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);
      break;
    }
  
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);

  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_HANDLE hHostDevice;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
    hHostDevice = g_hPhysDevs.hLi[i];
    result = OtxDrvCreateLogicalDevice(hHostDevice, OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(result);
    result = OtxDrvCreateLogicalDevice(hHostDevice, OTX_LDEVICE_HDLC_SENDER, g_hEventQueue, i, &g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;
  OTX_INT32 nCapacity;

  if (g_demoEnv.b56kbpsChannel) {
    DemoLogf("Configuring one 56kbps HDLC channel per Li\n");
    nCapacity = 56;
  } else {
    DemoLogf("Configuring one 64kbps HDLC channel per Li\n");
    nCapacity = 64;
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes.hLiIn[i]); DemoCheckRc(result);
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes.hLiOut[i]); DemoCheckRc(result);

    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    if (g_demoEnv.nTs == -1) {
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        ts = 24; 
      } else {
        // E1 mode
        ts = 16; 
      }
    } else {
      ts = g_demoEnv.nTs;
    }
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    if (g_demoEnv.b56kbpsChannel) {
      sqTimeSlotMask[ts]   = 0xFE; // Timeslot on the T1/E1 span
    } else {
      sqTimeSlotMask[ts]   = 0xFF; // Timeslot on the T1/E1 span
    }

    DemoLogf("Connecting HDLC to TS%d (%s kbps)\n", ts, (g_demoEnv.b56kbpsChannel) ? "56" : "64");

    result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(result);
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(result);

    // Connect HDLC Sender and Receiver to the external T1/E1 Pipes
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiIn[i], -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hHdlcSender[i], -1, g_hPipes.hLiOut[i], -1); DemoCheckRc(result);
  }

  if (
      g_demoEnv.nBoardType != OTX_DEVICE_SLEIPNIR_1_PCI     && 
      g_demoEnv.nBoardType != OTX_DEVICE_HUGIN_2_PC104      && 
      g_demoEnv.nBoardType != OTX_DEVICE_THOR_2_EXPRESSCARD &&
      g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS_8_PBX    &&
      g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS_4_PBX    &&
      g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS_2_PBX    &&
      g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS_2_CSI    &&
      g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS_4_CSI    &&
      g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS_4M_CSI   &&
      g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS_8_CSI    &&
      g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS_8M_CSI   &&
      g_demoEnv.nBoardType != OTX_DEVICE_THOR_4_EXPRESSCARD &&
      g_demoEnv.nBoardType != OTX_DEVICE_THOR_8M_USB        &&
      g_demoEnv.nBoardType != OTX_DEVICE_THOR_8_USB         &&
      g_demoEnv.nBoardType != OTX_DEVICE_THOR_4M_USB        &&
      g_demoEnv.nBoardType != OTX_DEVICE_THOR_2_USB         &&
      1) {

    if (g_demoEnv.bLoopHandsetsInternally) {
      // Connect codecs internally (via the time-space switch only)
      for (i = 0; i < 4; i++) {
        result = OtxDrvCreatePipe(64, &g_hPipes.hCodecCodec[i]); DemoCheckRc(result);
        
        // Create voice pipes for Codec0 <--> Codec1 and Codec2 <--> Codec3
        result = OtxDrvConnectPipe(g_hPipes.hCodecCodec[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(result);
        if (i % 2) {
          // Odd Codec
          result = OtxDrvConnectPipe(g_hPipes.hCodecCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i-1], -1, NULL, 0); DemoCheckRc(result);
        } else {
          // Even Codec
          result = OtxDrvConnectPipe(g_hPipes.hCodecCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i+1], -1, NULL, 0); DemoCheckRc(result);
        }
      }
    } else { 
      // Connect codecs via the T1/E1 span
  
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_1_PCI_PLUS) {

        for (i = 0; i < 4; i++) {

          // Create voice pipes for Li0 <--> Codec0 - Codec4 with different TS 
          result = OtxDrvCreatePipe(64, &g_hPipes.hLiCodec[i]); DemoCheckRc(result);
          result = OtxDrvCreatePipe(64, &g_hPipes.hCodecLi[i]); DemoCheckRc(result);
        
          // Connect Li to Codec, full duplex. Don't care Highway number for T1/E1 Li reverts to internal highway
          ts = 10 + i;
          memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
          sqTimeSlotMask[ts] = 0xFF;
        
          result = OtxDrvConnectPipe(g_hPipes.hLiCodec[i], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[0], -1, sqTimeSlotMask, 32); DemoCheckRc(result);
          result = OtxDrvConnectPipe(g_hPipes.hLiCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(result);

          result = OtxDrvConnectPipe(g_hPipes.hCodecLi[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(result);
          result = OtxDrvConnectPipe(g_hPipes.hCodecLi[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[0], -1, sqTimeSlotMask, 32); DemoCheckRc(result);
        }

      } else {
        
        for (i = 0; i < 4; i++) {

          // Create voice pipes for Li0 <--> Codec0 and Li1 <--> Codec1 and Li0 <--> Codec2 and Li1 <--> Codec3
          result = OtxDrvCreatePipe(64, &g_hPipes.hLiCodec[i]); DemoCheckRc(result);
          result = OtxDrvCreatePipe(64, &g_hPipes.hCodecLi[i]); DemoCheckRc(result);
        
          // Connect Li to Codec, full duplex. Don't care Highway number for T1/E1 Li reverts to internal highway
          // Codec0 and Codec1 --> timeslot 10 and Codec2 and Codec3 --> timeslot 11
          ts = 10 + i/2;
          memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
          sqTimeSlotMask[ts] = 0xFF;

          result = OtxDrvConnectPipe(g_hPipes.hLiCodec[i], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i%2], -1, sqTimeSlotMask, 32); DemoCheckRc(result);
          result = OtxDrvConnectPipe(g_hPipes.hLiCodec[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(result);
        
          result = OtxDrvConnectPipe(g_hPipes.hCodecLi[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(result);
          result = OtxDrvConnectPipe(g_hPipes.hCodecLi[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i%2], -1, sqTimeSlotMask, 32); DemoCheckRc(result);
        }
      }
    }
  }
}




/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Initialize the Hdlc Sender and Receiver
    result = OtxHdlcSenderInit(
      g_hLogicDevs.hHdlcSender[i],  // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,           // Fill character(s) to be send between HDLC frames
      (g_demoEnv.bHdlcSharedFlags) ? 1 : 2, // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,               // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion,  // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                     // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA);              // HDLC Transmission Rate for V.110 and X.30 modes
    DemoCheckRc(result);

    result = OtxHdlcReceiverInit(
      g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,            // Fill character(s) to be send between HDLC frames
      (g_demoEnv.bHdlcSharedFlags) ? 1 : 2, // Minimum No of interframe time characters (flags or AllOnes) between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,                // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion,   // If OTX_TRUE then all bits are inverted when transmitted and before received
      g_demoEnv.bHdlcCrcCheck,       // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC bytes are passed on to the receiving software (us)
      OTX_HDLC_TR_NA,                // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_FALSE);                    // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
    DemoCheckRc(result);

    if (g_demoEnv.bSs7FisuFiltering) {
      g_nSs7FilterBitMask = 
          OTX_HDLC_SS7_FILTER_FISU | 
          OTX_HDLC_SS7_FILTER_SHORT_LSSU | 
          OTX_HDLC_SS7_FILTER_LONG_LSSU | 
          0;
      result = OtxHdlcSS7SetFilter(g_hLogicDevs.hHdlcReceiver[i], g_nSs7FilterBitMask); DemoCheckRc(result);
    }

  }

}


/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Board
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  // If board type is Thor-8-PCI then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCIE) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_4_PBX || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_4_CSI || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_4M_CSI || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_8_PBX || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_8_CSI || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_8M_CSI) {
    // Enable Quad E1T1 device
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
    // Enable Quad E1T1 device
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD  || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_2_PBX || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_2_CSI) {
    // Enable Dual E1T1 device
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_DUAL_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
#if 0 // Enable only Li1
    if (i == 1) {
      continue;
    }
#endif
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Enable Logical HDLC Devices (Data Sinks)
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hHdlcSender[i], OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(result);

    result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(result);
  }
}

/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}




/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}

/*******************************************************************************
  @func
    Prints the current time in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintCurrentTime(
  void
)
{
  OTX_DATETIME  nTimeStamp;

  OtxDrvGetCurrentTime(&nTimeStamp);
  DemoPrintTimeStamp(nTimeStamp);
}


/*******************************************************************************
  @func

*******************************************************************************/
void  *DemoTimeStampTest(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  for (;;) {
    
    result = OtxDrvGetCurrentTime(&nTimeStamp); DemoCheckRc(result);
    
    //DemoLogf("%.8X%.8X (Current time)\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow);
    
    // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
    #if defined(__linux__)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
    #elif defined(WIN32)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
    #endif
    
    // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
    nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
    
    // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
    nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
    
    nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
    ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
    
    if (ptCurrentDayAndTime) {
      DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
        ptCurrentDayAndTime->tm_hour,
        ptCurrentDayAndTime->tm_min, 
        ptCurrentDayAndTime->tm_sec,
        nTimeStampMs.sUint32.nLow,
        asctime((struct tm *) ptCurrentDayAndTime)    // TODO: needs to be Osa function (OsaAscTime())
        );
    } else {
      DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
    }

    OsaThreadSleep(4000);
  }
  


  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
#if 0 // Single reads of events
      OtxEventDataS eventData;
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

#else // Multiple reads of events
#define DEMO_EVENT_DATA_BUF_SIZE 10
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        //DemoLogf("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);

        assert(coEventsFetched <= coEventsFetched);

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }

      }
#endif
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_LDEVICE_HDLC_RECEIVER:
    if (g_bHdlcVerbose) {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    }
    break;
  default:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    break;
  }
    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;
    
  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
      DemoLogf("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

#if 0 // Single Frame Read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
      OTX_UINT8  buf[1024];
      OTX_UINT32 nFrameLen;
      OTX_UINT32 nFrameSeq;
      OTX_UINT32 nFrameEndCode;
      OTX_DATETIME nTimeStamp;
      
      
      otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
              
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId]++;
        }
        
        
        if (g_bHdlcVerbose) {
          DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
          
          DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          for (i = 0; i < nFrameLen && i < 1024; i++) {
            DemoLogf("%.2X ", buf[i]);
          }
          DemoLogf("\n");
        } else {
          if (nFrameEndCode) {
            
            g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
            
            DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
            for (i = 0; i < nFrameLen && i < 256; i++) {
              DemoLogf("%.2X ", buf[i]);
            }
            DemoLogf("\n");
            
            if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
              DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
              otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(otxRc);
            }
            
            
          } else if (g_bHdlcPayloadCheck) {
            if ((pFrame->cbOctets != (OTX_UINT32)g_cbHdlcFrame) && (pFrame->cbOctets != 3)) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
            } else if (memcmp(pBuf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
            }
          }
        } 
        
        // Read next frame (if any)
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);

      } 
      if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#else // Multiple frame read
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
#define CB_BUF_SIZE 8192
      OTX_UINT8  buf[CB_BUF_SIZE];
      OTX_UINT8  *pBuf;
			OtxHdlcFrameInfoS *pFrame;
      OTX_UINT32 coFrames;
      OTX_UINT32 j;
      
      otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);
      while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      
        if (pEventData->m_nSourceId < 32) {
          g_coHdlcRxFrames[pEventData->m_nSourceId] += coFrames;
        }

        if (g_bHdlcVerbose) {
          DemoLogf("%d HDLC Frame(s) Received, Sequence Number = %d\n", coFrames, pEventData->m_nParam);
        }
        
        pFrame = (OtxHdlcFrameInfoS *) buf;

        for (j = 0; j < coFrames; j++) {

					pBuf = pFrame->sqData;

          if (g_bHdlcVerbose) {
            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
            for (i = 0; i < pFrame->cbOctets && i < CB_BUF_SIZE; i++) {
              DemoLogf("%.2X ", pBuf[i]);
            }
            DemoLogf("\n");
          } else {
            if (pFrame->nFrameEndCode) {
              
              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
              for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                DemoLogf("%.2X ", pBuf[i]);
              }
              DemoLogf("\n");
              
              if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
                DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
                otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(otxRc);
              }
              
              
            } else if (g_bHdlcPayloadCheck) {
              if ((pFrame->cbOctets != (OTX_UINT32)g_cbHdlcFrame) && (pFrame->cbOctets != 3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
                for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                  DemoLogf("%.2X ", pBuf[i]);
                }
                DemoLogf("\n");
                
              } else if (memcmp(pBuf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
                for (i = 0; i < pFrame->cbOctets && i < 256; i++) {
                  DemoLogf("%.2X ", pBuf[i]);
                }
                DemoLogf("\n");
                
              }
            }
          }  

          // Step pointer to next frame (if any)
					pFrame = OTX_STRUCT_ADD_BYTE_OFFSET(OtxHdlcFrameInfoS, pFrame, pFrame->nOffsetNext);
        }

        // Read next frame (if any)
        otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);

      } 
      if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadDataEx() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
#endif

    
  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;
    
  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

*******************************************************************************/
void DemoCloseDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;

  // Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
  for (i = 0; i < g_demoEnv.coLi; i++) {
    if (g_hLogicDevs.hHdlcSender[i]) {
      result = OtxDrvClose(g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(result);
    }
    if (g_hLogicDevs.hHdlcReceiver[i]) {
      result = OtxDrvClose(g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(result);
    }
  }

  // Incoming and outgoing Pipe representing the external T1/E1
  for (i = 0; i < g_demoEnv.coLi; i++) {
    if (g_hPipes.hLiIn[i]) {
      result = OtxDrvClose(g_hPipes.hLiIn[i]); DemoCheckRc(result);
    }
    if (g_hPipes.hLiOut[i]) {
      result = OtxDrvClose(g_hPipes.hLiOut[i]); DemoCheckRc(result);
    }
  }

  // Voice pipes 
  for (i = 0; i < g_demoEnv.coLi; i++) {
    if (g_hPipes.hLiCodec[i]) {
      result = OtxDrvClose(g_hPipes.hLiCodec[i]); DemoCheckRc(result);
    }
    if (g_hPipes.hCodecLi[i]) {
      result = OtxDrvClose(g_hPipes.hCodecLi[i]); DemoCheckRc(result);
    }
  }

}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_UNKNOWN;
  envP->coLi              = 2;
  envP->eLaw              = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->eT1E1Mode         = OTX_T1E1_MODE_T1;
  envP->bSs7FisuFiltering = OTX_FALSE;
  envP->b56kbpsChannel    = OTX_FALSE;
  envP->bLoopHandsetsInternally = OTX_FALSE;
  envP->pszDongleType     = NULL;
  envP->nLineLength       = 10;
  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->nTs               = -1;
  envP->bHdlcCrcCheck     = OTX_TRUE;
  envP->bHdlcSharedFlags  = OTX_FALSE;
  envP->bHdlcBitInversion = OTX_FALSE;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'H') {   // Longhaul mode
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'X') {   // Timeslot
          envP->nTs = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'I') {   // Internal handset loop
          envP->bLoopHandsetsInternally = OTX_TRUE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
          } else if (*(argv[i] + 2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          } else if (*(argv[i] + 2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if (*(argv[i] + 2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'W') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCIE;
          } else if (*(argv[i] + 2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
          } else if (*(argv[i] + 2) == 'H') {
            envP->nBoardType  = OTX_DEVICE_HUGIN_2_PC104;
          } else if (*(argv[i] + 2) == '1') {
            envP->nBoardType  = OTX_DEVICE_THOR_1_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'C') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_EXPRESSCARD;
          } else if (*(argv[i] + 2) == 'R') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_EXPRESSCARD;
          } else if (*(argv[i] + 2) == 'D') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_CSI;
          } else if (*(argv[i] + 2) == 'K') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4M_CSI;
          } else if (*(argv[i] + 2) == 'X') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8_CSI;
          } else if (*(argv[i] + 2) == 'Q') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8M_CSI;
          } else if (*(argv[i] + 2) == 'B') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_2_CSI;
          } else if (*(argv[i] + 2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_2_PBX;
          } else if (*(argv[i] + 2) == 'I') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_PBX;
          } else if (*(argv[i] + 2) == 'J') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8_PBX;
          } else if (*(argv[i] + 2) == 'U') {
            envP->nBoardType  = OTX_DEVICE_THOR_8M_USB;
          } else if (*(argv[i] + 2) == 'F') {
            envP->nBoardType  = OTX_DEVICE_THOR_4M_USB;
          } else if (*(argv[i] + 2) == 'A') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_USB;
          } else if (*(argv[i] + 2) == 'K') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_USB;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'D') {   // Dongle Type (for Thor-2-PCMCIA)
          if ( *(argv[i]+2) == '1') {
            envP->pszDongleType     = "HMA-1021-1-1.0";     // Thor-2-PCMCIA Standard Dongle (HMA-1021-1-1.0)
          } else if ( *(argv[i]+2) == '2') {
            envP->pszDongleType     = "HMA-1074-1-1.0-P1";  // Thor-2-PCMCIA Monitor Dongle  (HMA-1074-1-1.0-P1)
          } else if ( *(argv[i]+2) == '3') {
            envP->pszDongleType     = "HMA-1076-2-1.0-P1";  // Thor-2-PCMCIA-Plus Generator Dongle (HMA-1076-2-1.0-P1)
          } else if ( *(argv[i]+2) == '4') {
            envP->pszDongleType     = "HMA-1079-2-1.0";     // Thor-2-PCMCIA-Plus Monitor Dongle switchable (HMA-1079-2-1.0)
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->eLaw = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->eLaw = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'S') {   // SS7 FISU filtering
          envP->bSs7FisuFiltering = OTX_TRUE;
        } else if (*(argv[i]+1) == '5') {   // Use 56kbps HDLC channel (vs 64kpbs)
          envP->b56kbpsChannel = OTX_TRUE;
        } else if (*(argv[i]+1) == '9') {   // Don't CRC check the HDLC frames. Deliver the CRC bytes as part of the HDLC payload
          envP->bHdlcCrcCheck = OTX_FALSE;
        } else if (*(argv[i]+1) == '1') {   // Use shared flags
          envP->bHdlcSharedFlags = OTX_TRUE;
        } else if (*(argv[i]+1) == '2') {   // HDLC bit inversion
          envP->bHdlcBitInversion = OTX_TRUE;
        } else if (*(argv[i]+1) == 'N') {   // Line length
          envP->nLineLength = (OTX_UINT16)strtoul(argv[i]+2, NULL, 0);
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }


}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|1|T|E|8|P|M|S|H|C|R|D|X|Q|B|G|I|J|U] : Board Type used for the demo, where\n");
  DemoLogf("                    2  = Use Thor-2-PCI Board\n");
  DemoLogf("                    1  = Use Thor-1-PCI-Plus Board\n");
  DemoLogf("                    T  = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                    E  = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                    8  = Use Thor-8-PCI Board\n");
  DemoLogf("                    P  = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                    W  = Use Thor-8-PCIe Board\n");
  DemoLogf("                    4  = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                    S  = Use Sleipnir Board\n");
  DemoLogf("                    H  = Use Hugin Board\n");
  DemoLogf("                    C  = Use Thor-2-ExpressCard Board\n");
  DemoLogf("                    R  = Use Thor-4-ExpressCard Board\n");
  DemoLogf("                    D  = Use Alvis-4-CSI Board\n");
  DemoLogf("                    K  = Use Alvis-4M-CSI Board\n");
  DemoLogf("                    X  = Use Alvis-8-CSI Board\n");
  DemoLogf("                    Q  = Use Alvis-8M-CSI Board\n");
  DemoLogf("                    B  = Use Alvis-2-CSI Board\n");
  DemoLogf("                    G  = Use Alvis-2-PBX Board\n");
  DemoLogf("                    I  = Use Alvis-4-PBX Board\n");
  DemoLogf("                    J  = Use Alvis-8-PBX Board\n");
  DemoLogf("                    U  = Use Thor-8M-USB Board\n");
  DemoLogf("                    F  = Use Thor-4M-USB Board\n");
  DemoLogf("                    A  = Use Thor-8-USB Board\n");
  DemoLogf("                    K  = Use Thor-2-USB Board\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /S             : SS7 FISU filtering active\n");
  DemoLogf(" /5             : Use HDLC 56kbps channel (64kbps is default).\n");
  DemoLogf(" /9             : Do not check HDLC CRC. Deliver CRC bytes as part of payload\n");
  DemoLogf(" /1             : Shared flags. Default setting requires at least 2 flags.\n");
  DemoLogf(" /2             : HDLC Bit inversion. Default is not to invert the bits.\n");
  DemoLogf(" /CA            : A-law\n");
  DemoLogf(" /CU            : u-law\n");
  DemoLogf(" /ME            : E1 Li mode\n");
  DemoLogf(" /MT            : T1 Li mode\n");
  DemoLogf(" /I             : Internal Handset loop (not via the T1/E1 span)\n");
  DemoLogf(" /D[1|2|3|4]    : Dongle type (only for Thor-2-PCMCIA)\n");
  DemoLogf("                    1 = Thor-2-PCMCIA Standard Dongle (HMA-1021-1-1.0)\n");
  DemoLogf("                    2 = Thor-2-PCMCIA Monitor Dongle  (HMA-1074-1-1.0-P1)\n");
  DemoLogf("                    3 = Thor-2-PCMCIA-Plus Generator Dongle (HMA-1076-2-1.0-P1)\n");
  DemoLogf("                    4 = Thor-2-PCMCIA-Plus Monitor Dongle switchable (HMA-1079-2-1.0)\n");
  DemoLogf(" /N<lineLength> : T1/E1 span line length\n");
  DemoLogf(" /X<ts>         : Timeslot for HDLC device\n");
  DemoLogf(" /H             : Longhaul mode (default is shorthaul mode)\n");
  DemoLogf(" /F[E|S|9]      : T1 Frame Format, where\n");
  DemoLogf("                    E = ESF (default)\n");
  DemoLogf("                    4 = F4\n");
  DemoLogf("                    S = SF [D3/D4] (F12)\n");
  DemoLogf("                    9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]        : T1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]      : E1 Frame Format, where\n");
  DemoLogf("                    D = Doubleframe\n");
  DemoLogf("                    C = CRC4 Multiframe (default)\n");
  DemoLogf("                    G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]        : E1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    H = HDB3 (default)\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");

  DemoLogf("\nPress Any Key to Exit...\n"); OsaConsoleGetCh();

}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(const char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF((char *)formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(const char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF((char *)formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {
    
    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
  
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentLi = 0;
  OTX_RESULT result;



  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Li, this line is then used for subsequent commands\n");
    DemoLogf(" t   Send a HDLC fram on the current Li\n");
    DemoLogf(" e/E Enable/Disable current Li\n");
    DemoLogf(" c   Set Clock Source\n");
    DemoLogf(" s   Display the SS7 HDLC receive status of the active Li\n");
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" C   Query Line Status of both Lis (extended)\n");
    DemoLogf(" a   Start/Stop Sending AIS (All 1's) on current Li\n");
    DemoLogf(" A   Start/Stop Sending Remote (Yellow) alarm on current Li\n");
    DemoLogf(" f   Force Resynchronization on both Lis\n");
    DemoLogf(" b   Reset Current HDLC Receiver\n");
    DemoLogf(" B   Reset Current E1T1 device\n");
    DemoLogf(" z   Set T1E1 Line Termination\n");
    DemoLogf(" Z   Toggle front end Amplifier ON or OFF for current Li\n");
    DemoLogf(" *   Set Line Build Out (T1 mode only)\n");
    DemoLogf(" &   Read receive equalizer status\n");
    DemoLogf(" J   Set Receive Input Threshold for current Li\n");
    DemoLogf(" p   Set Loopback mode\n");
    DemoLogf(" o   Buffer 10 HDLC frames to be transmitted, then Flush the HDLC Sender\n");
    DemoLogf(" g   Get current time\n");
    DemoLogf(" m/M Enable/Disable Monitor mode\n");
    DemoLogf(" L   Set HDLC Frame length to be transmitted\n");
    DemoLogf(" [   Measure T1/E1 span voltage for current Li\n");
    DemoLogf(" .   Set AUX LED state\n");
    DemoLogf(" '   Set SS7 Frame Filter\n");
    DemoLogf(" v   Toggle verbose HDLC reception (g_bHdlcVerbose=%s)\n", g_bHdlcVerbose ? "TRUE" : "FALSE");
    DemoLogf(" ;   Toggle HDLC payload check (g_bHdlcPayloadCheck=%s)\n", g_bHdlcPayloadCheck ? "TRUE" : "FALSE");
#if 0
    // screenful, allow user to read and continue...
    OsaConsolePrintF("Press any key...\r", NULL);
    OsaConsoleGetCh();
    OsaConsolePrintF("                \r", NULL);
#endif
    DemoLogf(" +   Set number of consecutive frame to be send from the HDLC Sender\n");
    DemoLogf(" S   Set time to sleep (in ms) between transmitted HDLC Frames (%d)\n", g_nSleepTimeBetweenHdlcFrames);
    DemoLogf(" !   Customize transmit pulse shape for current Li\n");
    DemoLogf(" @   Set idle pattern to 0x01 (to facilitate pulse shape testing)\n"); 
    DemoLogf(" k/K Turn Amp On/Off for current dongle (HMA-1079-2-1.0 dongle only)\n"); 
    DemoLogf(" F   Query line frequency of the current clock source\n");
    DemoLogf(" P   Report only significant Li status messages\n");
    DemoLogf(" x   Configure board for RX/RX mode\n");
    DemoLogf(" X   Configure board for TX/RX mode\n");
    //DemoLogf("OTX_RES_END = %d \n", OTX_RES_END);
    //DemoLogf("OTX_T1E1_RES_BAD_RX_FRAME_FORMAT = %d \n", OTX_T1E1_RES_BAD_RX_FRAME_FORMAT);
    break;

  case '\'':
    {
      char strBuf[12];

      DemoLogf("Li%d: Set the SS7 Filter Bitmask (0x00 = no filtering)\n", nCurrentLi);

      // DemoLogf("Filter masks in the SS7 mode
      DemoLogf("Bit0 (0x01) OTX_HDLC_SS7_FILTER_FISU - Filters ident FISUs (LI=0)\n");
      DemoLogf("Bit1 (0x02) OTX_HDLC_SS7_FILTER_SHORT_LSSU - Filters ident short LSSUs (LI=1)\n");
      DemoLogf("Bit2 (0x04) OTX_HDLC_SS7_FILTER_LONG_LSSU - Filters ident consec long LSSUs (LI=2)\n");
      DemoLogf("Bit3 (0x08) OTX_HDLC_SS7_FILTER_MSU - Filters all MSUs (LI>2)\n");
      DemoLogf("Bit4 (0x10) OTX_HDLC_SS7_FILTER_FISU_ALL - Filters all FISUs (LI=0)\n");
      DemoLogf("Bit5 (0x20) OTX_HDLC_SS7_FILTER_SHORT_LSSU_ALL - Filters all short LSSUs (LI=1)\n");
      DemoLogf("Bit6 (0x40) OTX_HDLC_SS7_FILTER_LONG_LSSU_ALL - Filters all long LSSUs (LI=2)\n");
      DemoLogf("Which BitMask? [0x%.2X] ", g_nSs7FilterBitMask);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        g_nSs7FilterBitMask = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxHdlcSS7SetFilter(g_hLogicDevs.hHdlcReceiver[nCurrentLi], g_nSs7FilterBitMask);

      DemoLogf("Wrote 0x%.2X as the SS7 Filter Bitmask\n", g_nSs7FilterBitMask);
      DemoLogf("Done.\n");
    }
    break;


  case 'y':
    {

      DemoLogf("Change Li mode of both Lis \n");

      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        DemoLogf("Switching to E1 mode \n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_E1;
      } else {
        DemoLogf("Switching to T1 mode \n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_T1;
      }
      
      DemoCloseDevices();

      DemoConfigureE1T1Devices();
      
      DemoEnablePhysicalDevices();

      DemoCreateLogicalDevices();
      
      DemoConnectDevices();
      
      DemoInitializeLogicalDevices();
      
      DemoEnableLogicalDevices();
      
      DemoLogf("Done.\n");
    }
    break;

  case 'g':
    {
      OTX_DATETIME nTimeStamp;
      OTX_DATETIME nTimeStampMs;
      OTX_DATETIME nTimeStampSec;
      struct Osa_tm *ptCurrentDayAndTime;
      Osa_time_t nCurrentTimeSecs;
      
      result = OtxDrvGetCurrentTime(&nTimeStamp); DemoCheckRc(result);
      
      //DemoLogf("%.8X%.8X (Current time)\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow);
      
      // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
      #if defined(__linux__)
        nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
      #elif defined(WIN32)
        nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
      #endif
      
      // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
      nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
      
      // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
      nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
      
      nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
      ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
      
      if (ptCurrentDayAndTime) {
        DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
          ptCurrentDayAndTime->tm_hour,
          ptCurrentDayAndTime->tm_min, 
          ptCurrentDayAndTime->tm_sec,
          nTimeStampMs.sUint32.nLow,
          asctime((struct tm *) ptCurrentDayAndTime)    // TODO: needs to be Osa function (OsaAscTime())
          );
      } else {
        DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
      }
      

    }
    break;

  case 'G':
    {
      OsaThreadStart(DemoTimeStampTest);
    }
    break;


  case 'p':
    {
      OtxT1E1LoopE eLoopType;
      OTX_INT32 nLi;

      DemoLogf("Set Loopback mode for both Line Interfaces \n");

      DemoLogf("0: All Line Loops off                     \n");
      DemoLogf("1: Line Loop                              \n");
      DemoLogf("2: Remote Loop with Jitter Attenuation On \n");
      DemoLogf("3: Remote Loop with Jitter Attenuation Off\n");
      DemoLogf("4: Local Loop                             \n");
      
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      case 1:
        eLoopType = OTX_T1E1_LINE_LOOP;
        break;
      case 2:
        eLoopType = OTX_T1E1_REMOTE_ON_LOOP_JITTER_ATTEN;
        break;
      case 3:
        eLoopType = OTX_T1E1_REMOTE_OFF_LOOP_JITTER_ATTEN;
        break;
      case 4:
        eLoopType = OTX_T1E1_LOCAL_LOOP;
        break;
      default:
        DemoLogf("Invalid entry. Turning Loop off (OTX_T1E1_LOOPS_OFF)\n");
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      }

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Setting Loopback mode (%d) on Li #%d \n", (int)eLoopType, nLi);
        result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], eLoopType); DemoCheckRc(result);
      }

      DemoLogf("Done.\n");
    }
    break;

  case ':':
    {
      OTX_INT32 i;

      /*
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        DemoLogf("Switching both Lis to OTX_T1E1_MODE_E1 mode\n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_E1;
      } else {
        DemoLogf("Switching both Lis to OTX_T1E1_MODE_T1 mode\n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_T1;
      }
      */

      for (i = 0; i < g_demoEnv.coLi; i++) {
        
        // Set termination for the T1/E1 Front End
        result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);
        
        // Configure Transceivers
        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
        } else {
          // E1 mode
          result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
        }
        result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);
        
      }

      DemoLogf("Done\n");
  
    }
    break;

  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(result);
        DemoLogf("Done.");
      }
    }
    break;

#if 0 // For debugging

  case 'x':
    {
      OTX_UINT8 sqTimeSlotMask[32];
      OTX_INT32 ts;
      
      DemoLogf("Connecting HS3 to ASM Hwy0, TS0\n");
      ts = 4;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(result);

      ts = 0;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 10, sqTimeSlotMask, 32); DemoCheckRc(result);
      
      DemoLogf("Done.\n");
    }
    break;
#endif


  case 'a':
    {
      static OTX_BOOL g_bSendingAis = OTX_FALSE;

      if (g_bSendingAis) {
        DemoLogf("Stop Sending AIS (All Ones) on Li #%d \n", nCurrentLi);
        result = OtxT1E1AlarmOff(g_hPhysDevs.hLi[nCurrentLi], OTX_T1E1_ALARM_AIS); DemoCheckRc(result);
      } else {
        DemoLogf("Send AIS (All Ones) on Li #%d \n", nCurrentLi);
        result = OtxT1E1AlarmOn(g_hPhysDevs.hLi[nCurrentLi], OTX_T1E1_ALARM_AIS); DemoCheckRc(result);
      }
      g_bSendingAis = !g_bSendingAis;
      DemoLogf("Done.\n");
    }
    break;

  case 'A':
    {
      static OTX_BOOL g_bSendingRemoteAlarm = OTX_FALSE;

      if (g_bSendingRemoteAlarm) {
        DemoLogf("Stop Sending Remote (Yellow) Alarm on Li #%d \n", nCurrentLi);
        result = OtxT1E1AlarmOff(g_hPhysDevs.hLi[nCurrentLi], OTX_T1E1_ALARM_RAI); DemoCheckRc(result);
      } else {
        DemoLogf("Send Remote (Yellow) Alarm on Li #%d \n", nCurrentLi);
        result = OtxT1E1AlarmOn(g_hPhysDevs.hLi[nCurrentLi], OTX_T1E1_ALARM_RAI); DemoCheckRc(result);
      }
      g_bSendingRemoteAlarm = !g_bSendingRemoteAlarm;
      DemoLogf("Done.\n");
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case 'C':
    {
      OtxT1E1Layer1StatusExS sLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Li #%d line is ", nLi);
        result = OtxT1E1GetLayer1StatusEx(g_hPhysDevs.hLi[nLi], &sLayer1Status); DemoCheckRc(result);

        DemoLogf("Alarms: ");
        if (sLayer1Status.bRedAlarm) {
          DemoLogf("R");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bBlueAlarm) {
          DemoLogf("B");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bYellowAlarm) {
          DemoLogf("Y");
        } else {
          DemoLogf("-");
        }

        if (sLayer1Status.bFrameError) {
          DemoLogf(" FrmErr: YES");
        } else {
          DemoLogf(" FrmErr: NO ");
        }

        if (sLayer1Status.bMultiFrameError) {
          DemoLogf(" MFrmErr: YES");
        } else {
          DemoLogf(" MFrmErr: NO ");
        }
        DemoLogf("\n");
      }
    }
    break;

  case 'b':
    {
      DemoLogf("Reseting HdlcReceiver[%d]...\n",nCurrentLi);
      result = OtxDrvReset(g_hLogicDevs.hHdlcReceiver[nCurrentLi]); DemoCheckRc(result);
      DemoLogf("Done\n");
    }
    break;

  case 'B':
    {
      DemoLogf("Reseting Li #%d \n", nCurrentLi);
      result = OtxDrvReset(g_hPhysDevs.hLi[nCurrentLi]); DemoCheckRc(result);
      DemoLogf("Done\n");
    }
    break;

  case 's':
    {
      OTX_UINT32  nFisuCount;   // Number of FISUs received
      OTX_UINT32  nLssuCount;   // Number of LSSUs received
      OTX_UINT32  nMsuCount;    // Number of MSUs received
      OtxHdlcSS7FisuS fisu;     // The last received FISU on the pipe.

      result = OtxHdlcSS7GetReceiveStatus(g_hLogicDevs.hHdlcReceiver[nCurrentLi], &nFisuCount, &nLssuCount, &nMsuCount, &fisu); DemoCheckRc(result);

      if (OTX_SUCCESS(result)) {
		DemoLogf("HdlcReceiver[%d]: SS7 receive statistics\n",nCurrentLi);
        DemoLogf("FISUs received: %lu\n", nFisuCount);
        DemoLogf("LSSUs received: %lu\n", nLssuCount);
        DemoLogf("MSUs  received: %lu\n", nMsuCount);
        DemoLogf("Last FISU received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n",
               fisu.bsn, fisu.bib, fisu.fsn, fisu.fib, fisu.li);
      }

    }
    break;

#if 0
  case 'm':
    {
      DemoLogf("Disconnecting HS2 Pipes (Spkr and Mic)\n");
      result = OtxDrvDisconnectPipe(g_hPipes.hHs2MicPipe, OTX_PIPE_INPUT); DemoCheckRc(result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs2MicPipe, OTX_PIPE_OUTPUT); DemoCheckRc(result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs2SpkrPipe, OTX_PIPE_INPUT); DemoCheckRc(result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs2SpkrPipe, OTX_PIPE_OUTPUT); DemoCheckRc(result);
      
      DemoLogf("Done.\n");
    }
    break;

  case 'M':
    {
      DemoLogf("Disconnecting HS3 Pipes (Spkr and Mic)\n");
      result = OtxDrvDisconnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_INPUT); DemoCheckRc(result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_OUTPUT); DemoCheckRc(result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs3SpkrPipe, OTX_PIPE_INPUT); DemoCheckRc(result);
      result = OtxDrvDisconnectPipe(g_hPipes.hHs3SpkrPipe, OTX_PIPE_OUTPUT); DemoCheckRc(result);
      
      DemoLogf("Done.\n");
    }
    break;

  case 'V':
    {
      OTX_UINT8 sqTimeSlotMask[32];
      OTX_INT32 ts;
      
      DemoLogf("Connecting HS3 to H100 Hwy0 ");
      ts = 4;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(result);

      if (g_demoEnv.boardNo == 0) {
        ts = 0;
      } else {
        ts = 1;
      }
      DemoLogf("Mic to TS%d, ", ts);
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3MicPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 16, sqTimeSlotMask, 32); DemoCheckRc(result);
      
      if (g_demoEnv.boardNo == 0) {
        ts = 1;
      } else {
        ts = 0;
      }
      DemoLogf("Spkr to TS%d \n", ts);
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3SpkrPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 16, sqTimeSlotMask, 32); DemoCheckRc(result);

      ts = 4;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs3SpkrPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;
#endif

#if 0
  case 'v':
    {
      OTX_UINT8 sqTimeSlotMask[32];
      OTX_INT32 ts;
      
      DemoLogf("Connecting HS2 to H100 Hwy0 ");
      ts = 3;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs2MicPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(result);

      if (g_demoEnv.boardNo == 0) {
        ts = 0;
      } else {
        ts = 1;
      }
      DemoLogf("Mic to TS%d, ", ts);
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs2MicPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 16, sqTimeSlotMask, 32); DemoCheckRc(result);
      
      if (g_demoEnv.boardNo == 0) {
        ts = 1;
      } else {
        ts = 0;
      }
      DemoLogf("Spkr to TS%d \n", ts);
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs2SpkrPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hTss0, 16, sqTimeSlotMask, 32); DemoCheckRc(result);

      ts = 3;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hHs2SpkrPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hTss0, 14, sqTimeSlotMask, 32); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;
#endif

  case 'm':
    DemoLogf("Enabling Monitor (-20dB) mode for Li #%d \n", nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[nCurrentLi], OTX_TRUE); DemoCheckRc(result);
    DemoLogf("Done.\n");
    break;

  case 'M':
    DemoLogf("Disabling Monitor (-20dB) mode for Li #%d \n", nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[nCurrentLi], OTX_FALSE); DemoCheckRc(result);
    DemoLogf("Done.\n");
    break;

  case 'e':
    {
      OTX_RESULT result;
      OTX_INT32  nQuadE1T1 = nCurrentLi/4;
      
      // If board type is Thor-8-PCI then enable Quad E1T1 device
      switch (g_demoEnv.nBoardType) {
      case OTX_DEVICE_THOR_8_PCI:
      case OTX_DEVICE_THOR_8_PCI_PLUS:
      case OTX_DEVICE_THOR_8_PCIE:
      case OTX_DEVICE_THOR_4_PCI_PLUS:
      case OTX_DEVICE_DMP_ALVIS_4_PBX:
      case OTX_DEVICE_DMP_ALVIS_2_PBX:
      case OTX_DEVICE_DMP_ALVIS_2_CSI:
      case OTX_DEVICE_DMP_ALVIS_4_CSI:
      case OTX_DEVICE_DMP_ALVIS_4M_CSI:
      case OTX_DEVICE_THOR_4_EXPRESSCARD:
      case OTX_DEVICE_THOR_2_EXPRESSCARD:
        // Boards with Lis hosted by a QuadE1T1 device can only be enabled at the QuadT1E1 device level
        DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
        result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        break;
      case OTX_DEVICE_DMP_ALVIS_8_PBX:
      case OTX_DEVICE_DMP_ALVIS_8_CSI:
      case OTX_DEVICE_DMP_ALVIS_8M_CSI:
      case OTX_DEVICE_THOR_8M_USB:
      case OTX_DEVICE_THOR_4M_USB:
      case OTX_DEVICE_THOR_8_USB:
      case OTX_DEVICE_THOR_2_USB:
        // Boards with Lis hosted by a QuadE1T1 device can only be enabled at the QuadT1E1 device level
        DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", 0);
        result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[0], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        break;
      default:
        // These boards have Lis with invididual interrupt lines, so they can be disabled individually
        DemoLogf("Enable Li #%d \n", nCurrentLi);
        result = OtxDrvEnable(g_hPhysDevs.hLi[nCurrentLi], DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(result);
        break;
      }
      DemoLogf("Done.\n");
    }
    break;

  case 'E':
    {
      OTX_RESULT result;
      OTX_INT32  nQuadE1T1 = nCurrentLi/4;
      
      switch (g_demoEnv.nBoardType) {
      case OTX_DEVICE_THOR_8_PCI:
      case OTX_DEVICE_THOR_8_PCI_PLUS:
      case OTX_DEVICE_THOR_8_PCIE:
      case OTX_DEVICE_THOR_4_PCI_PLUS:
      case OTX_DEVICE_DMP_ALVIS_4_PBX:
      case OTX_DEVICE_DMP_ALVIS_2_PBX:
      case OTX_DEVICE_DMP_ALVIS_2_CSI:
      case OTX_DEVICE_DMP_ALVIS_4_CSI:
      case OTX_DEVICE_DMP_ALVIS_4M_CSI:
      case OTX_DEVICE_THOR_4_EXPRESSCARD:
      case OTX_DEVICE_THOR_2_EXPRESSCARD:
        // Boards with Lis hosted by a QuadE1T1 device can only be disabled at the QuadT1E1 device level
        DemoLogf("Disable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
        result = OtxDrvDisable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        break;
      case OTX_DEVICE_DMP_ALVIS_8_PBX:
      case OTX_DEVICE_DMP_ALVIS_8_CSI:
      case OTX_DEVICE_DMP_ALVIS_8M_CSI:
      case OTX_DEVICE_THOR_8M_USB:
      case OTX_DEVICE_THOR_4M_USB:
      case OTX_DEVICE_THOR_8_USB:
      case OTX_DEVICE_THOR_2_USB:
        // Boards with Lis hosted by a QuadE1T1 device can only be disabled at the QuadT1E1 device level
        DemoLogf("Disable OTX_DEVICE_QUAD_T1E1 #%d device\n", 0);
        result = OtxDrvDisable(g_hPhysDevs.hQuadE1T1[0], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        break;
      default:
        // These boards have Lis with invididual interrupt lines, so they can be disabled individually
        DemoLogf("Disable Li #%d \n", nCurrentLi);
        result = OtxDrvDisable(g_hPhysDevs.hLi[nCurrentLi], DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(result);
        break;
      }
      DemoLogf("Done.\n");
    }
    break;

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;
      int c;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      DemoLogf("4 = OTX_T1E1_TERM_300_OHM (for highly attenuated signals)\n");
      c = OsaConsoleGetCh();
      DemoLogf("'%c' pressed\n", c);
      switch (c - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      case 4:
        eTermMode = OTX_T1E1_TERM_300_OHM;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

  case '.':
    {
      OTX_UINT32 nLedVal;
      int c;
      
      DemoLogf("Set Aux LED state:\n");
      DemoLogf("0 = Off\n");
      DemoLogf("1 = Constant Red\n");
      DemoLogf("2 = Constant Green\n");
      c = OsaConsoleGetCh();
      DemoLogf("'%c' pressed\n", c);
      
      nLedVal = c - '0';
        
      DemoLogf("Setting AUX LED to state %d\n", nLedVal);
      
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_LED_STATE, nLedVal); DemoCheckRc(result);
      
      DemoLogf("Done.\n");
      
    }
    break;

  case 'Z':
    {
      OTX_BOOL bAmpOn;
      
      result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_AMPLIFIER, &bAmpOn); DemoCheckRc(result);
      
      if (bAmpOn) {
        bAmpOn=OTX_FALSE;
      } else {
        bAmpOn=OTX_TRUE;
      }
      
      DemoLogf("Turning front end Amplifier %s for Li%d \n", bAmpOn ? "ON": "OFF", nCurrentLi);
      
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_AMPLIFIER, bAmpOn); DemoCheckRc(result);
      
      DemoLogf("Done.\n");
      
    }
    break;

  case ')':
    {
      OTX_BOOL bAmpOn = OTX_FALSE;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Turning front end Amplifier %s for Li%d \n", bAmpOn ? "ON": "OFF", nLi);
        result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[nLi], OTX_ATTR_T1E1_RX_AMPLIFIER, bAmpOn); DemoCheckRc(result);
      }
      
      DemoLogf("Done.\n");
      
    }
    break;

  case '(':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Receiver and Transmitter Reset for for Li%d \n", nLi);
        result = OtxDrvDbgDump(g_hPhysDevs.hLi[nLi], 129); DemoCheckRc(result);
      }
      
      DemoLogf("Done.\n");
      
    }
    break;

  case 't':
    {
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;
      OTX_INT32 timeoutCount = 0;
      clock_t startTime, stopTime;

      if (g_cbHdlcFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      for (i = 0; i < g_cbHdlcFrame; i++) {
        g_sqHdlcFrame[i] = (OTX_UINT8) i;
      }

      DemoPrintCurrentTime();
      DemoPrintf(" Sending %d HDLC frame(s) from Li #%d (len=%d) ...", g_coHdlcSenderFrames, nCurrentLi, g_cbHdlcFrame);

      // Send HDLC Frame(s)

      startTime = clock();

      for (i = 0; (i < g_coHdlcSenderFrames) && (timeoutCount < 25); i++) {
        // If handle has been setup send the data
        if(g_hLogicDevs.hHdlcSender[nCurrentLi] != OTX_INVALID_HANDLE_VALUE) {
          result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nCurrentLi], g_sqHdlcFrame, g_cbHdlcFrame, DEMO_TASK_REF, 200); 
          if ((OTX_RESULT_CODE(result) == OTX_E_BUSY)) {
            // Driver IO Request buffer is full. We need to retransmit this frame.
            i--;
            busyCount++;
            timeoutCount++; // If it is busy 25 times in a row exit test
            OsaThreadSleep(10); // IO buffer is full, give it time to proceess
          } else {
            timeoutCount = 0; // reset the timeout counter
            DemoCheckRc(result);
            OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
          }
        }
      }
      if(timeoutCount >=25) {
        DemoPrintf("Test timed out waiting on IO request\n");
      }
      stopTime = clock();

      DemoPrintf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }
    break;


  case 'o':
    {
      OTX_UINT8 arMsg[256];
      OTX_INT32 i;
      OTX_INT32 nMsgLen = 256;
      OTX_HANDLE hHandle;

      if(g_hLogicDevs.hHdlcSender[nCurrentLi] != OTX_INVALID_HANDLE_VALUE) {

        hHandle = g_hLogicDevs.hHdlcSender[nCurrentLi];

        for (i = 0; i < nMsgLen; i++) {
          arMsg[i] = (OTX_UINT8) i;
        }

        DemoLogf("Buffer 10 HDLC frames to be transmitted, then Flush the HDLC Sender on Li #%d ...", nCurrentLi);

        // Send the HDLC frame
        for (i = 0; i < 10; i++) {
          result = OtxHdlcSendData(hHandle, arMsg, nMsgLen, DEMO_TASK_REF, 200); 
          DemoCheckRc(result);
        }

        DemoLogf("Sleeping...\n");
        OsaThreadSleep(50);

        result = OtxHdlcSenderFlush(hHandle); DemoCheckRc(result);
      }
      else {
        // Thor-4-ExpressCard support transmit on line 1 and 2, line 3 and 4
        // are receive only
        DemoLogf("HDLC Sender handle not initialized\n");
      }

      DemoLogf("Done.\n");
    }
    break;


  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI && nLi > 2) {
        DemoLogf("Invalid Li# for Thor-2-PCI board (%d)\n", nLi);
        break;
      }
      
      nCurrentLi = nLi;
      DemoLogf("LI#%d Selected ....\n", nCurrentLi);

    }
    break;


  case 'F':
    {
      // For Thor-8-PCI-Plus
      OTX_UINT32 nFrequencyCount;

      DemoLogf("Querying frequency of current clock source ...\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_FREQUENCY_COUNTER, &nFrequencyCount); DemoCheckRc(result);

      if (OTX_SUCCESS(result)) {
        OTX_UINT32 nFrequency;

        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          nFrequency = (nFrequencyCount+1)*193/2048;
        } else {
          // E1 mode
          nFrequency = (nFrequencyCount+1)/8;
        }
        DemoLogf("Line frequency is %d Hz (nFrequencyCount=%d)\n", nFrequency, nFrequencyCount);
      }

      DemoLogf("Done.\n");
    }
    break;

  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'u':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in QuadE1T1%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hQuadE1T1[nCurrentLi], nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("QuadE1T1%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'R':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hPhysDevs.hThorController0, nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("Brd Address 0x%X = %.4X\n", nAddress, (OTX_UINT16)readVal);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'j':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in QuadE1T1%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hQuadE1T1[nCurrentLi], nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote %.2X to QuadE1T1%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'W':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hPhysDevs.hThorController0, nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
    }
    break;

  case 'L':
    {
      char strBuf[12];
      DemoLogf("Enter HDLC Frame Length to be transmitted (%d) ", g_cbHdlcFrame);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_cbHdlcFrame = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("HDLC Frame Length is now set to %d\n", g_cbHdlcFrame);
      } else {
        DemoLogf("HDLC Frame Length remains unchanged\n");
      }
    }
    break;

  case 'S':
    {
      char strBuf[12];
      DemoLogf("Enter Time to Sleep (in ms) between transmitted HDLC Frames (%d) ", g_nSleepTimeBetweenHdlcFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_nSleepTimeBetweenHdlcFrames = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenHdlcFrames);
      } else {
        DemoLogf("Sleep time remains unchanged\n");
      }
    }
    break;

  case '+':
    {
      char strBuf[12];

      DemoLogf("Set the number of consecutive frame to sent when 't' are pressed (%d)\n", g_coHdlcSenderFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_coHdlcSenderFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
        DemoLogf("Consecutive frame count is now %d\n", g_coHdlcSenderFrames);
      } else {
        DemoLogf("Consecutive frame count remains unchanged\n");
      }
    }
    break;



  case 'v':
    {
      OTX_INT32 i;
      g_bHdlcVerbose = !g_bHdlcVerbose;
      DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE": "FALSE");
      for (i = 0; i < 32; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("g_coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_coHdlcRxFrames[i]);
          g_coHdlcRxFrames[i] = 0;
          g_coHdlcBadRxFrames[i] = 0;
        }
      }
    }
    break;

  case ';':
    {
      if (g_bHdlcVerbose) {
        DemoLogf("g_bHdlcVerbose must be FALSE to change g_bHdlcPayloadCheck (see the 'v' command)\n");
      } else {
        g_bHdlcPayloadCheck = !g_bHdlcPayloadCheck;
        DemoLogf("g_bHdlcPayloadCheck is now %s.\n", g_bHdlcPayloadCheck ? "TRUE": "FALSE");
      }
    }
    break;

  case 'V':
    {
      OTX_INT32 i;
      for (i = 0; i < 32; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("Tot[%.2d]=%.6d (Bad[%.2d]=%.6d)  ", i, g_coHdlcRxFrames[i], i, g_coHdlcBadRxFrames[i]);
          if (i && i % 2) {
            DemoLogf("\n");
          }
        }
      }
      DemoLogf("\n");
    }
    break;

  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;
      OTX_INT32 c;

      DemoLogf("Set Clock Source:\n");
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCIE) {
        DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span (QuadE1T1#0)\n");
        DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span (QuadE1T1#0)\n");
        DemoLogf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 receive span (QuadE1T1#0)\n");
        DemoLogf("3 = OTX_CLOCK_SOURCE_LOCAL_3  - LIs synchronized to Li3 receive span (QuadE1T1#0)\n");
        DemoLogf("4 = OTX_CLOCK_SOURCE_LOCAL_4  - LIs synchronized to Li4 receive span (QuadE1T1#1)\n");
        DemoLogf("5 = OTX_CLOCK_SOURCE_LOCAL_5  - LIs synchronized to Li5 receive span (QuadE1T1#1)\n");
        DemoLogf("6 = OTX_CLOCK_SOURCE_LOCAL_6  - LIs synchronized to Li6 receive span (QuadE1T1#1)\n");
        DemoLogf("7 = OTX_CLOCK_SOURCE_LOCAL_7  - LIs synchronized to Li7 receive span (QuadE1T1#1)\n");
        DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
          case 2:
          case 3:
          case 4:
          case 5:
          case 6:
          case 7:
            eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') ); 
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_8_CSI || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_8M_CSI || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_8_PBX || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8M_USB || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_USB) {
        DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span\n");
        DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span\n");
        DemoLogf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 receive span\n");
        DemoLogf("3 = OTX_CLOCK_SOURCE_LOCAL_3  - LIs synchronized to Li3 receive span\n");
        DemoLogf("4 = OTX_CLOCK_SOURCE_LOCAL_4  - LIs synchronized to Li4 receive span\n");
        DemoLogf("5 = OTX_CLOCK_SOURCE_LOCAL_5  - LIs synchronized to Li5 receive span\n");
        DemoLogf("6 = OTX_CLOCK_SOURCE_LOCAL_6  - LIs synchronized to Li6 receive span\n");
        DemoLogf("7 = OTX_CLOCK_SOURCE_LOCAL_7  - LIs synchronized to Li7 receive span\n");
        DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
          case 2:
          case 3:
          case 4:
          case 5:
          case 6:
          case 7:
            eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') ); 
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4M_USB) {
        DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 physical receive span\n");
        DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_5  - LIs synchronized to Li5 physical receive span\n");
        DemoLogf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li7 physical receive span\n");
        DemoLogf("3 = OTX_CLOCK_SOURCE_LOCAL_6  - LIs synchronized to Li6 physical receive span\n");
        DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0: eClkSource = OTX_CLOCK_SOURCE_LOCAL_1; break;
          case 1: eClkSource = OTX_CLOCK_SOURCE_LOCAL_5; break;
          case 2: eClkSource = OTX_CLOCK_SOURCE_LOCAL_2; break;
          case 3: eClkSource = OTX_CLOCK_SOURCE_LOCAL_6; break;
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_USB) {
        DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 physical receive span\n");
        DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 physical receive span\n");
        DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0: eClkSource = OTX_CLOCK_SOURCE_LOCAL_1; break;
          case 1: eClkSource = OTX_CLOCK_SOURCE_LOCAL_2; break;
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_4_CSI || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_4M_CSI) {
        DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span\n");
        DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span\n");
        DemoLogf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 receive span\n");
        DemoLogf("3 = OTX_CLOCK_SOURCE_LOCAL_3  - LIs synchronized to Li3 receive span\n");
        DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
          case 2:
          case 3:
            eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') ); 
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else {
        DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span\n");
        DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span\n");
        DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
            eClkSource = (OtxBrdClockSourceE) (OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') );
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }

      }

      result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;
  

  case 'O':
    {
      char szBuf[12];
      OTX_INT32 nCmd;
      
      DemoLogf("Driver Debug Command. Which Cmd?\n");
      DemoLogf("0: Dump QuadFalc register\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      DemoLogf("Issuing Driver Debug Command %d to QuadE1T1%d \n", nCmd, nCurrentLi);
      OtxDrvDbgDump(g_hPhysDevs.hQuadE1T1[nCurrentLi], nCmd);
    }
    break;

  case '9':
    {
      char szBuf[64];
      OTX_INT32 nCmd;
      
      DemoLogf("Board Debug Command. Which Cmd?\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      DemoLogf("Issuing Driver Debug Command %d to hThorController0\n", nCmd);
      OtxDrvDbgDump(g_hPhysDevs.hThorController0, nCmd);
    }
    break;

  case '`':
    {
      char szBuf[12];
      OTX_UINT32 nCmd;
      OTX_UINT32 nVal;
      OTX_UINT32 nValCmd;
      
      DemoLogf("TSS Debug Command. Which Cmd?\n");
      DemoLogf("5: Set GP Register\n");
      DemoLogf("6: UnSet GP Register\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_UINT32) strtol(szBuf, NULL, 0);

      DemoLogf("Which value (bitpattern)?\n");
      OsaConsoleGetStr(szBuf);
      nVal = (OTX_UINT32) strtol(szBuf, NULL, 0);

      nValCmd = (nVal << 16) | nCmd;

      DemoLogf("Issuing Driver Debug Command %d to TSS\n", nCmd);
      OtxDrvDbgDump(g_hPhysDevs.hTss0, nValCmd);
    }
    break;




#if 1 // Just for debugging
  case '{':
    {
      char strBuf[12];
      static OTX_UINT8 nTssRegister = 0;
      OTX_UINT32 nReadVal;

      DemoLogf("Read from TSS control register. Which control register? [0x%.2X] ", nTssRegister);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTssRegister = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x01, nTssRegister); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x02, 0x00); DemoCheckRc(result);

      result = OtxDrvReadReg8(g_hPhysDevs.hTss0,  0x03, &nReadVal);  DemoCheckRc(result);

      DemoLogf("Read 0x%.2X from TSS Control Register 0x%.2X\n", (OTX_UINT8)nReadVal, nTssRegister);
      DemoLogf("Done.\n");
    }
    break;


  case '}':
    {
      char strBuf[12];
      static OTX_UINT8 nTssRegister = 0;
      static OTX_UINT8 nWriteVal = 0;

      DemoLogf("Write to TSS control register. Which control register? [0x%.2X] ", nTssRegister);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTssRegister = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      DemoLogf("Which value? [0x%.2X] ", nWriteVal);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nWriteVal = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x01, nTssRegister); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0, 0x02, 0x00); DemoCheckRc(result);

      result = OtxDrvWriteReg8(g_hPhysDevs.hTss0,  0x03, nWriteVal);  DemoCheckRc(result);

      DemoLogf("Wrote 0x%.2X to TSS Control Register 0x%.2X\n", nWriteVal, nTssRegister);
      DemoLogf("Done.\n");
    }
    break;

  case 'i':
    {
      DemoLogf("Enable TSS... \n");
      result = OtxDrvEnable(g_hPhysDevs.hTss0, OTX_TASK_SYNC,0); DemoCheckRc(result);
      DemoLogf("Done.\n");
    }
    break;

  case 'I':
    {
      DemoLogf("Disable TSS... \n");
      result = OtxDrvDisable(g_hPhysDevs.hTss0, OTX_TASK_SYNC,0); DemoCheckRc(result);
      DemoLogf("Done.\n");
    }
    break;

  case '_':
    {
      DemoLogf("Reseting TSS... \n");
      result = OtxDrvReset(g_hPhysDevs.hTss0); DemoCheckRc(result);
      DemoLogf("Done.\n");
    }
    break;
#endif

#if 1  // Just for debugging
  case '*':
    {
      OTX_UINT32 nReadVal;
      OTX_INT32 nLboVal;

      DemoLogf("Set LBO of Li #%d (T1 long haul mode only)\n", nCurrentLi);
      DemoLogf("0 = 0 dB\n");
      DemoLogf("1 = -7.5  dB\n");
      DemoLogf("2 = -15   dB\n");
      DemoLogf("3 = -22.5 dB\n");

      nLboVal = OsaConsoleGetCh() - '0';
      switch (nLboVal) {
      case 0:  // 0 dB
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x26, 0x01); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x25, 0x2B); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x24, 0x7D); DemoCheckRc(result);
        break;
      case 1: // -7.5dB
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x26, 0x20); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x25, 0x02); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x24, 0x11); DemoCheckRc(result);
        break;
      case 2: // -15dB
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x26, 0x20); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x25, 0x01); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x24, 0x8E); DemoCheckRc(result);
        break;
      case 3: // -22.5dB
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x26, 0x20); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x25, 0x01); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x24, 0x09); DemoCheckRc(result);
        break;
      default:
        DemoLogf("Invalid entry.\n");
        break;
      }

      switch (nLboVal) {
      case 0:  // 0 dB
      case 1: // -7.5dB
      case 2: // -15dB
      case 3: // -22.5dB
        result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], 0x38, &nReadVal); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x38, (nReadVal & 0x3F) | (nLboVal << 6)); DemoCheckRc(result);
        break;
      default:
        break;
      }

      DemoLogf("Done.\n");
    }
    break;

  case '!':
    {
      OTX_UINT32 nPulseShapeLevel0;
      OTX_UINT32 nPulseShapeLevel1;
      OTX_UINT32 nPulseShapeLevel2;
      OTX_UINT32 nPulseShapeLevel3;
      char strBuf[12];

      DemoLogf("Customizing Transmit Pulse Shape Level of Li #%d \n", nCurrentLi);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_0, &nPulseShapeLevel0); DemoCheckRc(result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_1, &nPulseShapeLevel1); DemoCheckRc(result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_2, &nPulseShapeLevel2); DemoCheckRc(result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_3, &nPulseShapeLevel3); DemoCheckRc(result);
      
      DemoLogf("nPulseShapeLevel0 [%d]. Enter new value (0-31): ", nPulseShapeLevel0);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nPulseShapeLevel0 = strtoul(strBuf, NULL, 0);
      }
      
      DemoLogf("nPulseShapeLevel1 [%d]. Enter new value (0-31): ", nPulseShapeLevel1);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nPulseShapeLevel1 = strtoul(strBuf, NULL, 0);
      }
      
      DemoLogf("nPulseShapeLevel2 [%d]. Enter new value (0-31): ", nPulseShapeLevel2);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nPulseShapeLevel2 = strtoul(strBuf, NULL, 0);
      }
      
      DemoLogf("nPulseShapeLevel3 [%d]. Enter new value (0-31): ", nPulseShapeLevel3);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nPulseShapeLevel3 = strtoul(strBuf, NULL, 0);
      }
      
      DemoLogf("Writing Sub Pulse Mask %d,%d,%d,%d \n", nPulseShapeLevel0, nPulseShapeLevel1, nPulseShapeLevel2, nPulseShapeLevel3);

      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_0, nPulseShapeLevel0); DemoCheckRc(result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_1, nPulseShapeLevel1); DemoCheckRc(result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_2, nPulseShapeLevel2); DemoCheckRc(result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_PULSE_MASK_SUB_PULSE_3, nPulseShapeLevel3); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

  case '&':
    {
      OTX_UINT32 readVal;
      OTX_INT32 nLi;
      const char *pszEqStatus;
      
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        result = OtxDrvReadReg8(g_hPhysDevs.hLi[nLi], 115, &readVal); DemoCheckRc(result);
        switch ((readVal & 0xC0) >> 6) {
        case 0x00:
          pszEqStatus = "not valid, still adapting";
          break;
        case 0x01:
          pszEqStatus = "valid";
          break;
        case 0x02:
          pszEqStatus = "not valid";
          break;
        case 0x03:
          pszEqStatus = "valid but high noise floor";
          break;
        default:
          pszEqStatus = "unknown";
          break;
        }
        DemoLogf("Receive Equalizer of Li#%d: status=%s attenuation=%2.1fdB (0x%.2X)\n", nLi, pszEqStatus, (readVal & 0x1F)*1.4, readVal );
      }
      
    }
    break; 
    
  case 'J':
    {
      char strBuf[12];
      OTX_UINT32 nThresholdInMilliVolts;

      DemoLogf("Set Receive Input Threshold for Li%d \n", nCurrentLi);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(result);

      DemoLogf("Threshold (%d mV): ", nThresholdInMilliVolts);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nThresholdInMilliVolts = strtoul(strBuf, NULL, 0);

        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, nThresholdInMilliVolts); DemoCheckRc(result);

        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(result);
        DemoLogf("Receive Input Threshold is changed to %d mV.\n", nThresholdInMilliVolts);
      } else {
        DemoLogf("The Receive Input Threshold remains unchanged.\n");
      }

    }
    break;

  case 'k':
    {
      OtxBoardDeviceAttributesE nAttrib;

      DemoLogf("Turn Amp On for Li%d (HMA-1079-2-1.0 dongle only)\n", nCurrentLi);

      if (nCurrentLi == 0) {
        nAttrib = OTX_ATTR_BRD_PCMCIA_AMP_DONGLE_0;
      } else if (nCurrentLi == 1) {
        nAttrib = OTX_ATTR_BRD_PCMCIA_AMP_DONGLE_1;
      } else {
        DemoLogf("Invalid Li (%d) for this action\n", nCurrentLi);
        break;
      }

      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, nAttrib, OTX_TRUE); DemoCheckRc(result);

      DemoLogf("Done\n");
    }
    break;
  case 'K':
    {
      OtxBoardDeviceAttributesE nAttrib;

      DemoLogf("Turn Amp Off for Li%d (HMA-1079-2-1.0 dongle only)\n", nCurrentLi);

      if (nCurrentLi == 0) {
        nAttrib = OTX_ATTR_BRD_PCMCIA_AMP_DONGLE_0;
      } else if (nCurrentLi == 1) {
        nAttrib = OTX_ATTR_BRD_PCMCIA_AMP_DONGLE_1;
      } else {
        DemoLogf("Invalid Li (%d) for this action\n", nCurrentLi);
        break;
      }

      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, nAttrib, OTX_FALSE); DemoCheckRc(result);

      DemoLogf("Done\n");
    }
    break;



  case '@':
    {

      DemoLogf("Set idle pattern to 0x01 (to facilitate pulse shape testing) for Li #%d \n", nCurrentLi);

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x29, 0x01); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x30, 0xFF); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x31, 0xFF); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x32, 0xFF); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], 0x33, 0xFF); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

  case 'd':   // dump falc registers
    DemoDumpFalcRegisters();
    break;

  case '-':
    {
      OTX_INT32 i = 0;
      OTX_UINT32 nAddress, readVal;
      OTX_UINT32 cmdResult;
      FILE *fout2;
      OTX_INT32 nCmd = 0x100B;
      
      DemoLogf("Issuing Read FalcRegVal Debug Command 0x%X to hThorController0\n", nCmd);
      
      fout2 = fopen("./FalcRegFileFromDriver", "wt");
      if (!fout2) {
        DemoLogf("\nCould not open FalcRegFileFromDriver file\n");
        break;
      }
      
      while ((cmdResult = OtxDrvDbgDump(g_hPhysDevs.hThorController0, nCmd)) != 0xDEADBEEF) {
        if (i++ >= 4000) {
          DemoLogf("Too many register values. Aborting %d\n", i);
          break;
        }

        nAddress = (cmdResult >> 16) & 0xFFFF;
        readVal = cmdResult & 0xFF;

        fprintf(fout2, "0x%.4X,0x%.2X\n", nAddress, readVal); 
      }
      
      fclose(fout2);
      DemoLogf("Done. i=%d\n", i);
    }
    break;

  case 'x':
    DemoSetRxRxMode();
    break;
  case 'X':
    DemoSetTxRxMode();
    break;


#endif


  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;


    // E1/T1 span voltmeter
  case '[':
    {
      OTX_UINT32 nMilliVolts;
            
      DemoLogf("Measure T1/E1 Span Voltage for Li%d: \n", nCurrentLi);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_LINE_VOLTAGE, &nMilliVolts); DemoCheckRc(result);

      DemoLogf("  Voltage = %2.3f V   \n", (double)nMilliVolts/1000);
      
    }
    break;

  case 'P':
    {
      OTX_BOOL bReportOn;
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
          
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_REPORT_ALL_STATUS_MESSAGE, OTX_FALSE); DemoCheckRc(result);

      result = OtxDrvGetAttributeType(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_REPORT_ALL_STATUS_MESSAGE, &eAttribType, &cbAttribValue, &bReadOnly); 
      DemoLogf("eAttribType = %d, cbAttribValue = %d, bReadOnly = %d, Li%d \n", eAttribType, cbAttribValue, bReadOnly, nCurrentLi);
      
      result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_REPORT_ALL_STATUS_MESSAGE, &bReportOn); DemoCheckRc(result);
      
      if (bReportOn == OTX_FALSE) {
        DemoLogf("Report all status message OFF for Li%d \n", nCurrentLi);
      } else {
        DemoLogf("Report all status message ON for Li%d \n", nCurrentLi);
      }
            
    }
    break;
  
  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************

  @func

  Enumerates and finds the first board device

*******************************************************************************/
static OTX_RESULT FindBoardDevice(OTX_INT32 *peBrdType) 
{
  OTX_UINT32   eDevType;
  OTX_ITERATOR itrChildTypes;
  OTX_ITERATOR itrChildrenOfSameType;
  static char szBuf[256];
  OTX_RESULT otxRc;

  *peBrdType = OTX_DEVICE_UNKNOWN; // Defensive
  
  // Initialize the opaque iterator for the hosted types
  OtxDrvIteratorBegin(&itrChildTypes);
  while (OTX_SUCCESS(OtxDrvEnumDeviceType(0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
    OTX_HANDLE hChildDevice;
    // Initialize the opaque iterator for the devices within the same type
    OtxDrvIteratorBegin(&itrChildrenOfSameType);
    while (OTX_SUCCESS(OtxDrvEnumDevice(0, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
      if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {

        // Check device type
        switch (eDevType) {
          // These devices appear at the top (driver level) but they are not board devices, so ignore them here
        case OTX_DEVICE_OTPD_INTERFACE:
        case OTX_DEVICE_CT_BUS:
          break;
        default:
          otxRc = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_DEVICE_STRING, szBuf, 256);
          if (!OTX_SUCCESS(otxRc)) {
            DemoLogf("Could not get the OTX_ATTR_DEVICE_STRING attribute for device 0x%X\n", hChildDevice);
            return otxRc;
          } 
          DemoLogf("Found %s board (devType=%d)\n", szBuf, eDevType);
          *peBrdType = eDevType;
          return OTX_S_OK;
        }

      }
    }
  }

  return OTX_S_FALSE; // We did not find an OTX board
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("T1E1HostedHdlc, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("E1T1 Device Example program.\n");
  DemoLogf("For most OTX PCI, PCIe, ExpressCard, and CSI products\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    DemoLogf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  if (g_demoEnv.nBoardType == OTX_DEVICE_UNKNOWN) {
    // Board type was not supplied by a command line parameter, so enumerate and find a board
    result = FindBoardDevice(&g_demoEnv.nBoardType);
    if (result == OTX_S_FALSE) {
      DemoLogf("No OTX board was found\n");
      DemoExitQuestion();
    } else if (result != OTX_S_OK) {
      DemoExitQuestion();
    }
  }

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  DemoLogf("Setting clocks...\n");
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(result);
  //result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(result);

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing logical devices...\n");
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
#ifdef DAVINCI
    cbAttribValue += 4;
#endif
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
#ifdef DAVINCI
    cbAttribValue += 4;
#endif
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  // Read CPLD Build Numbers from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  if(g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
    // The Thor-4-ExpressCard can monitor 4 line (2 spans)
    // or receive and transmit on the 2 lines (2 spans)
    g_demoEnv.coLi = 2; // in this mode there are only 2 lines
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);

  DemoLogf("\nPress 'h' for help\n");

#ifdef DAVINCI // Don't use IDLE character
  {
    int i;
    for (i=0; i<g_demoEnv.coLi; i++) {
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 48, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 49, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 50, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 51, 0x00); DemoCheckRc(result);
    }
  }
#endif


  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}

