/* $Workfile: ThorDtmf.c $       $Revision: 3 $      $Modtime: 12/05/12 4:34p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    ThorDtmf.c

    A simple example console application demonstrating the use the following
    devices:
    
      OTX_LDEVICE_TONE_DTMF_DIALER, 
      OTX_LDEVICE_TONE_DTMF_DETECTOR, 
      OTX_LDEVICE_TONE_MF_DIALER 
      OTX_LDEVICE_TONE_MF_DETECTOR, 
      OTX_LDEVICE_TONE_SINEWAVE_GENERATOR,

   DTMF / MF digits are generated and detected, two methods to generate MF
   digits.
    
   The applications works with the following board combinations:

      OTX Thor PCI-Plus and PCIe boards,

      Thor-2[8]-PCI-Plus + Vidar-5x4-ASM-CST[EX,PRO]/Vidar-55x4-ASM
      Thor-2[8]-PCIe + Vidar-5x4-ASM-CST[EX,PRO]/Vidar-55x4-ASM
      
    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 8(2) E1 Lines and expects them to be looped with an external
    loop cables. It creates various tone device on the DSP, which are used to 
    detect and generate tones. 
    
    The total number of logical devices supported in this demo may vary by the 
    type of Vidar daugnter board, for Thor-8, it (nNoOfDevs) can be adjusted in 
    DemoParseCmdLineArgs() to the followings:

      for Vidar-55x4-ASM:    DTMF 128, MF 128  (the maximum number 32 TS * 4 Highway) 
          Vidar-5x4-ASM-EX:  DTMF 80,  MF 80 
          Vidar-5x4-ASM-PRO: DTMF 128, MF 128

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"


#define DEMO_TASK_REF 1L                    // Task Reference for a non-blocking call

#define DEMO_DEFAULT_LOG_FILE "ThorDtmf.log"
#define DEMO_DTMF 1                         // 1 for DTMF or 0 for MF

#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_MAX_OF_DEVS      128           

typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  OTX_INT32       nBoardType;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  int             boardNo;
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;                // T1 or E1 mode
  OtxDspSpmE      eProgramModule;           // Standard Program Module to be loaded in the DSP (from the Lib)
  int             coDsp;
  OtxMfToneE      eMfToneType;              // Type of MF Tone to detect
  int             nIteration;
  int             coLi;
  OTX_INT32       nNoOfDevs;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(const char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);

void *DemoDtmfDetector(void *lpParameter);
void *DemoDtmfDialer(void *lpParameter);

void *DemoMfDetector(void *lpParameter);
void *DemoSineWaveSender(void *lpParameter);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
const char *g_sqDtmfNumber = "0123456789ABCD*#";
const char *g_sqMfNumber   = "0123456789abcde";

unsigned int g_nOldID = 20;   // larger than 15

OTX_BOOL  g_sqDev[DEMO_MAX_OF_DEVS] = {OTX_FALSE};
char g_sqDialDigit[DEMO_MAX_OF_DEVS][20];
OTX_INT32 g_nDialer;

struct sPhysDevs {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2];    // Only used for  Thor-4(8)-PCI-Plus
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hCodec[4];
  OTX_HANDLE hDsp[4];
} g_hPhysDevs;


struct sPipes {
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
} g_hPipes;


#if  DEMO_DTMF

struct sDtmfDialer {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_HANDLE hHandle;   // logical device
  OTX_HANDLE hDspLiPipe;

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

  clock_t startTime, stopTime;

} g_sqDtmfDialer[DEMO_MAX_OF_DEVS];


struct sDtmfDetector {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_HANDLE hHandle;   // logical device
  OTX_HANDLE hLiDspPipe;

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

} g_sqDtmfDetector[DEMO_MAX_OF_DEVS];


#else  // MF

struct sMfDetector {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_HANDLE hHandle;   // logical device
  OTX_HANDLE hLiDspPipe;

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

  clock_t startTime, stopTime;

} g_sqMfDetector[DEMO_MAX_OF_DEVS];


struct sMfDialer {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_HANDLE hHandle;   // logical device
  OTX_HANDLE hDspLiPipe;

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

} g_sqMfDialer[DEMO_MAX_OF_DEVS];


struct sSineWaveSender {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_HANDLE hHandle;   // logical device
  OTX_HANDLE hDspLiPipe;

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

  clock_t startTime, stopTime;

} g_sqSineWaveSender[DEMO_MAX_OF_DEVS];


#endif


struct sT1E1CfgData {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;



/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func
    Initalize the global data for the DTMF / MF Dialer and Detector devices

*******************************************************************************/
void DemoInitializeGlobalData(
  void
)
{
  OTX_INT32 i;

#if DEMO_DTMF 

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
    g_sqDtmfDialer[i].nDev = g_sqDtmfDetector[i].nDev = i;
    
    // Li0, TS1, DSP0, DSP's TS is picked up by driver in this demo, see OtxDrvConnectPipe(...,NULL,0) of DemoConnectDevices() 
    // Li1, TS1, DSP1
    // Li2, TS1, DSP2
    // Li3, TS1, DSP3
    // Li4, TS1, DSP0
    // Li5, TS1, DSP1
    // Li6, TS1, DSP2
    // Li7, TS1, DSP3
    // Li0, TS2, DSP0
    // Li1, TS2, DSP1
    // Li2, TS2, DSP2
    // Li3, TS2, DSP3
    // Li4, TS2, DSP0
    // Li5, TS2, DSP1
    // Li6, TS2, DSP2
    // Li7, TS2, DSP3
    // ......

    g_sqDtmfDialer[i].nTs  = g_sqDtmfDetector[i].nTs  = i/g_demoEnv.coLi + 1;   // not using TS0
    g_sqDtmfDialer[i].nLi  = g_sqDtmfDetector[i].nLi  = i%g_demoEnv.coLi;       
    g_sqDtmfDialer[i].nDsp = g_sqDtmfDetector[i].nDsp = i%4;                          

    g_sqDtmfDialer[i].bActive = g_sqDtmfDetector[i].bActive = OTX_FALSE;
 
  }

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
    DemoLogf("TX[%.2d] Li=%.2d Li Ts=%.2d Dsp=%.2d \n", 
      i,
      g_sqDtmfDialer[i].nLi,       
      g_sqDtmfDialer[i].nTs,       
      g_sqDtmfDialer[i].nDsp
      );
  }
  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
    DemoLogf("RX[%.2d] Li=%.2d Li Ts=%.2d Dsp=%.2d \n", 
      i,
      g_sqDtmfDetector[i].nLi,       
      g_sqDtmfDetector[i].nTs,       
      g_sqDtmfDetector[i].nDsp      
    );
  }

#else  // MF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
    g_sqSineWaveSender[i].nDev = g_sqMfDetector[i].nDev = i;
    
    // Li0, TS1, DSP0, DSP's TS is picked up by driver in this demo, see OtxDrvConnectPipe(...,NULL,0) of DemoConnectDevices() 
    // Li1, TS1, DSP1
    // Li2, TS1, DSP2
    // Li3, TS1, DSP3
    // Li4, TS1, DSP0
    // Li5, TS1, DSP1
    // Li6, TS1, DSP2
    // Li7, TS1, DSP3
    // Li0, TS2, DSP0
    // Li1, TS2, DSP1
    // Li2, TS2, DSP2
    // Li3, TS2, DSP3
    // Li4, TS2, DSP0
    // Li5, TS2, DSP1
    // Li6, TS2, DSP2
    // Li7, TS2, DSP3
    // ......

    g_sqSineWaveSender[i].nTs  = g_sqMfDetector[i].nTs  = i/DEMO_NO_OF_LIS + 1;   // not using TS0
    g_sqSineWaveSender[i].nLi  = g_sqMfDetector[i].nLi  = i%DEMO_NO_OF_LIS;       // eight Lis
    g_sqSineWaveSender[i].nDsp = g_sqMfDetector[i].nDsp = i%4;                          

    g_sqSineWaveSender[i].bActive = g_sqMfDetector[i].bActive = OTX_FALSE;
 
  }

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
    DemoLogf("TX[%.2d] Li=%.2d Li Ts=%.2d Dsp=%.2d  \n", 
      i,
      g_sqSineWaveSender[i].nLi,       
      g_sqSineWaveSender[i].nTs,       
      g_sqSineWaveSender[i].nDsp      
      );
  }
  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
    DemoLogf("RX[%.2d] Li=%.2d Li Ts=%.2d Dsp=%.2d \n", 
      i,
      g_sqMfDetector[i].nLi,       
      g_sqMfDetector[i].nTs,       
      g_sqMfDetector[i].nDsp      
    );
  }

#endif

}


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // create a notification event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);						

  // create event queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // enable event queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDsp;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4(8)-PCI-Plus

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCIE:
    DemoLogf("Opening Thor-8-PCIe board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE)g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Get handles to T1/E1 line interface devices #0 - #7
  switch (g_demoEnv.nBoardType) {
  
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;

  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCIE:
    {
      // On the Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;

  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Get handles to Codec devices #0 - #3
  for (i = 0; i < 4; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);

    // Set correct companded law (u-law or a-law)
    result = OtxBrdCodecSetLaw(g_hPhysDevs.hCodec[i], g_demoEnv.law); DemoCheckRc(g_hPhysDevs.hCodec[i], result);
  }

  // search for Vidar type
  {
    OTX_UINT32   eDevType;
    OTX_ITERATOR itrChildTypes;
    OtxDrvIteratorBegin(&itrChildTypes);
    while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
      OTX_BOOL bOpenAsm = OTX_FALSE;
      switch (eDevType) {
      case OTX_DEVICE_VIDAR_5X4_ASM_CST:
        DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_EX:
        DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
        DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_55x4_ASM:
        DemoLogf("Opening Vidar-55x4-ASM board\n");
        bOpenAsm = OTX_TRUE;
        break;
      }

      if (bOpenAsm) {
        // Get a handle to to the ASM Daughter Board
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }      
    }
      
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
    } 

  }


  // Load DSP Program from OtxDrv.lib
  for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2) {
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2\n", nDsp);
    } else if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2EX) {
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2EX\n", nDsp);
    }

    result = OtxDspRunSpm(g_hPhysDevs.hDsp[nDsp], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);

  }
    
}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // configure transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // set termination for the T1/E1 front end
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4(8)-PCI-Plus

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  // If board type is Thor-8-PCI-(Plus) then enable Quad E1T1 device
  if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCIE) ) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }
    
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

#if DEMO_DTMF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
      
    // create a notification event and event queue for the DTMF Detector devices
    result = OtxDrvCreateEvent(&g_sqDtmfDetector[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_sqDtmfDetector[i].hNotificationEvent, &g_sqDtmfDetector[i].hEventQueue); DemoCheckRc(g_sqDtmfDetector[i].hEventQueue, result);
    result = OtxDrvEnable(g_sqDtmfDetector[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDetector[i].hEventQueue, result);

    // create DTMF Detector logical devices
    result = OtxDrvCreateLogicalDevice( 
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_DTMF_DETECTOR, 
      g_sqDtmfDetector[i].hEventQueue, 
      i,                  
      &g_sqDtmfDetector[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);


    // create a notification event and event queue for the DTMF Dialer devices
    result = OtxDrvCreateEvent(&g_sqDtmfDialer[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_sqDtmfDialer[i].hNotificationEvent, &g_sqDtmfDialer[i].hEventQueue); DemoCheckRc(g_sqDtmfDialer[i].hEventQueue, result);
    result = OtxDrvEnable(g_sqDtmfDialer[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[i].hEventQueue, result);

    // create DTMF Dialer logical devices
    result = OtxDrvCreateLogicalDevice(
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_DTMF_DIALER, 
      g_sqDtmfDialer[i].hEventQueue, 
      i, 
      &g_sqDtmfDialer[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);

  }

#else  // MF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
      
    // create a notification event and event queue for the MF Detector devices
    result = OtxDrvCreateEvent(&g_sqMfDetector[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_sqMfDetector[i].hNotificationEvent, &g_sqMfDetector[i].hEventQueue); DemoCheckRc(g_sqMfDetector[i].hEventQueue, result);
    result = OtxDrvEnable(g_sqMfDetector[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqMfDetector[i].hEventQueue, result);

    // create MF Detector logical devices
    result = OtxDrvCreateLogicalDevice( 
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_MF_DETECTOR, 
      g_sqMfDetector[i].hEventQueue, 
      i,                  
      &g_sqMfDetector[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);


    // create a notification event and event queue for the MF Dialer devices
    result = OtxDrvCreateEvent(&g_sqMfDialer[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_sqMfDialer[i].hNotificationEvent, &g_sqMfDialer[i].hEventQueue); DemoCheckRc(g_sqMfDialer[i].hEventQueue, result);
    result = OtxDrvEnable(g_sqMfDialer[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqMfDialer[i].hEventQueue, result);

    // Create MF Dialers Logical Devices
    result = OtxDrvCreateLogicalDevice(
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_MF_DIALER, 
      g_sqMfDialer[i].hEventQueue, 
      i, 
      &g_sqMfDialer[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);


    // create a notification event and event queue for the Sine Wave Generator devices
    result = OtxDrvCreateEvent(&g_sqSineWaveSender[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_sqSineWaveSender[i].hNotificationEvent, &g_sqSineWaveSender[i].hEventQueue); DemoCheckRc(g_sqSineWaveSender[i].hEventQueue, result);
    result = OtxDrvEnable(g_sqSineWaveSender[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqSineWaveSender[i].hEventQueue, result);

    // create Sine Wave Generator logical devices
    result = OtxDrvCreateLogicalDevice(
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_SINEWAVE_GENERATOR, 
      g_sqSineWaveSender[i].hEventQueue, 
      i, 
      &g_sqSineWaveSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);

  }

#endif

}

/*******************************************************************************

  @func
    Close Logical Devices

*******************************************************************************/
void DemoCloseLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;


#if DEMO_DTMF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {

    // Close DTMF Detector Logical Devices
    result = OtxDrvClose(g_sqDtmfDetector[i].hHandle); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);

    // Close DTMF Dialers Logical Devices
    result = OtxDrvClose(g_sqDtmfDialer[i].hHandle); DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);

  }

#else  // MF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {

    // Close MF Detector Logical Devices
    result = OtxDrvClose(g_sqMfDetector[i].hHandle); DemoCheckRc(g_sqMfDetector[i].hHandle, result);

    // Close MF Dialers Logical Devices
    result = OtxDrvClose(g_sqMfDialer[i].hHandle); DemoCheckRc(g_sqMfDialer[i].hHandle, result);

    // Close Sine Wave Generator Logical Devices
    result = OtxDrvClose(g_sqSineWaveSender[i].hHandle); DemoCheckRc(g_sqSineWaveSender[i].hHandle, result);

  }

#endif

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 nDev;
  OTX_INT32 nTs;
  OTX_INT32 nLi;
  OTX_INT32 nDsp;

  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];

  // connect Li pipes
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    // create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn[nLi]); DemoCheckRc(g_hPipes.hLiIn[nLi], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiOut[nLi]); DemoCheckRc(g_hPipes.hLiOut[nLi], result);

    // connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      nTs = 24; 
    } else {
      // E1 mode
      nTs = 16; 
    }
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[nLi], result);
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_INPUT, g_hPhysDevs.hLi[nLi], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiOut[nLi], result);

  }

#if DEMO_DTMF

  // DTMF devices connection
  for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {
    result = OtxDrvCreatePipe(64, &g_sqDtmfDetector[nDev].hLiDspPipe); DemoCheckRc(g_sqDtmfDetector[nDev].hLiDspPipe, result);
    result = OtxDrvCreatePipe(64, &g_sqDtmfDialer[nDev].hDspLiPipe); DemoCheckRc(g_sqDtmfDialer[nDev].hDspLiPipe, result);

  }                  

  // connect pipes for the DTMF Detector
  for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {
    
    // connect pipe, see DemoInitializeGlobalData()
    nTs  = g_sqDtmfDetector[nDev].nTs;
    nLi  = g_sqDtmfDetector[nDev].nLi;
    nDsp = g_sqDtmfDetector[nDev].nDsp;

    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF;

    result = OtxDrvConnectPipe(g_sqDtmfDetector[nDev].hLiDspPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDetector[nDev].hLiDspPipe, result);
    result = OtxDrvConnectPipe(g_sqDtmfDetector[nDev].hLiDspPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_sqDtmfDetector[nDev].hLiDspPipe, result);
  }

  // connect pipes for the DTMF Dialer 
  for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {

    // connect pipes, DemoInitializeGlobalData()
    nTs  = g_sqDtmfDialer[nDev].nTs;
    nLi  = g_sqDtmfDialer[nDev].nLi;
    nDsp = g_sqDtmfDialer[nDev].nDsp;

    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF;

    result = OtxDrvConnectPipe(g_sqDtmfDialer[nDev].hDspLiPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDialer[nDev].hDspLiPipe, result);
    result = OtxDrvConnectPipe(g_sqDtmfDialer[nDev].hDspLiPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_sqDtmfDialer[nDev].hDspLiPipe, result);
  }


  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {

    // Create the following Connections in receive direction:
    //  Li  ------> DtmfDetector
    result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[nDev].hLiDspPipe, -1, g_sqDtmfDetector[nDev].hHandle, -1); DemoCheckRc(g_sqDtmfDetector[nDev].hHandle, result);

    // Make the following connections in transmit direction:
    //              DtmfDialler   -----> Li
    result = OtxDrvConnectLogicalDevice(g_sqDtmfDialer[nDev].hHandle, -1, g_sqDtmfDialer[nDev].hDspLiPipe, -1); DemoCheckRc(g_sqDtmfDialer[nDev].hHandle, result);
  }

#else // MF

  // MF devices connection
  for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {
    result = OtxDrvCreatePipe(64, &g_sqMfDetector[nDev].hLiDspPipe); DemoCheckRc(g_sqMfDetector[nDev].hLiDspPipe, result);
    result = OtxDrvCreatePipe(64, &g_sqSineWaveSender[nDev].hDspLiPipe); DemoCheckRc(g_sqSineWaveSender[nDev].hDspLiPipe, result);

  }                  

  // connect pipes for the MF Detector
  for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {
    
    // connect pipe, see DemoInitializeGlobalData()
    nTs  = g_sqMfDetector[nDev].nTs;
    nLi  = g_sqMfDetector[nDev].nLi;
    nDsp = g_sqMfDetector[nDev].nDsp;

    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF;

    result = OtxDrvConnectPipe(g_sqMfDetector[nDev].hLiDspPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqMfDetector[nDev].hLiDspPipe, result);
    result = OtxDrvConnectPipe(g_sqMfDetector[nDev].hLiDspPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_sqMfDetector[nDev].hLiDspPipe, result);
  }

  // connect pipes for the Sine Wace Generator  
  for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {

    // connect pipes, DemoInitializeGlobalData()
    nTs  = g_sqSineWaveSender[nDev].nTs;
    nLi  = g_sqSineWaveSender[nDev].nLi;
    nDsp = g_sqSineWaveSender[nDev].nDsp;

    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF;

    result = OtxDrvConnectPipe(g_sqSineWaveSender[nDev].hDspLiPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqSineWaveSender[nDev].hDspLiPipe, result);
    result = OtxDrvConnectPipe(g_sqSineWaveSender[nDev].hDspLiPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_sqSineWaveSender[nDev].hDspLiPipe, result);
  }


  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {

    // Create the following Connections in receive direction:
    //  Li  ------> MfDetector
    result = OtxDrvConnectLogicalDevice(g_sqMfDetector[nDev].hLiDspPipe, -1, g_sqMfDetector[nDev].hHandle, -1); DemoCheckRc(g_sqMfDetector[nDev].hHandle, result);

    // Make the following connections in transmit direction:
    //              MfDialler   -----> Li
    result = OtxDrvConnectLogicalDevice(g_sqMfDialer[nDev].hHandle,      -1, g_sqSineWaveSender[nDev].hDspLiPipe, -1); DemoCheckRc(g_sqMfDialer[nDev].hHandle, result);

    result = OtxDrvConnectLogicalDevice(g_sqSineWaveSender[nDev].hHandle, -1, g_sqSineWaveSender[nDev].hDspLiPipe, -1); DemoCheckRc(g_sqSineWaveSender[nDev].hHandle, result);
  }

#endif

}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

#if DEMO_DTMF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {

    // Initialize the DTMF Detector
    result = OtxToneDtmfDetectorInit(
      g_sqDtmfDetector[i].hHandle, 
      g_demoEnv.law, 
      OTX_TRUE); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);


    // Initialize the DTMF Diallers
    result = OtxToneDtmfDiallerInit(
      g_sqDtmfDialer[i].hHandle, 
      g_demoEnv.law, 
      8, 
      8, 
      OTX_TONE_GAIN_0DB); 
    DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);

  }

#else // MF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {

    // Initialize the MF Tone Detector
    result = OtxToneMfDetectorInit(
      g_sqMfDetector[i].hHandle, 
      g_demoEnv.law,
      g_demoEnv.eMfToneType,  // MF Tone type (OTX_TONE_MF_R1, OTX_TONE_MF_R2_FORWARD, OTX_TONE_MF_R2_BACKWARD)
      OTX_TRUE,               // If OTX_TRUE, generate events and store the received digits in the Event QueusFifo, If OTX_FALSE Don't store. 
      OTX_FALSE,              // If OTX_TRUE the detector generates an event when a tone ends    
      25);                    // Minimum duration (in milliseconds) required to validate a digit 
    DemoCheckRc(g_sqMfDetector[i].hHandle, result);


    // Initialize the MF Diallers
    result = OtxToneMfDiallerInit(
      g_sqMfDialer[i].hHandle, 
      g_demoEnv.law, 
      g_demoEnv.eMfToneType,
      50, 
      25, 
      OTX_TONE_GAIN_0DB); 
    DemoCheckRc(g_sqMfDialer[i].hHandle, result);

  }

#endif

}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

#if DEMO_DTMF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
   
    result = OtxDrvEnable(g_sqDtmfDetector[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);
    result = OtxDrvEnable(g_sqDtmfDialer[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);
  }

#else  // MF

  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
   
    result = OtxDrvEnable(g_sqMfDetector[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqMfDetector[i].hHandle, result);    
    result = OtxDrvEnable(g_sqMfDialer[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqMfDialer[i].hHandle, result);
    
    result = OtxDrvEnable(g_sqSineWaveSender[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqSineWaveSender[i].hHandle, result);
  }

#endif

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*********************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}


/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Print Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}


#if DEMO_DTMF 
/*******************************************************************************

  @func

  Detect DTMF digits

  @comm

*******************************************************************************/
void *DemoDtmfDetector(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 nDev;
  int nIndex = 0;

  clock_t startTime = 0, stopTime;
  
  // validate parameters
  nDev = *((OTX_INT32 *) lpParameter);
  if (nDev < 0 || nDev > g_demoEnv.nNoOfDevs) {
    DemoLogf("DemoDtmfDetector: Thread was passed an invalid nDev(%d). Thread exiting...\n", nDev);
    return 0;
  }
        
  startTime = clock();

  g_sqDtmfDetector[nDev].bActive = OTX_TRUE;
    
    
  OtxDrvWaitForSingleEvent(g_sqDtmfDetector[nDev].hNotificationEvent, OTX_TIME_INFINITY);
    
  // Check for events from this device
  {
    OtxEventDataS eventData;
    OtxEventDataS *pEventData;
    char msgStr[128];
      
    do  {
      result = OtxDrvGetEventData(g_sqDtmfDetector[nDev].hEventQueue, &eventData); 
      OsaThreadSleep(1); // VERY IMPORTANT: give the driver some time, otherwise, CPU load is 100% 
        
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        pEventData = &eventData;

        DemoPrintTimeStamp(pEventData->m_nTimeStamp);

        switch (pEventData->m_eDeviceType) {
            
        case OTX_LDEVICE_TONE_DTMF_DETECTOR:
        result = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
          
        if (OTX_SUCCESS(result)) {
          // Event from Dtmf Detector
          if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
            OTX_INT32 i = pEventData->m_nSourceId;

            switch (pEventData->m_nParam) {
            case 0:  // 0  (941Hz/1336Hz)
            case 1:  // 1  (697Hz/1209Hz)
            case 2:  // 2  (697Hz/1336Hz)
            case 3:  // 3  (697Hz/1477Hz)
            case 4:  // 4  (770Hz/1209Hz)
            case 5:  // 5  (770Hz/1336Hz)
            case 6:  // 6  (770Hz/1477Hz)
            case 7:  // 7  (852Hz/1209Hz)
            case 8:  // 8  (852Hz/1336Hz)
            case 9:  // 9  (852Hz/1477Hz)
            DemoLogf("DtmfDetector[%d,Li%d,Ts%d,Dsp%d] detected DTMF digit %d \n", pEventData->m_nSourceId, g_sqDtmfDetector[i].nLi, g_sqDtmfDetector[i].nTs, g_sqDtmfDetector[i].nDsp, pEventData->m_nParam);
            break;

            case 10: // A  (697Hz/1633Hz)
            case 11: // B  (770Hz/1633Hz)
            case 12: // C  (852Hz/1633Hz)
            case 13: // D  (941Hz/1633Hz)
            DemoLogf("DtmfDetector[%d,Li%d,Ts%d,Dsp%d] detected DTMF key %C \n", pEventData->m_nSourceId, g_sqDtmfDetector[i].nLi, g_sqDtmfDetector[i].nTs, g_sqDtmfDetector[i].nDsp, pEventData->m_nParam - 10 + 'A');
            break;

            case 14: // *  (941Hz/1209Hz)
            DemoLogf("DtmfDetector[%d,Li%d,Ts%d,Dsp%d] detected DTMF key '*'\n", pEventData->m_nSourceId, g_sqDtmfDetector[i].nLi, g_sqDtmfDetector[i].nTs, g_sqDtmfDetector[i].nDsp);
            break;

            case 15: // #  (941Hz/1477Hz)
            DemoLogf("DtmfDetector[%d,Li%d,Ts%d,Dsp%d] detected DTMF key '#'\n", pEventData->m_nSourceId, g_sqDtmfDetector[i].nLi, g_sqDtmfDetector[i].nTs, g_sqDtmfDetector[i].nDsp);
            break;

            default: 
            DemoLogf("DtmfDetector[%d,Li%d,Ts%d,Dsp%d] Unrecognized DTMF tone (%d)\n", pEventData->m_nSourceId, g_sqDtmfDetector[i].nLi, g_sqDtmfDetector[i].nTs, g_sqDtmfDetector[i].nDsp, pEventData->m_nParam);
            break;
            }
          }

          // compare the detected and the sent
          if ( g_nOldID == pEventData->m_nSourceId) {
              nIndex++;
            if (nIndex == (int)strlen(g_sqDtmfNumber) ) {
              nIndex = 0;
            }
          } else {
            g_nOldID = pEventData->m_nSourceId;
            nIndex = 0;
          }

          switch (pEventData->m_nParam) {
          case 0:

            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '0') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '0') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '0') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '0') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            }
            break;

          case 1:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '1') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '1') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '1') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '1') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 2:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '2') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '2') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '2') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '2') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 3:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '3') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '3') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '3') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '3') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 4:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '4') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '4') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '4') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '4') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 5:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '5') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '5') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '5') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '5') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 6:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '6') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '6') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '6') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '6') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 7:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '7') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '7') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '7') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '7') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 8:

            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '8') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '8') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '8') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '8') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 9:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '9') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '9') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '9') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '9') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 10:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == 'A') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'A') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == 'A') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'A') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 11:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == 'B') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'B') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == 'B') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'B') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 12:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == 'C') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'C') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == 'C') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'C') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 13:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == 'D') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'D') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == 'D') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'D') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 14:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '*') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '*') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '*') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '*') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          case 15:
            
            if (g_nDialer % 2 == 0) { // dialer is even number

              if (g_sqDialDigit[pEventData->m_nSourceId - 1][nIndex] == '#') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '#') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }

            } else { // dialer is odd number

              if (g_sqDialDigit[pEventData->m_nSourceId + 1][nIndex] == '#') {  // compare with saved data
                DemoLogf("loop between Lis: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else if (g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '#') {
                DemoLogf("loop to itself: correct DTMF digit = %d is detected \n",  pEventData->m_nParam);
              }
              else {
                DemoLogf("The loop cable either loops back to itself or Li0<->Li1, etc. \n");
              }
            }
            break;

          default:
            DemoLogf("unrecognized DTMF digit\n" );
            break;
          }
        } else {
          // Unexpected event from this device
          DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
        }
        break;
            
        default:
          // Unhandled event
          DemoLogf("DemoDtmfDetector[%d]: Event from an unknown object 0x%X\n", nDev, pEventData->m_hDevice);
          OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
          break;
        }
      }
                
    } while(g_sqDtmfDetector[nDev].bActive);
    
  }
  
  stopTime = clock();
  
  DemoLogf("DemoDtmfDetector[%d] Thread Exiting (time=%2.1f seconds).\n", nDev, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
  
  return 0;
}


/*******************************************************************************

  @func

  Dial DTMF digits

  @comm

*******************************************************************************/
void *DemoDtmfDialer(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 nDialer;
  
  OTX_INT32 nDigits = strlen(g_sqDtmfNumber);
  OTX_HANDLE hHandle;
  const char *cDigit = "0";  
  float fRandom;
  int   nDialDigit;
  clock_t startTime = 0, stopTime;
  
  // validate parameters
  nDialer = *((OTX_INT32 *) lpParameter);
  if (nDialer < 0 || nDialer > g_demoEnv.nNoOfDevs) {
    DemoLogf("DemoDtmfDetector: Thread was passed an invalid nDialer(%d). Thread exiting...\n", nDialer);
    return 0;
  }

  srand( (unsigned)time(NULL) );

  rand();
  rand();
  rand();
  rand();
  fRandom = (float) rand();
  nDialDigit = (int) (fRandom / RAND_MAX * 16.0);  // 16 different digits

  switch (nDialDigit) {

  case 0:
    cDigit = "0";  
    break;
  case 1:
    cDigit = "1";  
    break;
  case 2:
    cDigit = "2";  
    break;
  case 3:
    cDigit = "3";  
    break;
  case 4:
    cDigit = "4";  
    break;
  case 5:
    cDigit = "5";  
    break;
  case 6:
    cDigit = "6";  
    break;
  case 7:
    cDigit = "7";  
    break;
  case 8:
    cDigit = "8";  
    break;
  case 9:
    cDigit = "9";  
    break;
  case 10:
    cDigit = "A";  
    break;
  case 11:
    cDigit = "B";  
    break;
  case 12:
    cDigit = "C";  
    break;
  case 13:
    cDigit = "D";  
    break;
  case 14:
    cDigit = "*";  
    break;
  case 15:
    cDigit = "#";  
    break;
  default:
    DemoLogf("wrong digit \n");
  }

  nDigits = strlen(cDigit);

  // save the digit to be compared
  if (nDialDigit < 10) {  // '0'-'9', 'A'-'D', "*" or '#'
    g_sqDialDigit[nDialer][0] = (char) ('0' + nDialDigit);
  } else if ( nDialDigit >= 10 && nDialDigit <= 13 ) {
    g_sqDialDigit[nDialer][0] = (char) ('A' + nDialDigit - 10);
  } else if ( nDialDigit == 14  ) {
    g_sqDialDigit[nDialer][0] = (char) ('*');
  } else if ( nDialDigit == 15 ) {
    g_sqDialDigit[nDialer][0] = (char) ('#');
  }

  // save the dialer (device number)  
  g_nDialer = nDialer;
  
  startTime = clock();

  g_sqDtmfDialer[nDialer].bActive = OTX_TRUE;
                
  hHandle = g_sqDtmfDialer[nDialer].hHandle;

  DemoLogf("\nSending digit %s on Device#%d Li%d,Ts%d, Dsp%d\n", cDigit ,nDialer, g_sqDtmfDialer[nDialer].nLi, g_sqDtmfDialer[nDialer].nTs, g_sqDtmfDialer[nDialer].nDsp);
  // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
  result = OtxToneDtmfDial(hHandle, (OTX_UINT8 *)cDigit, nDigits, DEMO_TASK_REF, OTX_TIME_INFINITY); 

  if (OTX_RESULT_CODE(result) == OTX_DTMF_DIALER_RES_BUSY) {
    DemoLogf("Dialer busy.\n");
  }   else {
    DemoCheckRc(hHandle, result);
    DemoLogf("Done.\n");
  }

  OsaThreadSleep(1000);      
  
  stopTime = clock();
  
  DemoLogf("DemoDtmfDialer[%d] Thread Exiting (time=%2.1f seconds).\n\n", nDialer, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
  
  return 0;
}

#else // MF


/*******************************************************************************

  @func

  Detect MF digits

  @comm

*******************************************************************************/
void *DemoMfDetector(
  void *lpParameter   // thread data
)
{  

  OTX_RESULT result;
  OTX_UINT32 nDev;
  int nIndex = 0;

  clock_t startTime = 0, stopTime;
  
  // validate parameters
  nDev = *((OTX_INT32 *) lpParameter);
  if (nDev < 0 || nDev > g_demoEnv.nNoOfDevs) {
    DemoLogf("DemoMfDetector: Thread was passed an invalid nDev(%d). Thread exiting...\n", nDev);
    return 0;
  }
        
  startTime = clock();

  g_sqMfDetector[nDev].bActive = OTX_TRUE;
    
    
  OtxDrvWaitForSingleEvent(g_sqMfDetector[nDev].hNotificationEvent, OTX_TIME_INFINITY);
    
  // Check for events from this device
  {
    OtxEventDataS eventData;
    OtxEventDataS *pEventData;
    char msgStr[128];
      
    do  {
      result = OtxDrvGetEventData(g_sqMfDetector[nDev].hEventQueue, &eventData); 
      OsaThreadSleep(1); // VERY IMPORTANT: give the driver some time, otherwise, CPU load is 100% 
        
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        pEventData = &eventData;
          
        switch (pEventData->m_eDeviceType) {
            
        case OTX_LDEVICE_TONE_MF_DETECTOR:
          result = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
          if (OTX_SUCCESS(result)) {
            // Event from Dtmf Detector
            if (pEventData->m_nCode == OTX_MF_DETECTOR_EC_DIGIT) {
              DemoLogf("MfDetector[%d] detected MF digit %.2d \n", pEventData->m_nSourceId, pEventData->m_nParam);
            } else if (pEventData->m_nCode == OTX_MF_DETECTOR_EC_VERBOSE_STATUS){
              DemoLogf("MfDetector[%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
            } else {
              if (pEventData->m_nParam) {
                DemoLogf("MfDetector[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
              } else {
                DemoLogf("MfDetector[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
              }
            }
          } else {
            DemoLogf("MfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
          }
        

          // compare the detected and the sent
          if ( g_nOldID == pEventData->m_nSourceId) {
              nIndex++;
            if (nIndex == (int)strlen(g_sqMfNumber) ) {
              nIndex = 0;
            }
          } else {
            g_nOldID = pEventData->m_nSourceId;
            nIndex = 0;
          }

          switch (pEventData->m_nParam) {
          case 0:
            if ( g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '0' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '0' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 1:
            if ( g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '1' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '1' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 2:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '2' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '2' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 3:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '3' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '3' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 4:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '4' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '4' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 5:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '5' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '5' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 6:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '6' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '6' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 7:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '7' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '7' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 8:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '8' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '8' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 9:                                                   
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == '9' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit '9' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 10:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'a' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit 'a' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 11:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'b' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit 'b' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 12:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'c' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit 'c' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 13:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'd' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit 'd' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;
          case 14:
            if ( (char) g_sqDialDigit[pEventData->m_nSourceId][nIndex] == 'e' ) {
              DemoLogf("correct MF digit = %d is detected \n",  pEventData->m_nParam );
            } else {
              DemoLogf("error: MF digit 'e' is detected, but digit %c was sent \n", g_sqDialDigit[pEventData->m_nSourceId][nIndex] );
            }
            break;

          default:
              DemoLogf("unrecongnized MF digit\n" );
            break;
          }

          break;
            
        default:
          // Unhandled event
          DemoLogf("DemoMfDetector[%d]: Event from an unknown object 0x%X\n", nDev, pEventData->m_hDevice);
          OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
          break;
        }
      }
                
    } while(g_sqMfDetector[nDev].bActive);
    
  }
  
  stopTime = clock();
  
  DemoLogf("DemoMfDetector[%d] Thread Exiting (time=%2.1f seconds).\n\n", nDev, (double)(stopTime - startTime) / CLOCKS_PER_SEC);

  
  return 0;
}



/*******************************************************************************

  @func

    Dial MF digits by Sine Wave

  @comm

*******************************************************************************/
void *DemoSineWaveSender(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  static OTX_INT32 nMfDigit = 0;
  char *szMfDigitStr = "?";
  OTX_INT32 nDialer;

  float fRandom;

  // Initialize the SineWave Sender 
  OtxToneSineWaveDataS sqSineWaveData[5];

  // validate parameters
  nDialer = *((OTX_INT32 *) lpParameter);
  if (nDialer < 0 || nDialer > g_demoEnv.nNoOfDevs) {
    DemoLogf("DemoDtmfDetector: Thread was passed an invalid nDialer(%d). Thread exiting...\n", nDialer);
    return 0;
  }

  srand( (unsigned)time(NULL) );

  rand();
  rand();
  rand();
  rand();
  fRandom = (float) rand();
  nMfDigit = (int) (fRandom / RAND_MAX * 15.0);  // 15 different digits 

  sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[0].nPhase       = 0;  

  sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[1].nPhase       = 0;  

  switch (nMfDigit) {
  case 0:
    // MF digit 0 
    szMfDigitStr = "0";
    sqSineWaveData[0].nFrequency   = 1300; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 1:
    // MF digit 1 
    szMfDigitStr = "1";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 900; // in Hz       
    break;
  case 2:
    // MF digit 2 
    szMfDigitStr = "2";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 1100; // in Hz       
    break;
  case 3:
    // MF digit 3 
    szMfDigitStr = "3";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 1100; // in Hz       
    break;
  case 4:
    // MF digit 4 
    szMfDigitStr = "4";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 1300; // in Hz       
    break;
  case 5:
    // MF digit 5 
    szMfDigitStr = "5";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 1300; // in Hz       
    break;
  case 6:
    // MF digit 6
    szMfDigitStr = "6";
    sqSineWaveData[0].nFrequency   = 1100; // in Hz
    sqSineWaveData[1].nFrequency   = 1300; // in Hz       
    break;
  case 7:
    // MF digit 7 
    szMfDigitStr = "7";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 8:
    // MF digit 8 
    szMfDigitStr = "8";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 9:
    // MF digit 9
    szMfDigitStr = "9";
    sqSineWaveData[0].nFrequency   = 1100; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 10:
    // MF KP
    szMfDigitStr = "KP(10)";
    sqSineWaveData[0].nFrequency   = 1100; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  case 11:
    // MF ST
    szMfDigitStr = "ST(11)";
    sqSineWaveData[0].nFrequency   = 1500; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  case 12:
    // MF ST1(x)
    szMfDigitStr = "ST1(12)";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  case 13:
    // MF ST2(y)
    szMfDigitStr = "ST2(13)";
    sqSineWaveData[0].nFrequency   = 1300; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  case 14:
    // MF ST3(z)
    szMfDigitStr = "ST3(14)";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  }

  // save the digit to be compared, odd number of Li connects to even number of Li
  if (nDialer % 2 == 0) {
    if (nMfDigit < 10) { // '0'-'9' or 'a'-'e'
      g_sqDialDigit[nDialer + 1][0] = (char) ('0' + nMfDigit);
    } else {
      g_sqDialDigit[nDialer + 1][0] = (char) ('a' + nMfDigit - 10);
    }
  } else {
    if (nMfDigit < 10) { 
      g_sqDialDigit[nDialer - 1][0] = (char) ('0' + nMfDigit);
    } else {
      g_sqDialDigit[nDialer - 1][0] = (char) ('a' + nMfDigit - 10);
    }
  }

    
  DemoLogf("Initalizing SineWave Generator[%d] for R1 MF tone %s (%d Hz / %d Hz)\n", nDialer, szMfDigitStr, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
  result = OtxToneSineWaveGeneratorInit(
    g_sqSineWaveSender[nDialer].hHandle, 
    g_demoEnv.law, 
    sqSineWaveData,
    2, // coSineWaveData
    OTX_TONE_GAIN_0DB); DemoCheckRc(g_sqSineWaveSender[nDialer].hHandle, result);
  
  
  DemoLogf("Enabling SineWave Generator[%d]... \n", nDialer);
  result = OtxDrvEnable(g_sqSineWaveSender[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqSineWaveSender[nDialer].hHandle, result);
  
  OsaThreadSleep(500);
  
  DemoLogf("Disabling SineWave Generator[%d]... \n", nDialer);
  result = OtxDrvDisable(g_sqSineWaveSender[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqSineWaveSender[nDialer].hHandle, result);
   
  return 0;
}

#endif



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr      = NULL;
  envP->bExitOnError     = OTX_FALSE;
  envP->boardNo          = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType       = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coDsp            = 4;
  envP->eT1E1Mode        = OTX_T1E1_MODE_E1;
  envP->eProgramModule   = OTX_DSP_SPM_OTXSPM2;
  envP->eMfToneType      = OTX_TONE_MF_R1;
  envP->law              = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->nIteration       = 1;
  envP->coLi             = 8;
  envP->nNoOfDevs        = 80;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {          // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'I') {   // Test Iterations 
          envP->nIteration = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
          if (*(argv[i]+2) == '2') {          // SPM2 
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2;
          } else if (*(argv[i]+2) == 'X') {   // SPM2EX
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2EX;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }

        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coDsp      = 4;
            envP->coLi       = 8;
          } else if ( *(argv[i]+2) == 'W') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCIE;
            envP->coDsp      = 4;
            envP->coLi       = 8;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coDsp      = 4;
            envP->coLi       = 2;
            envP->nNoOfDevs  = 62;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coDsp      = 4;
            envP->coLi       = 2;
            envP->nNoOfDevs  = 62;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {        // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') { // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {        // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') { // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}   



/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[P|T|E]    : Board Type used for the demo, where\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /P[2|X]      : DSP module \n");
  DemoLogf("                  2 = SPM2 (default)\n");
  DemoLogf("                  X = SPM2EX\n");
  DemoLogf(" /I<count>    : Test Iteration \n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(const char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF((char *) formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    if (DEMO_DTMF) { 
      DemoLogf(" d   Sending random DTMF digits \n");
      DemoLogf(" g   Sending DTMF digit string \n");
      DemoLogf(" G   Sending DTMF digit string (short string) \n");
    } else {
      DemoLogf(" k   Sending random MF digits \n");
      DemoLogf(" f   Sending MF digit string \n");
    }
		DemoLogf(" l   Query Line Status of the two Lis\n");
    
    break;

#if DEMO_DTMF 

  case 'd':
    {
      OTX_INT32 i,j;
      
      for (j = 0; j < g_demoEnv.nIteration; j++) {
        
        DemoLogf("Loop %d of %d: \n", j, g_demoEnv.nIteration);

        for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {      
          OsaThreadStartEx(DemoDtmfDialer,  (void *) &g_sqDtmfDialer[i].nDev);
          OsaThreadSleep(2000);
        }
      }
    }
    break;

  case 'g':
    {
      OTX_INT32 j;
      OTX_HANDLE hHandle;      
      int nDev = 0;

      for (j = 0; j < g_demoEnv.nIteration; j++) {
        for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {
                    
          // digits to be received by deivce
          strcpy( g_sqDialDigit[nDev], g_sqDtmfNumber);
          
          DemoLogf("\n Sending digits '%s' on Device#%d Li%d, Ts%d, Dsp%d \n", g_sqDialDigit[nDev], nDev, g_sqDtmfDialer[nDev].nLi, g_sqDtmfDialer[nDev].nTs, g_sqDtmfDialer[nDev].nDsp);
          hHandle = g_sqDtmfDialer[nDev].hHandle;   
            
          // save the dialer (device number)  
          g_nDialer = nDev;

          // start Dialing DTMF number, Non-Blocking call, will receive Event when completed
          result = OtxToneDtmfDial(hHandle, (OTX_UINT8 *)g_sqDialDigit[nDev], strlen(g_sqDialDigit[nDev]), DEMO_TASK_REF, OTX_TIME_INFINITY); 
        
          if (OTX_RESULT_CODE(result) == OTX_DTMF_DIALER_RES_BUSY) {
            DemoLogf("Dialer busy.\n");
          } else {
            DemoCheckRc(hHandle, result);
            DemoLogf("Done.\n");
          }
        
          OsaThreadSleep(4000);      
        
        } //for
      }
    }
    break;
 
  case 'G':
    {
      OTX_INT32 j;
      OTX_HANDLE hHandle;      
      int nDev = 0;
      const char *sqDtmfNumber = "123";

      for (j = 0; j < g_demoEnv.nIteration; j++) {
        for (nDev = 0; nDev < 8; nDev++) {
                    
          // digits to be received by deivce
          strcpy( g_sqDialDigit[nDev], sqDtmfNumber);
          
          DemoLogf("\n Sending digits '%s' on Device#%d Li%d, Ts%d, Dsp%d \n", g_sqDialDigit[nDev], nDev, g_sqDtmfDialer[nDev].nLi, g_sqDtmfDialer[nDev].nTs, g_sqDtmfDialer[nDev].nDsp);
          hHandle = g_sqDtmfDialer[nDev].hHandle;   
            
          // save the dialer (device number)  
          g_nDialer = nDev;

          // start Dialing DTMF number, Non-Blocking call, will receive Event when completed
          result = OtxToneDtmfDial(hHandle, (OTX_UINT8 *)g_sqDialDigit[nDev], strlen(g_sqDialDigit[nDev]), DEMO_TASK_REF, OTX_TIME_INFINITY); 
        
          if (OTX_RESULT_CODE(result) == OTX_DTMF_DIALER_RES_BUSY) {
            DemoLogf("Dialer busy.\n");
          } else {
            DemoCheckRc(hHandle, result);
            DemoLogf("Done.\n");
          }
        
          OsaThreadSleep(1500);      
        
        } //for
      }
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }

    }
    break;

 
  case 'r':
    {
      OTX_INT32 nDialler;
      
      DemoLogf("Disable/Enable DTMF dialler\n");
      for (nDialler = 0; nDialler < g_demoEnv.nNoOfDevs; nDialler++) {
        result = OtxDrvDisable(g_sqDtmfDialer[nDialler].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[nDialler].hHandle, result);
      }
      for (nDialler = 0; nDialler < g_demoEnv.nNoOfDevs; nDialler++) {
        result = OtxDrvEnable(g_sqDtmfDialer[nDialler].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[nDialler].hHandle, result);
      }
      DemoLogf("Done.");
    }
    break;

#else  // MF

  case 'f':
    {
      OTX_INT32 j;
      OTX_HANDLE hHandle;
      int nDev = 0;

      for (j = 0; j < g_demoEnv.nIteration; j++) {

        for (nDev = 0; nDev < g_demoEnv.nNoOfDevs; nDev++) {
        
          switch (g_demoEnv.eMfToneType) {
          case OTX_TONE_MF_R1:
            strcpy( g_sqDialDigit[nDev], g_sqMfNumber );
            break;
          case OTX_TONE_MF_R2_FORWARD:
          case OTX_TONE_MF_R2_BACKWARD:
            strcpy( g_sqDialDigit[nDev], g_sqMfNumber);
            break;
          default:
            strcpy( g_sqDialDigit[nDev], "1");
            DemoLogf("Unsupported value for g_demoEnv.eMfToneType (%d) \n", g_demoEnv.eMfToneType);
            break;
          }
        
          if (nDev % 2 == 0) {
            // odd number of device dials, even number of device receives 
            hHandle = g_sqMfDialer[nDev + 1].hHandle;   
            DemoLogf("Sending digits on Device#%d  '%s'\n", nDev + 1, g_sqDialDigit[nDev]);
        
          } else {  
            // enen number of device dials, odd number of device receives 
            hHandle = g_sqMfDialer[nDev - 1].hHandle;   
            DemoLogf("Sending digits on Device#%d  '%s'\n", nDev - 1, g_sqDialDigit[nDev]);
          }
                    
          // start Dialing MF digit string, Non-Blocking call, will receive Event when completed
          result = OtxToneMfDial(hHandle, (OTX_UINT8 *)g_sqDialDigit[nDev], strlen(g_sqDialDigit[nDev]), DEMO_TASK_REF, OTX_TIME_INFINITY);
          if (OTX_RESULT_CODE(result) == OTX_MF_DIALER_RES_BUSY) {
            DemoLogf("Dialer busy.\n");
          } else {
            DemoCheckRc(hHandle, result);
            DemoLogf("Done.\n");
          }
          
          OsaThreadSleep(4000);      
        
        } //for
      }
    }
    break;


  case 'k':
    {
      OTX_INT32 i,j;

      for (j = 0; j < g_demoEnv.nIteration; j++) {
        
        for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {      
          OsaThreadStartEx(DemoSineWaveSender,  (void *) &g_sqSineWaveSender[i].nDev);
          OsaThreadSleep(2000);
        }
        DemoLogf("Done.\n");
      }
    }
    break;

#endif

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;      // OTX Standard Return Code
  char commandChar;
  int i;

  OsaConsoleInit();

  DemoLogf("ThorDtmf (Thor-PCI-Plus), %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("DTMF, MF tone example program on PCI Plus board.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  DemoInitializeGlobalData();  

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set Li0 as the clock source
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // makes LED lights freeze
#if 0  // Temporary  
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);
  OsaThreadSleep(300); // To let DemoIdleFunction start up

#if DEMO_DTMF
  // start DTMF Detector threads
  DemoLogf("\nStarting DTMF Detector...\n");
  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
    
    OsaThreadStartEx(DemoDtmfDetector,  (void *) &g_sqDtmfDetector[i].nDev);
    OsaThreadSleep(10);
  }

#else
  // start MF Detector threads
  DemoLogf("\nStarting MF Detector...\n");
  for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
    
    OsaThreadStartEx(DemoMfDetector,  (void *) &g_sqMfDetector[i].nDev);
    OsaThreadSleep(10);
  }

#endif

  DemoLogf("\nPress 'h' for help\n");
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

#if DEMO_DTMF 
  // abort threads
  DemoLogf("\nExiting...\n");
  {
    for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
      g_sqDtmfDialer[i].bActive = OTX_FALSE;
      g_sqDtmfDetector[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1000);   // give threads a chance to exit
  }

#else
  // abort threads
  DemoLogf("\nExiting...\n");
  {
    for (i = 0; i < g_demoEnv.nNoOfDevs; i++) {
      g_sqSineWaveSender[i].bActive = OTX_FALSE;
      g_sqMfDetector[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1000);   // give threads a chance to exit
  }
#endif      


  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

