/* $Workfile: TimeStamp.c $       $Revision: 1 $      $Modtime: 4/03/12 1:32p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    TimeStamp.c

    A simple example console application demonstrating the use of timestamps
    provided by the OTX Driver.

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "DspIoCtl.h"
#include <windows.h>

#define DEMO_DEFAULT_LOG_FILE "TimeStamp.log"

typedef struct DemoEnvT {
  OTX_BOOL            bExitOnError;
  char               *logFileName;
  OTX_BOOL            bResyncActive;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoGetLocalTime(SYSTEMTIME *pstLocal);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static OTX_INT32 DemoProcessUserCommand(char commandChar);
static void *DemoResyncContiniously(void * lpParameter);


/*******************************************************************************
 Global Data
*******************************************************************************/
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bTimeThreadActive = OTX_FALSE;

CRITICAL_SECTION g_demoLogfCritSec;

/*******************************************************************************
*******************************************************************************/


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampUs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif
  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  // Calculate number of microseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampUs.nUint64 = nTimeStamp.nUint64 / 10 - (nTimeStampSec.nUint64 * 1000000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.6d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampUs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}


/*******************************************************************************
  @func
    Call OtxDrvGetCurrentTime() at a regular interval
*******************************************************************************/
void *DemoTimeThread(
  void *lpParameter   // thread data
)
{
  //OTX_RESULT result;
  OTX_DATETIME nTimeStampPrev;
  OTX_DATETIME nTimeStamp;
  int nSleepTime = 5000;
  SYSTEMTIME stLocal;

  do {
    // Retrieve the SystemTime
    DemoGetLocalTime(&stLocal);
    
  } while ((stLocal.wSecond % 5) != 0);


  OtxDrvGetCurrentTime(&nTimeStampPrev);
  while (g_bTimeThreadActive) {
    
    OsaThreadSleep(2000); 
    do {
      // Retrieve the SystemTime
      DemoGetLocalTime(&stLocal);
      
    } while ((stLocal.wSecond % 5) != 0);


    OtxDrvGetCurrentTime(&nTimeStamp);

    DemoLogf("DemoTimeThread() Actual Diff %d. Expected diff %d \n", nTimeStamp.sUint32.nLow - nTimeStampPrev.sUint32.nLow, nSleepTime*10*1000);
    
    memcpy(&nTimeStampPrev, &nTimeStamp, sizeof(OTX_DATETIME));

  }

  DemoLogf("DemoTimeThread() terminating.\n");

  return 0;

    
}

void DemoGetLocalTime(SYSTEMTIME *pstLocal)
/*******************************************************************************

  @func

  Get the local time from the system

  @comm

*******************************************************************************/
{
  FILETIME x;
  SYSTEMTIME stUTC;
  
  // Retrieve the SystemTime
  GetSystemTimeAsFileTime(&x);
  
  // Convert the time to local time.
  FileTimeToSystemTime(&x, &stUTC);
  SystemTimeToTzSpecificLocalTime(NULL, &stUTC, pstLocal);
  
}

/*******************************************************************************
*******************************************************************************/
void *DemoResyncContiniously(void * lpParameter)
{
  OTX_UINT32 nDelay;

  nDelay = (OTX_UINT32) lpParameter;

  while (g_demoEnv.bResyncActive) {
    DemoProcessUserCommand('r');
    OsaThreadSleep(nDelay);
  }

  DemoLogf("DemoResyncContiniously() terminating.\n");
  return 0;
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  //OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" t   Get time from driver\n");
    DemoLogf(" s   Start/Stop time thread\n");
    DemoLogf(" r   Resync time\n");
    DemoLogf(" R   Resync time continiously (turn %s)\n", g_demoEnv.bResyncActive ? "off" : "on");
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case 't':
    {
      OTX_DATETIME nTimeStamp;
      SYSTEMTIME stLocal;

      // Retrieve the SystemTime
      OtxDrvGetCurrentTime(&nTimeStamp);
      DemoGetLocalTime(&stLocal);

      DemoLogf("OTX Driver time: ");
      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("\n");

      // Print the date and time.
      DemoLogf("System time:     %02d/%02d %02d:%02d:%02d.%03d\n",
        stLocal.wMonth, stLocal.wDay, 
        stLocal.wHour, stLocal.wMinute,
        stLocal.wSecond, stLocal.wMilliseconds
        );

      DemoLogf("\n");
    }
    break;

  case 's':
    {
      g_bTimeThreadActive = !g_bTimeThreadActive;
      
      if (g_bTimeThreadActive) {
        DemoLogf("Starting Time Thread\n");
        OsaThreadStart(DemoTimeThread);
      }
    }
    break;

  case 'r':
    {
      OTX_BOOL bSaveTimeStampFrom1601;

      DemoLogf("Before resync\n");
      DemoProcessUserCommand('t');
      OtxDrvGetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, &bSaveTimeStampFrom1601);
      OtxDrvSetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, bSaveTimeStampFrom1601);
      DemoLogf("Time resyncronized\n");
      DemoProcessUserCommand('t');
      break;
    }

  case 'R':
    if (g_demoEnv.bResyncActive) {
      g_demoEnv.bResyncActive = OTX_FALSE;
    } else {
      g_demoEnv.bResyncActive = OTX_TRUE;
      OsaThreadStartEx(DemoResyncContiniously, (void *) 1000);
    }
    break;

  case 'f':
    {
      OTX_UINT32 nTimeStampTickFrequency;
      OTX_BOOL bTimeStampFrom1601;

      OtxDrvGetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, &bTimeStampFrom1601);
      OtxDrvSetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, OTX_FALSE);
      OtxDrvGetAttributeValueUint32(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_TICK_FREQUENCY, &nTimeStampTickFrequency);
      OtxDrvSetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, bTimeStampFrom1601);

      DemoLogf("Tick Frequency: %u\n", nTimeStampTickFrequency);
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}                      




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    strupr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        envP->logFileName = argv[i]+2;
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      exit(1);
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  vfprintf(stdout, formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  LeaveCriticalSection(&g_demoLogfCritSec);
}


int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  printf("TIMESTAMP, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Timestamping Example program.\n");

  InitializeCriticalSection(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    exit(-1);
  }

  // Connect to the Driver
  result = OtxDrvConnectLib();
  if (OTX_RESULT_CODE(result) == OTX_RES_FAIL) {
    DemoLogf("Error: Could not connect to lib (OTX Driver (OtxDrv.sys) possibly not loaded).\n");
    OtxDrvReleaseLib();
    exit(1);
  }
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);


  DemoLogf("\nPress '?' for help, 'q' to quit\n");
  
  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();

  return 0;
}
